/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/Manager.java,v 1.3 2000/10/19 16:59:36 andrewf Exp $

<<ICECOPYRIGHT>>
*/

package com.adobe.rice.common;

import com.adobe.rice.util.time.*;
//$JDK1.1$ import com.sun.java.util.collections.*;

/**
 * Classes Implementing the Managable interface must provide six routines.
 *
 *
 * These manager routines are designed to assist in managing cross-dependencies during
 * initialization at startup and finalization at shutdown using startup and shutdown
 * phases, <em>pre-initialization</em>, <em>initialization</em>, and 
 * <em>post-initialization</em> for startup; and <em>post-finalization</em>, 
 * <em>finalization</em>, and <em>post-finalization</em>.
 *
 * <dl>
 *   <dt>pre-initialization</dt>
 *      <dd>provides an initial call to each manager so that it can
 *          setup it's internal structures and get ready to provide external services to
 *          other managers in process of initializing that may depend on it.
 *          A manager MUST NOT call on other managers during this phase because order
 *          of invocation of mangers is not guaranteed.</dd>
 *
 *   <dt>initialization</dt>
 *      <dd>provides a second call to each manager, so that it knows that all other
 *          managers have been pre-initialized.  Thus, during this phase, a manager may 
 *          call another manager for services in order to complete it's setup to provide
 *          services.</dd>
 *
 *   <dt>post-initialization</dt>
 *      <dd>provides a last call to each manager, so that it knows that all
 *          other managers have been initialized.  Thus, during this phase, a manager may
 *          perform final setup knowing that all other managers are initialized. A manager
 *          performing setup at this stage should not affect any other routines; but
 *          MAY use other routine services.</dd>
 *
 *   <dt>run</dt>
 *       <dd>provides an entry for a single task service manager.</dd>
 *
 *
 *   <dt>pre-finalize</dt>
 *      <dd>provides an initial call to each manager, to notify it that shutdown is in progress.
 *          The manager can ready itself to support other managers as they break dependency
 *          relationships to it.  A manager MUST NOT call on other managers at this phase
 *	        for any finalization services because the order of invocation of managers during
 *          this phase is not guaranteed.</dd>
 *   <dt>finalize</dt>
 *      <dd>provides a second call to each manager so that it can call on other managers for
 *          finalization services, knowing that those managers are ready for the call. This
 *          manager may complete it's internal finalization in this phase if it doesn't have
 *          to wait for other managers to disengage resources.</dd>
 *   <dt>post-finalize</dt>
 *      <dd>provides a final call to each manager so that it can complete internal finalization
 *          knowing that all other managers have had the opportunity to dis-engage it.
 *          A manager MUST NOT call another manager during this phase because the order of
 *          invocation is not guaranteed and therefore the called manager may not exist.
 *  </dl>
 **/

/**
 * Manager.java
 *
 *
 * Created: Sun Nov  7 12:06:17 1999
 *
 * @author V. Bruce Hunt, Andrew Francis
 * @version
 */

public interface Manager extends Runnable
{
    // Execution Activity Logic
    
    // Startup Initialization Logic.
    public void preInitialize();    // initially run for pre-initialization
    public void initialize();		// run after all managers have been pre-initialized.
    public void postInitialize();	// run after all managers have been initialized.

    // Running Logic.
    public void run();

    // Shutdown Finalization Logic:
    public void preFinalize();		// Run at shutdown to notify of pending finalization.
    // public void finale();		       run after all managers have been pre-finalized.
    public void postFinalize();		// run after all managers have been finalized.
    
} // Manager




