/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/Parsergen/Grammar.java,v 1.1.1.1 2000/09/19 19:39:44 bhunt Exp $


ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

One or more classes in ICEcubes have been written by:

V. Bruce Hunt,                  Technical Lead & Vice Chair, ICE-AG, 
    vbhunt@sourceforge.net      Adobe Systems, Inc.
    
Jonathan Herbach                Developer           
    jherbach@adobe.com          Princeton University & Adobe Systems, Inc.
    
Will Conant                     Developer
                                University of Utah & Adobe Systems Inc.

Andrew Francis                  Developer           
    afrancis@adobe.com          Adobe Systems, Inc.
    
Anna Shamble                    Developer           
    ashamble@adobe.com          Santa Clara University & Adobe Systems, Inc.

Authors of various classes are noted in the headings 

*/
/*
*
* $Author: bhunt $
* $Revision: 1.1.1.1 $
* $Log: Grammar.java,v $
* Revision 1.1.1.1  2000/09/19 19:39:44  bhunt
* Initial Population
*
* Revision 1.1.1.1  2000/08/17 17:50:08  bhunt
* Initial Code Upload
*
* Revision 1.2  1999/09/19 07:47:00  bhunt
* Grammar is now it's own java file.
*
* Revision 1.1  1999/09/19 07:18:57  bhunt
* Initial revision
*
*/
/*----------------------------------------------------------------------------------+
 |                                                                                  |
 |                                  Grammar Class                                   |
 |                                                                                  |
 +----------------------------------------------------------------------------------*/

package com.adobe.rice.common.Parsergen;

import java.io.*;
import java.text.*;
import java.util.*;
//$JDK1.1$ import com.sun.java.util.collections.*;

/*----------------------------------------------------------------------------------+
 |                                                                                  |
 |                   Generate LALR(1) parse tables for a grammar                    |
 |                                                                                  |
 +----------------------------------------------------------------------------------*/
/**
@class Grammar
   This class contains the grammar under construction.  This grammar is always 
   augmented as it is constructed.  The root symbol is Symbol.root ( @ROOT ) 
   and the first production is always:  
           @ROOT -> r ,  
   where r is the left hand side of the first production added;
   or the initial symbol when the grammar is first constructed.

   A grammar is a tupple:
 
                   G ::=  [ T, N, R, P ::= {  N x (N u T)* }, L ::= { T* } ]
 
   where T is a set of terminal symbols, 
         N is a set of non terminal symbols, 
         R is a non terminal symbol called the root symbol, 
         P is a set of productions which is a mapping from the set of 
           nonterminals into a set of strings composed of terminals and 
           non-terminals and 
         L is the set of all strings of terminal symbols that are derived 
           from R using P.
   
   The productions are connected such that there is no production whose 
   first element (one of the non-terminal symbols.) is not derivable from R.  
   A terminal symbol or non-terminal symbol x is directly derivable from a 
   non-terminal symbol N in a grammar G, if there exists a production in the 
   grammar whose first element is N and one of the elements of the string of
   symbols in the second element is x.  A symbol (terminal or non-terminal) 
   x is derivable from a non-terminal symbol N if there is a sequence of 
   directly derivable non-terminals N, N1, N2, ... Nk such that N1 is 
   directly derivable from N, Ni+1 is directly derivable from Ni and x is 
   directly derivable from Nk.  (i.e. x is in the transitive closure of N ).  
   A string, alpha, is a sentence of L using G if there are no non-terminal 
   symbols in alpha and alpha is derivable from G (i.e. from the root, R, 
   in G ).  L is the set of sentences of G.

   There are distinquished in this implementation three additional symbols, 
   @ROOT, @EMPTY and @END.

   @ROOT is a unique non-terminal symbol in N that is used to create an 
     augmented grammar with the addition of a production, @ROOT -> R, where 
     R is the root symbol.  This is done so that the parser can recognize 
     when to accept a grammar as valid.

   @EMPTY is a unique terminal symbol in T that represents the empty 
     terminal symbol.  It is used keep track of possibly empty derivations.

   @END is a unique terminal symbol that never appears in a Production.  
     It represents the end of parse input and is used to signal that no 
     more terminal symbols are in the input.

**/
public class Grammar
{
    Symbol		root;		// nonterminal root of the grammar
    List		prods;		// productions of the grammar
    TreeSet		symbols;	// Grammar Symbols (both nonterminal and terminal)
    HashMap		firstmap;	// FIRST mapping

    int			terminalCount;	// number of terminal symbols in grammar
    int			nontermCount;	// number of non terminal symbols in grammar.

    int			maxSymLength;	// length of the biggest symbol
    public Grammar()
    {
        root = Symbol.root;
        prods = new ArrayList( LRconst.PRODUCTION_COUNT );
        symbols = new TreeSet( );
        firstmap = new HashMap( LRconst.SYMBOL_COUNT );
        //
        terminalCount = 0;
        nontermCount = 0;
        maxSymLength = 0;		// we don't have any symbols yet.
    }

    public Grammar( Symbol r )
    {
        this();
        // augment the grammar so that state tables can be built.
        //
        //	root -> r
        //
        Production ar = new Production( root );
        ar.add( r );
        // put the symbols of the production into symbols:
        addSymbols( ar );
        prods.add( ar );
        // 
        // place the endParse Symbol (@END) in the symbol list.
        //
        symbols.add( Symbol.endParse );
    }

    //
    // Methods
    //
    public int getTerminalCount() { return terminalCount; }
    public int getNontermCount() { return nontermCount; }

    //
    // note that the start symbol is the root of the
    // user provided grammar before the grammar is augmented.
    // It is NOT @ROOT -- which is the 
    // start symbol of the augmented grammar.
    public Symbol getStartSymbol() 
    {
        if ( prods.size() > 0 )
        {
            return ( (Production) prods.get(1) ).getLHS( );
        }
        else
            return null;
    }   

    public void add( Production p )
    {
        // if this is the first addition, augment the grammar.
        if ( prods.size() == 0 )
        {
            Production ar = new Production( root );
            ar.add( p.getLHS() );
            addSymbols( ar );
            prods.add( ar );
            // make sure that we computer FIRST( @END ) too.
            symbols.add( Symbol.endParse );
        }
        prods.add( p );
        // put the symbols of the production into symbols.
        addSymbols( p );
    }
	
    public Production get( int i )
    {
        if ( ( i >= 0 ) && ( i < prods.size() ) )
            return (Production) prods.get(i);
        return null;
    }

    // return the index of the production in the grammar.
    // this is used in building the parser table.
    //
    public int getProductionIndex( Production p )
    {
        for ( int i = 0; i < prods.size(); i++ )
        {
            Production pr = (Production) prods.get(i);
            if ( pr.equals(p) )
                return i;
        }
        return -1;
    }

    public boolean containsSymbol( Symbol s )
    {
        if ( symbols.contains( s ) )
            return true;
        return false;
    }
	
    void addSymbols( Production p )
    {
        // put the symbols of the production into symbols.
        // first the lhs non-terminal
        if ( !symbols.contains( p.getLHS() ) )
        {
            Symbol lhs = p.getLHS();
            // make sure that bozo's don't get through.
            if ( lhs.equals( Symbol.endParse ) || lhs.equals( Symbol.empty ) )
            {
                // really dumb production assignment.
                //
                // throw an exception!
            }
            // force the lhs symbol into being a non-terminal
            if ( lhs.isTerminal() )
            {
                Symbol nlhs = new Symbol( lhs.getS(), LRconst.NONTERMINAL );
                p.setLHS( nlhs ) ;
                lhs = nlhs;
            }
            // add it to our list.
            symbols.add( lhs );
            nontermCount++;
            // following for cosmetic layout purposes
            if ( lhs.length() > maxSymLength )
                maxSymLength = lhs.length();
        }
        // then each of the symbols.
        for ( int i = 0; i < p.size(); i++ )
        {
            Symbol s = p.get(i);
            if ( !symbols.contains( s ) )
            {
                symbols.add( s );
                // count the terminals and non terminals.
                if ( s.isNonTerminal() ) 
                    nontermCount++;
                else
                    terminalCount++;
                // following for cosmetic layout purposes
                if ( s.length() > maxSymLength )
                    maxSymLength = s.length();
            }
        }
    }
	
    public int size() { return prods.size(); }
    public TreeSet getSymbols() { return symbols; }
    public int symbolsSize() { return symbols.size(); }
    public Symbol getRoot() { return root; }
    public void setRoot( Symbol r ) { root = r; }
	
    public HashMap getFirst() { return firstmap; }
    public void setFirst( HashMap m ) { firstmap = m; }
    public int getMaxSymLength( ) { return maxSymLength; }

    // construct a string suitable for output with indents
    public String toStringCore( int ind )
    {
        final int DEBUGLEVEL = 8;

        String gr = LRconst.indent(ind) + "Grammar --> [ " + root.toString() + ",\n" ;
        ind++;
        gr += LRconst.indent(ind) + "Productions = [\n";
        ind ++;
        for ( int i = 0 ; i < prods.size(); i++ )
        {
            Production p = (Production) prods.get(i);
            gr += p.toStringCore(ind) + "\n";
        }
        ind--;
        gr += LRconst.indent(ind) + "],\n";
        Iterator si = symbols.iterator();
        gr += LRconst.indent(ind) + "Symbols = [\n";
        ind++;
        while ( si.hasNext() )
        {
            Symbol sym = (Symbol) si.next();
            gr += sym.toStringCore(ind);
            if ( sym.length() < getMaxSymLength() )
            {
                if ( LRconst.DEBUG > DEBUGLEVEL )
                {
                    gr += "|sym.length() = " + 
                          (new Integer(sym.length())).toString() + "|";
                    gr += "|getMaxSymLength() = " + 
                          ( new Integer(getMaxSymLength())).toString() + "|";
                }
                gr += LRconst.genBlanks( getMaxSymLength() - 
                      sym.length() + (sym.isNonTerminal()?0:1) );
            }
            if ( firstContainsSymbol( sym ) )
            {
                gr += " : ";
                SymbolFirst symFirst = firstGet( sym ); 
                gr += symFirst.toString();
            }
            gr += "\n";
        }
        ind--;
        gr += LRconst.indent(ind) + "]\n";
        ind--;
        gr += LRconst.indent(ind) + "]\n";
        return gr;
    }

    public String toString() { return toStringCore( 0 ); }
	
    void firstPut( Symbol x, SymbolFirst l ) { firstmap.put( x, l ); }
    SymbolFirst firstGet( Symbol x ) { return (SymbolFirst) firstmap.get( x ); }
    boolean firstContainsSymbol( Symbol x ) { return firstmap.containsKey( x ); }
    boolean firstContainsList( SymbolFirst sf ) { return firstmap.containsValue( sf ); }
    int firstSize() { return firstmap.size(); } 
    //
    // After adding all of the productions for a grammar, then call
    // complete to sort the symbols so that the terminal symbols are first
    // in the grammar. The terminal symbols @END is, by definition, the
    // last terminal.  Similarly, @ROOT, by definition, is the last non-terminal
    // Also compute FIRST for the grammar.
    // 
    public void complete()
    {
        // First go get the terminal symbols:
        //
        Iterator si = symbols.iterator();

        // Compute the core FIRST based on the Root symbol.
        if ( firstSize() == 0 )
            FIRST( getRoot() );
        // Now compute any other FIRSTs needed 
        // 
        Iterator iterSym = symbols.iterator();
        while ( iterSym.hasNext() )
        {
            Symbol s = (Symbol) iterSym.next();
            if ( !firstContainsSymbol( s ) )
            {
                // There exist non terminals and terminals that aren't in the FIRST of
                // the root symbol.  
                FIRST( s );
            }
        }
    }

    /**
        @method boolean inFIRST( List symlist, Symbol x )
	@parameter symlist - array list of symbols.
	@parameter x - a terminal symbol; including 
       empty == Symbol( "@EMPTY", TERMINAL ) or
	   endParse == Symbol("@END", TERMINAL);
	
	This routine returns true if x is in FIRST( symlist );
		otherwise it returns false.
	
	This tells us that x can be derived from symlist; i.e. 

                 symlist =>...=> x a1 a2 ... ak.
	
	This routine relies on the computation of the HashMap first.  
    first is a mapping from each symbol of the grammar to a list of 
    symbols that are in the FIRST of that symbol.  That is for any 
    symbol c in the grammar.  first(c) returns a HashSet
	containing all of the symbols that are in FIRST(c).
	
	The algorithm works as follows.  symlist is a list of symbols; 
    and we want to see if x is in the FIRST( symlist );  Recall that 
    FIRST( alpha) is defined as the set of terminals that begin the 
    strings derived from alpha; where alpha is any string of symbols 
    (i.e. either terminal or non-terminal ).
	
	So, to find out if x is in FIRST( symlist) where symlist is a 
    sequence c1, c2, ..., ck, it follows that x must be in FIRST(c1).  
    It could only be in FIRST(c2) if FIRST(c1) contains "@EMPTY".  
    Similarly, if x is not in FIRST(c1) and FIRST(c2), the only way
	it could be in FIRST(c3) is if both FIRST(c1) and FIRST(c2) contains
    the empty string ( so that it is possible for both to derive the 
    empty string).  If either doesn't, then the symbol is not in 
    the FIRST( symlist).
	
    **/
	
    public boolean inFIRST( List symlist, Symbol x )
    {
        for ( int i = 0; i < symlist.size(); i++ )
        {
            Symbol c = (Symbol) symlist.get(i);
            SymbolFirst sf = firstGet(c);

            if ( sf.contains( x ) )
                return true;
			
            if ( ! sf.contains( Symbol.empty ) )
                return false;
        }
        return false;  // got here because all firsts contained empty
    }
    /**
        @method FIRST
	compute FIRST(X) for all grammar symbols X for given grammar
	see page 189 in H,S,A

	To compute FIRST(X) for all grammar symbols X, apply the following 
    rules until no more terminals or empty can be added to any FIRST set:
		
	1.  if X is a terminal, then FIRST(X) is { X }.
		
	2.  if X -> empty is a production, then add empty to FIRST(X).
		
	3.  if X -> Y1 Y2 ... Yk is a production, then place a in FIRST(X) 
        if for some i, a is in FIRST( Yi ) and empty is in all of 
        FIRST( Y1 ), FIRST( Y2 ), ..., FIRST( Y(i-1) ).
	    That is, Y1 ... Y(i-1) *=> empty.  If empty is in FIRST( Yj ) 
        for all j such that 1 <= j <= k, then add empty to FIRST( X ).  
        Everything in FIRST(Y1) is in FIRST(X). If Y1 does not derive 
        empty, then add nothing more to FIRST(X).  If Y1 *=> empty, then
	    we add FIRST(Y2) and so on.

	The Grammar contains a HashMap named first.  The grammar field, 
    first, consists of a collection of pairs of class instances: 
    [ Symbol, SymbolFirst ] where SymbolFirst is itself a pair, 
    [ Symbol, HashSet ].   The HashSet has an entry for every symbol
	that is in FIRST of Symbol. (i.e. For every symbol A, and every 
    instance of SymbolFirst, [ A, HashSet ], the HashSet instance 
    has an entry for every x that is a symbol of the grammar 
    iff x is in FIRST(A).)  

	Thus we have a tree representation that allows rapid lookup of 
    the set of symbols that are in the FIRST(A) for any A in the 
    symbol set of the grammar. Note that A can either be a terminal 
    or non-terminal.   

    **/
    void FIRST( Symbol x )
    {
        // if we're already computing it, don't recurse (just curse!).
        if ( firstContainsSymbol( x ) )
            return;
		
        int xSize = 0;
        SymbolFirst xfirst = new SymbolFirst(x);
		
        // repeatedly apply the rules until xfirst stops growing.
        do
        {
            xSize = xfirst.size();
            if ( x.isTerminal() )
            {	
                // put it in the first list if not already there
                if ( ! xfirst.contains( x ) )
                    xfirst.add( x );
                if ( ! firstContainsSymbol( x ) )
                    firstPut( x, xfirst );
            }
            else // x is a non-terminal.
            {
                // for all productions of the form x -> gamma
                for ( int i = 0; i < this.size(); i++ )
                {
                    Production p = this.get(i);
                    // if x is not the left side of the production, continue looking.
                    if ( !(p.getLHS()).equals( x ) )
                        continue;
                    // production p is of the form x -> gamma
                    if ( p.size() == 1 )
                    {   // if X -> empty is a production, then add empty to FIRST(x).
                        if ( (p.get(0)).equals( Symbol.empty ) )
                        {
                            if ( !xfirst.contains( Symbol.empty ) )
                                xfirst.add( Symbol.empty );
                            if ( ! firstContainsSymbol( x ) )
                                firstPut( x, xfirst );
                            continue;
                        }
                    }
                    // production p is of the form X -> Y1 Y2 ... Yk
                    // step one is to put( x, xfirst ) in FIRST to
                    // signal that we are computing it if it
                    // isn't already there.  (This is to prevent recursively 
                    // computing FIRST(X)).
                    if ( ! firstContainsSymbol( x ) )
                        firstPut( x, xfirst );
                    for ( int j = 0; j < p.size(); j++ )
                    {
                        Symbol Yj = p.get(j);
			    
                        if ( !firstContainsSymbol( Yj ) )
                            FIRST(Yj);
                        // if Yj is a terminal, then Yj is in FIRST(x), 
                        // we can't derive empty, so break.
                        if ( Yj.isTerminal() )
                        {
                            if ( !xfirst.contains( Yj ) )
                                xfirst.add( Yj );
                            break;
                        }
                        // Can only get here if Yj is a non-terminal
                        // now add every unique terminal in FIRST( Yj ) to xfirst.
                        Iterator iYj = firstGet( Yj ).iterator();
                        while ( iYj.hasNext() )
                        {
                            Symbol s = (Symbol) iYj.next();
                            if ( !xfirst.contains( s ) )
                                xfirst.add( s );
                        }
                        // if it is not the case that Yi *=> empty then cannot add more.
                        if ( !(firstGet( Yj )).contains( Symbol.empty ) )
                            break;
                        if ( j == p.size() )
                        { // possible that @EMPTY must be added to xfirst.
                            if ( !xfirst.contains( Symbol.empty ) )
                                xfirst.add( Symbol.empty );
                        }
                    } // end scan through the k elements of X -> Y1 Y2 ... Yi ... Yk
                } // end scan of productions looking for productions of X -> Y1 ... Yk
            } // end X is a non terminal
        } 
        while ( xSize < xfirst.size() );  // end repeatedly do until growth stops.
    } // end FIRST
}














