/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/XMLassembly/ContentReference.java,v 1.3 2000/12/02 23:31:44 bhunt Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

*/
/*----------------------------------------------------------------------------------+
 |                                                                                  |
 |                         Content Reference for XMLassembly                        |
 |                                                                                  |
 +----------------------------------------------------------------------------------*/

package com.adobe.rice.common.XMLassembly;

/**
 		Class	ContentReference
  ContentReference is a typed pointer class that defines the components of an
  XML document. It defines all of the possible generic XML components as defined
  in the XML 1.0 Spec. It is intended to be used as the means for composing a 
  (deflated) XML document that can later be inflated with the actual content.  
  Effectively it is the foundation of a pattern for an XML document that will 
  later be instanciated and then inflated.  The first step builds the XML 
  document but does not insert the content -- rather places references to it 
  in the document instance where it will be inflated.

  The reason for doing this is to be able to have 1) document instance templates
  (For example, we intend that for each ICE Payload type, there is an XML 
  document template that can be used to extremely rapidly construct a 
  document instance. The instance has all of the standard XML document 
  apparatus with pointers to the appropriate values defined in the ICE 
  internal objects that can be "inflated" as soon as it becomes necessary 
  to launch the payload.  In addition, the XML document instance can be 
  populated with character data and CDSECT content references (using 
  this object). These, too, can be expanded with content at the time 
  it is necessary to transmit the payload.  The second reason for doing 
  this is so that RICE can manage a very large number of payload documents 
  and hence support a very large number of subscribers and syndicators.
 **/

public class ContentReference implements XMLexpandable
{
    private ContentType 	ctype = null;
    private XMLexpandable  	ref  = null;
    private String          text = null;
    private int             nest = 0;
    private boolean         okToFormat = false;
  
    public ContentReference() throws ContentException
    {
        try { ctype = new ContentType("CharData"); }
        catch ( ContentException ce ) { throw ce; }
        ref = null;
    }
	
    public ContentReference( String s ) throws ContentException
    {
        this();
        try { ctype = new ContentType( s ); }
        catch ( ContentException ce ) { throw ce; }
    }
    
    public ContentReference( String s, XMLexpandable x ) throws ContentException
    {
        this();
        try { ctype = new ContentType( s ); }
        catch ( ContentException ce ){ throw ce; }
        ref = x;
    }
    
    public ContentReference( String s, String actual ) throws ContentException
    {
        this();
        try { ctype = new ContentType(s); }
        catch ( ContentException ce ) { throw ce; }
        ref = null;
        text = actual;
    }
	
    public boolean setCtype( String s ) 
    {
        try { ctype = new ContentType(s); }
        catch ( ContentException ce ) { return false; }
        return true;
    }
    
    public String getCtypeString() { return ctype.getString(); }
    
    public ContentType getCtype() { return ctype; }
    
    public void setCtype( byte b ) { ctype.set( b ); }
    
    public void setRef( XMLexpandable e ) { ref = e; }
    public XMLexpandable getRef() { return ref; }
    public String getText() { return text; }
    
    public void setNest( int newNest ) { this.nest = newNest; }
    public int getNest() { return this.nest; }
    
    public void setOkToFormat(boolean ok ) { this.okToFormat = ok; }
    public boolean getOkToFormat() { return this.okToFormat; }
    
    //
    // implement XMLexpandable
    //
    public String toXMLstring() // returns string as XML data
    {
        switch ( this.ctype.get() )
        {
        case ContentType.PROLOG:	    // envelope for xmldecl, doctype 
        // decl, comments & PIs at beginning of XML document
        case ContentType.XMLDECL:	    // XML version, encoding and 
        // standalone declaration.
        case ContentType.DOCTYPEDECL:	// Doctype declarations
        case ContentType.TEXTDECL:	    // External Entity definition
        case ContentType.EXTSUBSETDECL :	// External subset declaration
        case ContentType.IGNORE:	    // IGNORE section
        case ContentType.INCLUDE:	    // INCLUDE section
        case ContentType.PARMENTITYREF:	// parameter entity reference
        // the following types are valid content for an element:
        case ContentType.ATTRIBUTE:	    // XML element attribute content 
                                                // reference (name & value)
        case ContentType.ATTRVALUE:	    // XML element attribute value 
            return ref.toXMLstring();
        case ContentType.ELEMENT:	    // XML element
            Element el = (Element) this.ref;
            if ( el.getOkToFormat() )
            {
                el.setNest( this.nest );
            }
            return el.toXMLstring();
                                                // content references possible:
        case ContentType.COMMENT:	    // XML comment
            return ref.toXMLstring();
        case ContentType.PROCINST:	    // processing instruction declaration
            return ref.toXMLstring();
        case ContentType.GENENTITYREF:	// general entity reference
            return ref.toXMLstring();
        case ContentType.CHARREF:	    // character reference
            return ref.toXMLstring();
        case ContentType.CHARDATA:	    // regular characters (not markup)
            return this.toCharData();
        case ContentType.CDATA:	        // CDATA section
            return this.toCDSect();
        default:
            return "<!-- $ERROR$ Unknown Content Reference type! -->";
        }
    };    
    
    public boolean canDoCharData() // returns true if implements toCDSect() AND toCharData
    { 
        return true; 
    } 
    
    public String toCDSect() // returns content in a CDATA sections as much as possible
    {
        return CDATAquote( this.text );
    }
        
    public String toCharData() // returns content in CharData Section
    {
        if ( this.okToFormat )
        {
            StringBuffer result = new StringBuffer( this.text.length() + 100 );
            
            String theQuotedText = XMLquote(this.text).trim();
            
            for( int i = theQuotedText.indexOf("\n"); i > -1; i = theQuotedText.indexOf("\n") )
            {
                result.append( Element.getWhiteSpace(this.nest, 2) );
                // if ( i > 0 )
                result.append( theQuotedText.substring(0, i ) );
                // if ( i == (theQuotedText.length()-1) )
                //    break;
                result.append("\n");
                theQuotedText = theQuotedText.substring(i+1);
            }
            if ( theQuotedText.length() > 0 )
            {
                result.append( Element.getWhiteSpace(this.nest, 2) );
                result.append( theQuotedText );
            }
            return result.toString();
        }
        return XMLquote( this.text );
    }
    
    /**
     * Quote a string of text by turning it into chardata.  That is, certain
     * characters are turned into their equivalent character entities so that
     * the text won't be confused with markup.
     *
     * @param source - String that contains the string of characters to be XMLquoted.
     *
     * @return result - StringBuffer containing the XML quoted string.
     *
     **/
    
    static public String XMLquote( String source )
    {
        StringBuffer result = new StringBuffer( source.length() );
        char ch ;
        for ( int i = 0; i < source.length(); i++ )
        {
            ch = source.charAt(i);
            switch ( ch )
            {
            case '<':
                result.append("&lt;");
                break;
            case '>':
                result.append("&gt;");
                break;
            case '&':
                result.append("&amp;");
                break;
            case '"':
                result.append("&quot;");
                break;
            case '\'':
                result.append("&apos;");
                break;
            default:
                result.append( ch );
                break;
            }
        }
        return result.toString();
    }
    
    /**
    * Turn a string of characters into a sequence of CDATA sections
    * with chardata interspersed to handle the ending sequence ( "]]>" ).
    * 
    * @param source - String that contains the content to be turned into
    *   CDATA.  
    *
    * @return result - String that contains a sequence of CDATA sections
    *   interspersed with chardata ( ]]&amp;gt; ) representing the ending
    *   sequence.  The algorithm works as follows:
    *
    *   source     result
    *     null       ""  -- the empty string
    *      ""        ""  -- the empty string
    *   "]]>xyz..." "]]&gt;<![CDATA[xyz...]]>"
    *   "]]>"       "]]&gt;" 
    *   "xxxxyz]]>" "<![CDATA[xxxxyz]]>]]&gt;"
    *   "x]]>y"     "<![CDATA[x]]>]]&gt;<![CDATA[y]]>"
    *   "xyz"       "<![CDATA[xyz]]>"
    *
    *    
    **/
    static public String CDATAquote( String source )
    {
        final String boxes = "]]&gt;";
        final String cdata = "<![CDATA[";
        final String ender = "]]>";
        if ( source == null )
            return "";
            
        StringBuffer result = new StringBuffer( source.length()+10 );
        
        if ( source.length() > 0 )
        {
            for ( int i = source.indexOf(ender); i > -1; i = source.indexOf(ender)  )
            {
                // something of the form "xxxxxxxx]]>yyyyyy" in source.
                if ( i > 0 )  // must be at least one non-ender character to append.
                {
                    result.append( cdata );
                    result.append( source.substring(0, i-1) );
                    result.append( ender );
                }
                result.append( boxes );  // keep the ]]> as chardata
                // reform the source to be the stuff past the ender.
                if ( ( i + ender.length() ) < source.length() )
                    source = source.substring( i + ender.length() );
                else
                    source = "";
            }
            if ( source.length() > 0 )
            {
                result.append( cdata );
                result.append( source );
                result.append( ender );
            }
        }
        return result.toString();
    }
}













