/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/XMLassembly/XMLscanner.java,v 1.3 2000/12/02 23:31:44 bhunt Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

*/
/*----------------------------------------------------------------------------------+
 |                                                                                  |
 |                             XMLscanner - Scanner for XML                         |
 |                                                                                  |
 +----------------------------------------------------------------------------------*/

package com.adobe.rice.common.XMLassembly;

import java.io.*;
import java.util.*;
import java.text.*;
import com.sun.java.util.collections.*;

// $TEST$
//
//  23-06-2000 vbh - Changes to support automated scanning of PI's and XML decls.
//                   Now scanner recognizes { '<?' whitespace? name } and distinguishes
//                   between name being case invariant of "xml" and any other name (which is
//                   a PI target).  
//                   When an XML decl is detected.scanner tokenizes normally
//                   so that internal parameters to the XML decl can be scanned.  For example:
//                   the scanner returns the following tokens for the string "<?xml version="1.0" ?>
//                   (assuming that Whitespace is not a token).
//
//                   Type         Status      String       Comment
//                   
//                   HUH_NAME      OK          "xml"       Note that the opening "<?" is reported by type.
//                   NAME          OK          "version"
//                   SPECIAL       OK          "="
//                   QUOTED_STRING OK          "1.0"
//                   PI_END        OK          "?>"
//                   
//                   In this case a '?>' is syntactically required to complete the XML decl after the attributes
//                   have been scanned.  This results in a new token, ScanToken.PI_END being returned.
//                   The scanner now always looks for a token of this type when processing normal
//                   whitespace. (So it will be reported if it occurs in random scanned text.)
//
//                   When a PI target is detected, { '<?' whitespace? target }, the target name
//                   is returned and the scanner is set to resume scanning for the processing instruction.
//                   { '<?' whitespace? target whitespace+ processing_instruction '?>' }
//                   The reason for this is that the Processing instruction is simply a sequence
//                   of non-whitespace text following the target up to the first occurrence of '?>'
//                   which terminates the PI and the processing instruction in the PI.
//                   The scanner returns the following tokens for the string "<?silly attempt at a ? > PI ?>"
//                   (assuming that whitespace is not a token):
//
//                   Type         Status  String                 Comment
//                   
//                   HUH_NAME      OK     "silly"                the opening "<?" is reported by type.
//                   PROC_INSTR    OK     "attempt at a ? > PI " the closing '?>" is reported by type since
//                                                               it closes the processing instruction.
//
//                   Note also that the particularly ugly case "<? <?????>" will be reported as:
//
//                   MARKUP_QUESTION OK   "<?"
//                   MARKUP_QUESTION OK   "<?"
//                   SPECIAL         OK   "?"                    repeated 3 times!
//                   SPECIAL         OK   "?"
//                   SPECIAL         OK   "?"
//                   PI_END          OK   "?>"
//
//
//
//	Things to fix.
//
//	(D) - DONE,  (HL) - Higher Level, (P) - partial
//
// (D)  1.  Have begin/end line number and positions for multi-line tokens.
// (D)  2.  build interface to expand entities inline
//      3.  construct option to return entity references as tokens or expand entities inline.
// (HL) 4.  do name = "value" attribute scans
// (HL) 5.  do <element ...> scans
// (HL) 6.  do <element1 attribute-list> char* <element2>char*<element2/> </element1> & <element /> scans
//	    7.  perform entity VC and WFC checking for various constraints on entities
// (P)  7a.  Entity Declared ( This implementation requires that a DTD is available, so the
//           constraint for well-formedness is moot.) 
//	    7b.  No unparsed entity references
//	    7c.  No recursion
// (D)  7d.  In DTD checking
// (D)  8.  Provide option to Keep track of Whitespace as a token.
//      9.  Provide option to return Section Markers ( Section Begin,  Section End )
//     10.  Provide option to return IGNORE section text.
	 

public class XMLscanner
{
    /*----------------------------------------------------------------------------------+
     |                                                                                  |
     |                              XML Scanner Constants                               |
     |                                                                                  |
     +----------------------------------------------------------------------------------*/
    private static final byte C_ILLEGAL    = 0; /* XML illegal characters */
    private static final byte C_WHITESPACE = 1;	/* tab, return, line feed, blank */
    private static final byte C_SPECIAL    = 2; /* a special character */
    private static final byte C_QUOTE 	   = 3; /* one of apostrophe or quotation mark */
    private static final byte C_DIGIT      = 4; /* one of 0,1,2,3,4,5,6,7,8,9 */
    private static final byte C_NAMECHAR   = 5;	/* Name character is base, ideographic, combining, 
                                                   extender, digit, extended digit, period, 
                                                   underscore, hyphen, colon */
    private static final byte C_LETTER     = 6;	/* Letter is base, ideographic, period, underscore */
    private static final byte C_PERCENT	   = 7;	/* Parameter entity escape character -- '%' */
    private static final byte C_AMPERSAND  = 8;	/* General entity escape character -- '&' */
    private static final byte C_SHARP      = 9;	/* Sharp, Hash or Pound sign -- '#' 
						                           signals numeric field in enity character reference */
    private static final byte C_HEXCHAR    = 10; /* 'A'..'F' or 'a'..'f' -- used for hex digits 
						                           in Characters references. */
    private static final byte C_SEMICOLON  = 11; /* Semi colon -- ';' ends entity references */
    private static final byte C_MARKUPSTART = 12;/* Left angle bracket character
						                          -- '<' lead in to XML markup */
    private static final byte C_MARKUPEND = 13; /* Right angle bracket Character -- '>' */
    private static final byte C_BANG	  = 14; /* Exclamation character -- '!' */
    private static final byte C_QUESTION  = 15; /* Question character -- '?' */
    private static final byte C_DASH	  = 16; /* Hyphen, minus or dash character -- '-' */
    private static final byte C_LEFTBOX	  = 17; /* Left square bracket -- '[' */
    private static final byte C_RIGHTBOX  = 18; /* Right square bracket -- ']' */
    private static final byte C_X         = 19; /* x character -- 'x' signals hex character entity */
    private static final byte C_SLASH     = 20;	/* Slash or division character -- '/' */
    private static final byte C_STAR      = 21; /* Asterisk or star character -- '*' */
    private static final byte C_PLUS      = 22;	/* plus sign -- '+' */
    private static final byte C_EQUAL     = 23; /* equal sign -- '=' */
    private static final byte C_BAR       = 24;	/* bar symbol -- '|' */
    private static final byte C_LEFTPAREN = 25;	/* Left parenthesis -- '(' */
    private static final byte C_RIGHTPAREN= 26;	/* Right parenthesis -- ')' */
    private static final byte C_END       = 27; /* End of line (scan string end seen) */

    private static final byte C_TYPES     = 28;	/* number of character types in table */
	
    private static final char[][] whiteSpaceRange = 
    {
    	{ '\t', '\n' },		
    	{ '\r', '\r' },
    	{ '\u0020', '\u0020' }
    };
    
    private static final char[][] specialRange =
    {
    	{ '!', '!' },			/* 0x21 - 0x21 */
    	{ '#', '&' },			/* 0x23 - 0x26 */
    	{ '(', '/' },			/* 0x28 - 0x2F */
    	{ ';', '@' },			/* 0x3B - 0x40 -- note 0x3A colon missing. */
    	{ '[', '^' },			/* 0x5B - 0x5E -- note 0x5F underscore missing. */
    	{ '`', '`' },			/* 0x60 - 0x60 */
    	{ '{', '~' }			/* 0x7B - 0x7E */
    };
    
    private static final char[][] quoteRange =
    {
    	{ '"' , '"'  },			/* 0x22 - 0x22 -- the quotation mark */
    	{ '\'', '\'' }			/* 0x27 - 0x27 -- the apostrophe */
    };

    private static final char[][] digitRange  = 
    {
    	{ '\u0030', '\u0039' }
    };
    
    public static final char[][] baseCharRange =
    {
        { '\u0041', '\u005A' }, { '\u0061', '\u007A' },	{ '\u00C0', '\u00D6' },	{ '\u00D8', '\u00F6' },
        { '\u00F8', '\u00FF' }, { '\u0100', '\u0131' },	{ '\u0134', '\u013E' },	{ '\u0141', '\u0148' },
        { '\u014A', '\u017E' }, { '\u0180', '\u01C3' },	{ '\u01CD', '\u01F0' },	{ '\u01F4', '\u01F5' },
        { '\u01FA', '\u0217' }, { '\u0250', '\u02A8' },	{ '\u02BB', '\u02C1' },	{ '\u0386', '\u0386' },
        { '\u0388', '\u038A' }, { '\u038C', '\u038C' },	{ '\u038E', '\u03A1' },	{ '\u03A3', '\u03CE' },
        { '\u03D0', '\u03D6' }, { '\u03DA', '\u03DA' }, { '\u03DC', '\u03DC' }, { '\u03DE', '\u03DE' },
        { '\u03E0', '\u03E0' }, { '\u03E2', '\u03F3' }, { '\u0401', '\u040C' }, { '\u040E', '\u044F' },
        { '\u0451', '\u045C' }, { '\u045E', '\u0481' }, { '\u0490', '\u04C4' }, { '\u04C7', '\u04C8' },
        { '\u04CB', '\u04CC' }, { '\u04D0', '\u04EB' }, { '\u04EE', '\u04F5' }, { '\u04F8', '\u04F9' },
        { '\u0531', '\u0556' }, { '\u0559', '\u0559' }, { '\u0561', '\u0586' }, { '\u05D0', '\u05EA' },
        { '\u05F0', '\u05F2' }, { '\u0621', '\u063A' }, { '\u0641', '\u064A' }, { '\u0671', '\u06B7' },
        { '\u06BA', '\u06BE' }, { '\u06C0', '\u06CE' }, { '\u06D0', '\u06D3' }, { '\u06D5', '\u06D5' },
        { '\u06E5', '\u06E6' }, { '\u0905', '\u0939' }, { '\u093D', '\u093D' }, { '\u0958', '\u0961' },
        { '\u0985', '\u098C' }, { '\u098F', '\u0990' }, { '\u0993', '\u09A8' }, { '\u09AA', '\u09B0' },
        { '\u09B2', '\u09B2' }, { '\u09B6', '\u09B9' }, { '\u09DC', '\u09DD' }, { '\u09DF', '\u09E1' },
        { '\u09F0', '\u09F1' }, { '\u0A05', '\u0A0A' }, { '\u0A0F', '\u0A10' }, { '\u0A13', '\u0A28' },
        { '\u0A2A', '\u0A30' }, { '\u0A32', '\u0A33' }, { '\u0A35', '\u0A36' }, { '\u0A38', '\u0A39' },
        { '\u0A59', '\u0A5C' }, { '\u0A5E', '\u0A5E' }, { '\u0A72', '\u0A74' }, { '\u0A85', '\u0A8B' },
        { '\u0A8D', '\u0A8D' }, { '\u0A8F', '\u0A91' }, { '\u0A93', '\u0AA8' }, { '\u0AAA', '\u0AB0' },
        { '\u0AB2', '\u0AB3' }, { '\u0AB5', '\u0AB9' }, { '\u0ABD', '\u0ABD' }, { '\u0AE0', '\u0AE0' },
        { '\u0B05', '\u0B0C' }, { '\u0B0F', '\u0B10' }, { '\u0B13', '\u0B28' }, { '\u0B2A', '\u0B30' },
        { '\u0B32', '\u0B33' }, { '\u0B36', '\u0B39' }, { '\u0B3D', '\u0B3D' }, { '\u0B5C', '\u0B5D' },
        { '\u0B5F', '\u0B61' }, { '\u0B85', '\u0B8A' }, { '\u0B8E', '\u0B90' }, { '\u0B92', '\u0B95' },
        { '\u0B99', '\u0B9A' }, { '\u0B9C', '\u0B9C' }, { '\u0B9E', '\u0B9F' }, { '\u0BA3', '\u0BA4' },
        { '\u0BA8', '\u0BAA' }, { '\u0BAE', '\u0BB5' }, { '\u0BB7', '\u0BB9' }, { '\u0C05', '\u0C0C' },
        { '\u0C0E', '\u0C10' }, { '\u0C12', '\u0C28' }, { '\u0C2A', '\u0C33' }, { '\u0C35', '\u0C39' },
        { '\u0C60', '\u0C61' }, { '\u0C85', '\u0C8C' }, { '\u0C8E', '\u0C90' }, { '\u0C92', '\u0CA8' },
        { '\u0CAA', '\u0CB3' }, { '\u0CB5', '\u0CB9' }, { '\u0CDE', '\u0CDE' }, { '\u0CE0', '\u0CE1' },
        { '\u0D05', '\u0D0C' }, { '\u0D0E', '\u0D10' }, { '\u0D12', '\u0D28' }, { '\u0D2A', '\u0D39' },
        { '\u0D60', '\u0D61' }, { '\u0E01', '\u0E2E' }, { '\u0E30', '\u0E30' }, { '\u0E32', '\u0E33' },
        { '\u0E40', '\u0E45' }, { '\u0E81', '\u0E82' }, { '\u0E84', '\u0E84' }, { '\u0E87', '\u0E88' },
        { '\u0E8A', '\u0E8A' }, { '\u0E8D', '\u0E8D' }, { '\u0E94', '\u0E97' }, { '\u0E99', '\u0E9F' },
        { '\u0EA1', '\u0EA3' }, { '\u0EA5', '\u0EA5' }, { '\u0EA7', '\u0EA7' }, { '\u0EAA', '\u0EAB' },
        { '\u0EAD', '\u0EAE' }, { '\u0EB0', '\u0EB0' }, { '\u0EB2', '\u0EB3' }, { '\u0EBD', '\u0EBD' },
        { '\u0EC0', '\u0EC4' }, { '\u0F40', '\u0F47' }, { '\u0F49', '\u0F69' }, { '\u10A0', '\u10C5' },
        { '\u10D0', '\u10F6' }, { '\u1100', '\u1100' }, { '\u1102', '\u1103' }, { '\u1105', '\u1107' },
        { '\u1109', '\u1109' }, { '\u110B', '\u110C' }, { '\u110E', '\u1112' }, { '\u113C', '\u113C' },
        { '\u113E', '\u113E' }, { '\u1140', '\u1140' }, { '\u114C', '\u114C' }, { '\u114E', '\u114E' },
        { '\u1150', '\u1150' }, { '\u1154', '\u1155' }, { '\u1159', '\u1159' }, { '\u115F', '\u1161' },
        { '\u1163', '\u1163' }, { '\u1165', '\u1165' }, { '\u1167', '\u1167' }, { '\u1169', '\u1169' },
        { '\u116D', '\u116E' }, { '\u1172', '\u1173' }, { '\u1175', '\u1175' }, { '\u119E', '\u119E' },
        { '\u11A8', '\u11A8' }, { '\u11AB', '\u11AB' }, { '\u11AE', '\u11AF' }, { '\u11B7', '\u11B8' },
        { '\u11BA', '\u11BA' }, { '\u11BC', '\u11C2' }, { '\u11EB', '\u11EB' }, { '\u11F0', '\u11F0' },
        { '\u11F9', '\u11F9' }, { '\u1E00', '\u1E9B' }, { '\u1EA0', '\u1EF9' }, { '\u1F00', '\u1F15' },
        { '\u1F18', '\u1F1D' }, { '\u1F20', '\u1F45' }, { '\u1F48', '\u1F4D' }, { '\u1F50', '\u1F57' },
        { '\u1F59', '\u1F59' }, { '\u1F5B', '\u1F5B' }, { '\u1F5D', '\u1F5D' }, { '\u1F5F', '\u1F7D' },
        { '\u1F80', '\u1FB4' }, { '\u1FB6', '\u1FBC' }, { '\u1FBE', '\u1FBE' }, { '\u1FC2', '\u1FC4' },
        { '\u1FC6', '\u1FCC' }, { '\u1FD0', '\u1FD3' }, { '\u1FD6', '\u1FDB' }, { '\u1FE0', '\u1FEC' },
        { '\u1FF2', '\u1FF4' }, { '\u1FF6', '\u1FFC' }, { '\u2126', '\u2126' }, { '\u212A', '\u212B' },
        { '\u212E', '\u212E' }, { '\u2180', '\u2182' }, { '\u3041', '\u3094' }, { '\u30A1', '\u30FA' },
        { '\u3105', '\u312C' }, { '\uAC00', '\uD7A3' }
    };

    //
    // Ideographic Character range
    //

    public static final char[][] ideographicRange =
    {
        { '\u3007', '\u3007' },	{ '\u3021', '\u3029' },	{ '\u4E00', '\u9FA5' }
    };
    
    //
    // Combining Character range
    //
    public final static char[][] combiningRange =
    {
        { '\u0300', '\u0345' }, { '\u0360', '\u0361' }, { '\u0483', '\u0486' }, { '\u0591', '\u05A1' }, 
        { '\u05A3', '\u05B9' }, { '\u05BB', '\u05BD' }, { '\u05BF', '\u05BF' }, { '\u05C1', '\u05C2' }, 
        { '\u05C4', '\u05C4' }, { '\u064B', '\u0652' }, { '\u0670', '\u0670' }, { '\u06D6', '\u06DC' }, 
        { '\u06DD', '\u06DF' }, { '\u06E0', '\u06E4' }, { '\u06E7', '\u06E8' }, { '\u06EA', '\u06ED' }, 
        { '\u0901', '\u0903' }, { '\u093C', '\u093C' }, { '\u093E', '\u094C' }, { '\u094D', '\u094D' }, 
        { '\u0951', '\u0954' }, { '\u0962', '\u0963' }, { '\u0981', '\u0983' }, { '\u09BC', '\u09BC' }, 
        { '\u09BE', '\u09BE' }, { '\u09BF', '\u09BF' }, { '\u09C0', '\u09C4' }, { '\u09C7', '\u09C8' }, 
        { '\u09CB', '\u09CD' }, { '\u09D7', '\u09D7' }, { '\u09E2', '\u09E3' }, { '\u0A02', '\u0A02' }, 
        { '\u0A3C', '\u0A3C' }, { '\u0A3E', '\u0A3E' }, { '\u0A3F', '\u0A3F' }, { '\u0A40', '\u0A42' }, 
        { '\u0A47', '\u0A48' }, { '\u0A4B', '\u0A4D' }, { '\u0A70', '\u0A71' }, { '\u0A81', '\u0A83' }, 
        { '\u0ABC', '\u0ABC' }, { '\u0ABE', '\u0AC5' }, { '\u0AC7', '\u0AC9' }, { '\u0ACB', '\u0ACD' }, 
        { '\u0B01', '\u0B03' }, { '\u0B3C', '\u0B3C' }, { '\u0B3E', '\u0B43' }, { '\u0B47', '\u0B48' }, 
        { '\u0B4B', '\u0B4D' }, { '\u0B56', '\u0B57' }, { '\u0B82', '\u0B83' }, { '\u0BBE', '\u0BC2' }, 
        { '\u0BC6', '\u0BC8' }, { '\u0BCA', '\u0BCD' }, { '\u0BD7', '\u0BD7' }, { '\u0C01', '\u0C03' }, 
        { '\u0C3E', '\u0C44' }, { '\u0C46', '\u0C48' }, { '\u0C4A', '\u0C4D' }, { '\u0C55', '\u0C56' }, 
        { '\u0C82', '\u0C83' }, { '\u0CBE', '\u0CC4' }, { '\u0CC6', '\u0CC8' }, { '\u0CCA', '\u0CCD' }, 
        { '\u0CD5', '\u0CD6' }, { '\u0D02', '\u0D03' }, { '\u0D3E', '\u0D43' }, { '\u0D46', '\u0D48' }, 
        { '\u0D4A', '\u0D4D' }, { '\u0D57', '\u0D57' }, { '\u0E31', '\u0E31' }, { '\u0E34', '\u0E3A' }, 
        { '\u0E47', '\u0E4E' }, { '\u0EB1', '\u0EB1' }, { '\u0EB4', '\u0EB9' }, { '\u0EBB', '\u0EBC' }, 
        { '\u0EC8', '\u0ECD' }, { '\u0F18', '\u0F19' }, { '\u0F35', '\u0F35' }, { '\u0F37', '\u0F37' }, 
        { '\u0F39', '\u0F39' }, { '\u0F3E', '\u0F3E' }, { '\u0F3F', '\u0F3F' }, { '\u0F71', '\u0F84' }, 
        { '\u0F86', '\u0F8B' }, { '\u0F90', '\u0F95' }, { '\u0F97', '\u0F97' }, { '\u0F99', '\u0FAD' }, 
        { '\u0FB1', '\u0FB7' }, { '\u0FB9', '\u0FB9' }, { '\u20D0', '\u20DC' }, { '\u20E1', '\u20E1' }, 
        { '\u302A', '\u302F' }, { '\u3099', '\u3099' }, { '\u309A', '\u309A' } 
    };
    
    //
    //	Extended Digit range
    //
    private static final char[][] extendedDigitRange =
    {
    	{ '\u0660', '\u0669' },	{ '\u06F0', '\u06F9' },	{ '\u0966', '\u096F' },	{ '\u09E6', '\u09EF' },
        { '\u0A66', '\u0A6F' },	{ '\u0AE6', '\u0AEF' },	{ '\u0B66', '\u0B6F' },	{ '\u0BE7', '\u0BEF' },
        { '\u0C66', '\u0C6F' },	{ '\u0CE6', '\u0CEF' },	{ '\u0D66', '\u0D6F' },	{ '\u0E50', '\u0E59' },
        { '\u0ED0', '\u0ED9' },	{ '\u0F20', '\u0F29' }
    };

    //
    // Character Extender range
    //
    private static final char [][] extenderRange =
    {
        { '\u00B7', '\u00B7' }, { '\u02D0', '\u02D1' }, { '\u0387', '\u0387' }, { '\u0640', '\u0640' },
        { '\u0E46', '\u0E46' }, { '\u0EC6', '\u0EC6' }, { '\u3005', '\u3005' }, { '\u3031', '\u3035' },
        { '\u309D', '\u309E' }, { '\u30FC', '\u30FE' }
    };
  
    //
    // Hex character Range
    //
    private static final char [][] hexCharRange  = 
    {
    	{ 'A', 'F' },		
    	{ 'a', 'f' }
    };
    
    	
    private static final String[][] builtInEntities =
    {
        { "amp",	"&"  },
        { "apos",   "'"  },
        { "gt", 	">"  },
        { "lt", 	"<"  },
        { "quot",   "\"" }
    };
	
    private static byte [ /* 65536 */ ] characterType = new byte[65536];
  	
    private static boolean characterTypeInitialized = false;

    /*----------------------------------------------------------------------------------+
     |                                                                                  |
     |                XML Scanner Finite State Machine STATE DEFINITIONS                |
     |                                                                                  |
     +----------------------------------------------------------------------------------*/

     /* don't change these state definitions without repairing table */
	
    static final byte S0 = 0, S_WHITETOKEN  = S0; /* white space token state */
    static final byte S1 = 1, S_NAME        = S1; /* XML Name state */
    static final byte S2 = 2, S_NAMETOKEN   = S2; /* XML NMTOKEN state */
    static final byte S3 = 3, S_COMMENT     = S3; /* "<!--" seen and therefore in XML comment state */
    static final byte S4 = 4, S_MARKUPBGN   = S4; /* "<" seen state */
    static final byte S5 = 5, S_LESSBANG    = S5; /* "<!" seen  */
    static final byte S6 = 6, S_DIGIT       = S6; /* integer seen */
    static final byte S7 = 7, S_QUOTE       = S7; /* quoted string -- apostrophe or quotation seen */
    static final byte S8 = 8, S_SGEREF      = S8; /* "&" seen, Start of a general entity reference */
    static final byte S9 = 9, S_REFCHAR     = S9; /* "&#" seen, Reference character state */
    static final byte SA = 10,S_SPEREF      = SA; /* % seen, start PEreference state */
    static final byte SB = 11,S_HEXCHAR     = SB; /* Reference Char hex character state */
    static final byte SC = 12,S_PENAME      = SC; /* PEreference name state */
    static final byte SD = 13,S_REFNAME     = SD; /* Reference Char name state */
    static final byte SE = 14,S_LESSQUEST   = SE; /* "<?" seen */
    static final byte SF = 15,S_LBDASH1     = SF; /* "<!-" seen */
    static final byte SG = 16,S_ECDASH1     = SG; /* "<!--" char* "-" seen */
    static final byte SH = 17,S_ECDASH2     = SH; /* "<!--" char* "--" seen */
    static final byte SI = 18,S_LBBOX       = SI; /* <![ seen */
    static final byte SJ = 19,S_SPECIAL     = SJ; /* Special character seen */
    static final byte SK = 20,S_LBBOXNAME   = SK; /* <![name (building name) state */
    static final byte SL = 21,S_LBNAME      = SL; /* <!name state - e.g. (<!DOCTYPE, ...) */
    static final byte SM = 22,S_LQNAME      = SM; /* <?name state (<?XML, or <?pitarget */
    static final byte SN = 23,S_INCLUDE     = SN; /* "<![ name [" seen, in CDATA, INCLUDE or IGNORE */
    static final byte SO = 24,S_INCLUDEBOX  = SO; /* in Section, "]" seen */
    static final byte SP = 25,S_INCLUDEBB   = SP; /* in Section, "]]" seen */
    static final byte SQ = 26,S_MARKUPEND   = SQ; /* saw a markup end ">" character */
    static final byte SR = 27,S_ELEMENTEND  = SR; /* saw "</" signalling end tag of element markup */

    static final byte SS = 28,S_SLASH       = SS; /* saw "/" signalling possible empty element ("/>") */
    static final byte ST = 29,S_WHITESPACE  = ST; /* ignoring whitespace state */
    static final byte SU = 30,S_LBBBLANK    = SU; /* "<![" S? (whitespace after less bang box) state */
    static final byte SV = 31,S_LBBNAMEBL   = SV; /* "<![" S? name S? (whitespace after lbbname state */

    static final byte SW = 32,S_TAGNAME     = SW; /* <name state - e.g.( <tagname, ...) */
    

    /* section state machine for IGNORE and <unknown> sections */

	static final byte SX = 33, SC_START     = SX; 
    static final byte SY = (byte)(0 + SC_START), S_IGNORE    = SY; /* IGNORE common collection state */
    static final byte SZ = (byte)(1 + SC_START), S_IGNLESS   = SZ; /* "<" seen in IGNORE section */
    static final byte Sa = (byte)(2 + SC_START), S_IGNLESSB  = Sa; /* "<!" seen  IGNORE section */
    static final byte Sb = (byte)(3 + SC_START), S_IGNLBOX   = Sb; /* "<![" seen IGNORE section */
    static final byte Sc = (byte)(4 + SC_START), S_IGNRBOX   = Sc; /* "]" seen IGNORE section */
    static final byte Sd = (byte)(5 + SC_START), S_IGNRBXRBX = Sd; /* "]]" seen IGNORE section */
    static final byte Se = (byte)(6 + SC_START), S_IGNGREAT  = Se; /* "]]>" seen IGNORE section */
    
    static final byte Sf = (byte)(7 + SC_START), S_PIWHITE   = Sf; /* "<?" name with possible white space seen */
    static final byte Sg = (byte)(8 + SC_START), S_PIINST    = Sg; /* "<?" name whitespace non-white (char-'?')* seen */
    static final byte Sh = (byte)(9 + SC_START), S_LONEQUEST = Sh; /* "?" seen in "<?" name whitespace non-white (char-'?')* "?" string */
    static final byte Sj = (byte)(10+ SC_START), S_LQWHITE   = Sj; /* "<?" whitespace seen */
    static final byte Sk = (byte)(11+ SC_START), S_QUESTION  = Sk; /* "?" seen -- look for ">" */

    static final byte S_STATES = (byte)( Sk + 1 ); /* non-ending scan states are less than this */

    /* Ending states for token scanner */

    static final byte T0 = (byte)(S_STATES),    IS_START      = T0; /* token ending states */
	
    static final byte T1 = (byte)(IS_START +0), IS_CHARREFHEX = T1; /* character reference in HEX */
    static final byte T2 = (byte)(IS_START +1), IS_CHARREFDIG = T2; /* character reference in digits */
    static final byte T3 = (byte)(IS_START +2), IS_GEREF      = T3; /* general entity reference name */
    static final byte T4 = (byte)(IS_START +3), IS_PEREF	  = T4; /* PEreference */
    static final byte T5 = (byte)(IS_START +4), IS_QUOTED	  = T5; /* quoted string */
    static final byte T6 = (byte)(IS_START +5), IS_COMMENT    = T6; /* comment */
    static final byte T7 = (byte)(IS_START +6), IS_NAME 	  = T7; /* Name (or NameToken ) */
    static final byte T8 = (byte)(IS_START +7), IS_NMTOKEN    = T8; /* NameToken -- but NOT a name! */
    static final byte T9 = (byte)(IS_START +8), IS_LBNAME	  = T9; /* "name" after a BangMarkup token */
    static final byte TA = (byte)(IS_START +9), IS_LBBOXNAME  = TA; /* "<![name" (where name probably is 
                                                       CDATA, INCLUDE or IGNORE) */
    static final byte TB = (byte)(IS_START +10),IS_LQNAME	  = TB; /* token is "<?name"  where name is 
                                                                probably xml or a pitarget */
    static final byte TC = (byte)(IS_START +11),IS_SPECIAL    = TC; /* special character seen */
    static final byte TD = (byte)(IS_START +12),IS_MARKUP	  = TD; /* token is "<" */
    static final byte TE = (byte)(IS_START +13),IS_BANGMARKUP = TE; /* token is "<!" */
    static final byte TF = (byte)(IS_START +14),IS_QUESMARKUP = TF; /* token is "<?" */
    static final byte TG = (byte)(IS_START +15),IS_EMPTYELEM  = TG; /* "/>" signals empty markup element*/
    static final byte TH = (byte)(IS_START +16),IS_BBMARKUP   = TH; /* "<![" ( bang box markup) almost section */
    static final byte TI = (byte)(IS_START +17),IS_LBDMARKUP  = TI; /* "<!-" almost a comment */
    static final byte TJ = (byte)(IS_START +18),IS_ENDTAG	  = TJ; /* "</" signals end tag of element */
    static final byte TK = (byte)(IS_START +19),IS_MARKUPEND  = TK; /* ">" and ends markup */
    static final byte TL = (byte)(IS_START +20),IS_SECTION    = TL; /* "<![name[ char* ]]>", a section */
    static final byte TM = (byte)(IS_START +21),IS_WHITESPACE = TM; /* token is whitespace */
    static final byte TN = (byte)(IS_START +22),IS_PERCENT    = TN; /* token is "%" */
    static final byte TO = (byte)(IS_START +23),IS_RBOX       = TO; /* token is "]" */
    static final byte TP = (byte)(IS_START +24),IS_RBOXRBOX   = TP; /* token is "]]" */
    static final byte TQ = (byte)(IS_START +25),IS_TAGNAME    = TQ; /* token is tagname */
    static final byte TR = (byte)(IS_START +26),IS_PI         = TR; /* token is processing instruction */
    static final byte TS = (byte)(IS_START +27),IS_QUESTION   = TS; /* token is "?" */
    static final byte TT = (byte)(IS_START +28),IS_QMARKEND   = TT; /* token is "?>",  XML decl or PI end */

	static final byte T_STATES = (byte)( TT + 1 );  /* Good Token ending states less than this */
    
    /* Errors ending token scan */
	
    static final byte U0 = (byte)(T_STATES),     IN_START      = U0; /*token terminated(eos, eof, syntax, ... ) */
	
    static final byte U1 = (byte)(IN_START + 0), IN_CHARREF    = U1; /*in char ref and EOL or EOS seen */
    static final byte U2 = (byte)(IN_START + 1), IN_QUOTED	   = U2; /*in quoted string, no final quote */
    static final byte U3 = (byte)(IN_START + 2), IN_COMMENT    = U3; /*in comment -- no final "-->" */
    static final byte U4 = (byte)(IN_START + 3), IN_SECTION    = U4; /*in section -- no final "]]>" */
    static final byte U5 = (byte)(IN_START + 4), IN_MARKUP	   = U5; /*in markup("<" and no ">")*/
    static final byte U6 = (byte)(IN_START + 5), IN_MARKUPBANG = U6; /*in markup bang("<! and no  ">") */
    static final byte U7 = (byte)(IN_START + 6), IN_MARKUPQUES = U7; /*in markup question("<?" & no "?>" */
    static final byte U8 = (byte)(IN_START + 7), IN_ENTITY	   = U8; /*in parm or general entity */
    static final byte U9 = (byte)(IN_START + 8), IN_NAMETOKEN  = U9; /*in name token */
    static final byte UA = (byte)(IN_START + 9), IN_NAME	   = UA; /*in name */
    static final byte UB = (byte)(IN_START + 10),IN_GEREF	   = UB; /*in ge ref (i.e. & seen) */
    static final byte UC = (byte)(IN_START + 11),IN_PEREF	   = UC; /*in pe ref (i.e. % seen) */
    static final byte UD = (byte)(IN_START + 12),IN_LBDASH1    = UD; /* "<!-" and EOL or EOS */
    static final byte UE = (byte)(IN_START + 13),IN_LBBOX	   = UE; /* "<![" and EOL or EOS */
    static final byte UF = (byte)(IN_START + 14),IN_LBBOXNAME  = UF; /* "<![nameprefix" and EOL or EOS */
    static final byte UG = (byte)(IN_START + 15),IN_LBNAME	   = UG; /* "<!nameprefix" and EOL or EOS */
    static final byte UH = (byte)(IN_START + 16),IN_LQNAME	   = UH; /* "<?nameprefix" and EOL or EOS */
    static final byte UI = (byte)(IN_START + 17),IN_SLASH	   = UI; /* "/" and EOL or EOS */
    static final byte UJ = (byte)(IN_START + 18),IN_IGNORE     = UJ; /*in IGNORE & EOL or EOS seen */
    static final byte Uk = (byte)(IN_START + 19),IN_PI         = Uk; /*in PI and EOL or EOS seen */
	
    /* Natural termination states */
	
    static final byte E0 = (byte)(Uk + 1),        END_START	    = E0;
    static final byte E1 = (byte)(END_START + 0), END_EMPTY	    = E1; /* Ended in whitespace */
    static final byte E2 = (byte)(END_START + 1), END_ILLEGAL   = E2; /* Ended in illegal char 
           (eg. #x20DD-#x20E0 see Appendix B, "Character Classes" in the XML spec.) */

    /* End of Scan */
    static final byte E3 = (byte)(END_START + 2), IS_EOS	    = E3; /*string/stream end, tokens rtn'd */
	
    static final byte [/*S_STATES*/][/*C_TYPE*/] nextState = 
    {
/*                     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                       | I| W| S| Q| D| N| L| P| A| S| H| S  M| M| B| Q| D| L| R| X| S| S| P| E| B| L| R| E|
                       | L| H| P| U| I| A| E| E| M| H| E| E| A| A| A| U| A| E| I|  | L| T| L| Q| A| E| I| N|
                       | L| I| E| O| G| M| T| R| P| A| X| M| R| R| N| E| S| F| G|  | A| A| U| U| R| F| G| D|
                       | E| T| C| T| I| E| T| C| E| R| C| I| K| K| G| S| H| T| H|  | S| R| S| A|  | T| H|  |
                       | G| E| I| E| T| C| E| E| R| P| H| C| U| U|  | T|  | B| T|  | H|  |  | L|  | P| T|  |
                       | A| S| A|  |  | H| R| N| S|  | A| O| P| P|  | I|  | O| B|  |  |  |  |  |  | A| P|  |
                       | L| P| L|  |  | A|  | T| A|  | R| L| S| E|  | O|  | X| O|  |  |  |  |  |  | R| A|  |
                       |  | A|  |  |  | R|  |  | N|  |  | O| T| N|  | N|  |  | X|  |  |  |  |  |  | E| R|  |
                       |  | C|  |  |  |  |  |  | D|  |  | N| A| D|  |  |  |  |  |  |  |  |  |  |  | N| E|  |
                       |  | E|  |  |  |  |  |  |  |  |  |  | R|  |  |  |  |  |  |  |  |  |  |  |  |  | N|  |
                       |  |  |  |  |  |  |  |  |  |  |  |  | T|  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
                       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                       | 0| 1| 2| 3| 4| 5| 6| 7| 8| 9|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|
/*                     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
/* S_WHITETOKEN = S0*/ {TM,S0,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,TM,E3},
/* S_NAME	    = S1*/ {T7,T7,T7,T7,S1,S1,S1,T7,T7,T7,S1,T6,T7,T7,T7,T7,S1,T7,T7,S1,T7,T7,T7,T7,T7,T7,T7,UA},
/* S_NAMETOKEN  = S2*/ {T8,T8,T8,T8,S2,S2,S2,T8,T8,T8,S2,T8,T8,T8,T8,T8,S2,T8,T8,S2,T8,T8,T8,T8,T8,T8,T8,U9},
/* S_COMMENT	= S3*/ {S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,SG,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,U3},
/* S_MARKUPBGN  = S4*/ {TD,TD,TD,TD,TD,TD,SW,SA,S8,TD,TD,TD,TD,TD,S5,SE,TD,TD,TD,TD,SR,TD,TD,TD,TD,TD,TD,U5},
/* S_LESSBANG   = S5*/ {TE,TE,TE,TE,TE,TE,SL,TE,TE,TE,SL,TE,TE,TE,TE,TE,SF,SI,TE,SL,TE,TE,TE,TE,TE,TE,TE,U6},
/* S_DIGIT	    = S6*/ {U1,U1,U1,U1,S6,U1,U1,U1,U1,U1,U1,T2,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1},
/* S_QUOTE	    = S7*/ {S7,S7,S7,T5,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,S7,U2},
/* S_SGEREF     = S8*/ {TC,TC,TC,TC,TC,TC,SD,TC,TC,S9,SD,TC,TC,TC,TC,TC,TC,TC,TC,SD,TC,TC,TC,TC,TC,TC,TC,UB},
/* S_REFCHAR	= S9*/ {U1,U1,U1,U1,S6,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,SB,U1,U1,U1,U1,U1,U1,U1,U1},
/* S_SPEREF	    = SA*/ {TN,TN,TN,TN,TN,TN,SC,TN,TN,TN,SC,TN,TN,TN,TN,TN,TN,TN,TN,TN,TN,TN,TN,TN,TN,TN,TN,UC},
/* S_HEXCHAR	= SB*/ {U1,U1,U1,U1,SB,U1,U1,U1,U1,U1,SB,T1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1,U1}, 
/* S_PENAME     = SC*/ {U8,U8,U8,U8,SC,SC,SC,U8,U8,U8,SC,T4,U8,U8,U8,U8,SC,U8,U8,SC,U8,U8,U8,U8,U8,U8,U8,U8},
/* S_REFNAME    = SD*/ {U8,U8,U8,U8,SD,SD,SD,U8,U8,U8,SD,T3,U8,U8,U8,U8,SD,U8,U8,SD,U8,U8,U8,U8,U8,U8,U8,U8},
/* S_LESSQUEST  = SE*/ {TF,Sj,TF,TF,TF,TF,SM,TF,TF,TF,SM,TF,TF,TF,TF,TF,TF,TF,TF,SM,TF,TF,TF,TF,TF,TF,TF,U7},
/* S_LBDASH1    = SF*/ {TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,S3,TI,TI,TI,TI,TI,TI,TI,TI,TI,TI,UD},
/* S_ECDASH1	= SG*/ {S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,SH,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,U3},
/* S_ECDASH2	= SH*/ {S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,S0,S3,S3,SH,S3,S3,S3,S3,S3,S3,S3,S3,S3,S3,U3}, 
/* S_LBBOX	    = SI*/ {TH,SU,TH,TH,TH,TH,SK,SA,TH,TH,SK,TH,TH,TH,TH,TH,TH,TH,TH,TH,SK,TH,TH,TH,TH,TH,TH,UE},
/* S_SPECIAL	= SJ*/ {TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC},
/* S_LBBOXNAME  = SK*/ {TA,SV,TA,TA,SK,SK,SK,TA,TA,TA,SK,TA,TA,TA,TA,TA,SK,TA,TA,SK,TA,TA,TA,TA,TA,TA,TA,UF},
/* S_LBNAME	    = SL*/ {T9,T9,T9,T9,SL,SL,SL,T9,T9,T9,SL,T9,T9,T9,T9,T9,SL,T9,T9,SL,T9,T9,T9,T9,T9,T9,T9,T9},
/* S_LQNAME	    = SM*/ {TB,TB,TB,TB,SM,SM,SM,TB,TB,TB,SM,TB,TB,TB,TB,TB,SM,TB,TB,SM,TB,TB,TB,TB,TB,TB,TB,UH},
/* S_INCLUDE    = SN*/ {SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SN,SO,SN,SN,SN,SN,SN,SN,SN,SN,U4},
/* S_INCLUDEBOX = SO*/ {TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,TO,SP,TO,TO,TO,TO,TO,TO,TO,TO,TO},
/* S_INCLUDEBB  = SP*/ {TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TL,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP,TP},
/* S_MARKUPEND  = SQ*/ {TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK,TK},
/* S_ELEMENTEND = SR*/ {TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ,TJ},
/* S_SLASH	    = SS*/ {TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,SR,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC,TC},
/* S_WHITESPACE = ST*/ {SJ,ST,SJ,S7,S2,S2,S1,SA,S8,SJ,S1,SJ,S4,SQ,SJ,Sk,S2,SJ,SO,S1,SS,SJ,SJ,SJ,SJ,SJ,SJ,E3},
/* S_LBBBLANK   = SU*/ {TH,SU,TH,TH,TH,TH,SK,SK,TH,TH,SK,TH,TH,TH,TH,TH,TH,TH,TH,SK,TH,TH,TH,TH,TH,TH,TH,UG},
/* S_LBBNAMEBL  = SV*/ {TA,SV,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,TA,UF},
/* S_TAGNAME    = SW*/ {TQ,TQ,TQ,TQ,SW,SW,SW,TQ,TQ,TQ,SW,TQ,TQ,TQ,TQ,TQ,SW,TQ,TQ,SW,TQ,TQ,TQ,TQ,TQ,TQ,TQ,TQ}, 
/*                     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                       | 0| 1| 2| 3| 4| 5| 6| 7| 8| 9|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|
                       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                       | I| W| S| Q| D| N| L| P| A| S| H| S| M| M| B| Q| D| L| R| X| S| S| P| E| B| L| R| E|
                       | L| H| P| U| I| A| E| E| M| H| E| E| A| A| A| U| A| E| I|  | L| T| L| Q| A| E| I| N|
                       | L| I| E| O| G| M| T| R| P| A| X| M| R| R| N| E| S| F| G|  | A| A| U| U| R| F| G| D|
                       | E| T| C| T| I| E| T| C| E| R| C| I| K| K| G| S| H| T| H|  | S| R| S| A|  | T| H|  |
                       | G| E| I| E| T| C| E| E| R| P| H| C| U| U|  | T|  | B| T|  | H|  |  | L|  | P| T|  |
                       | A| S| A|  |  | H| R| N| S|  | A| O| P| P|  | I|  | O| B|  |  |  |  |  |  | A| P|  |
                       | L| P| L|  |  | A|  | T| A|  | R| L| S| E|  | O|  | X| O|  |  |  |  |  |  | R| A|  |
                       |  | A|  |  |  | R|  |  | N|  |  | O| T| N|  | N|  |  | X|  |  |  |  |  |  | E| R|  |
                       |  | C|  |  |  |  |  |  | D|  |  | N| A| D|  |  |  |  |  |  |  |  |  |  |  | N| E|  |
                       |  | E|  |  |  |  |  |  |  |  |  |  | R|  |  |  |  |  |  |  |  |  |  |  |  |  | N|  |
                       |  |  |  |  |  |  |  |  |  |  |  |  | T|  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
                       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                       | 0| 1| 2| 3| 4| 5| 6| 7| 8| 9|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|
                       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+*/
/* S_IGNORE     = SY*/ {SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SZ,SY,SY,SY,SY,SY,Sc,SY,SY,SY,SY,SY,SY,SY,SY,UJ},
/* S_IGNLESS    = SZ*/ {SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SZ,SY,Sa,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,UJ},
/* S_IGNLESSB   = Sa*/ {SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SZ,SY,SY,SY,SY,Sb,SY,SY,SY,SY,SY,SY,SY,SY,SY,UJ},
/* S_IGNLBOX    = Sb*/ {SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SZ,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,UJ},
/* S_IGNRBOX    = Sc*/ {SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SZ,SY,SY,SY,SY,SY,Sd,SY,SY,SY,SY,SY,SY,SY,SY,UJ},
/* S_IGNRBXRBX  = Sd*/ {SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SZ,Se,SY,SY,SY,SY,Sd,SY,SY,SY,SY,SY,SY,SY,SY,UJ},
/* S_IGNGREAT   = Se*/ {SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SZ,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,SY,UJ},
/* Note that the behavior of a general entity in a PI is unclear -- we ignore them -- check the 
   eratta of the XML spec to see what it says and clean up this comment when you figure it out! $FIX$ */
/* S_PIWHITE    = Sf*/ {Sg,Sf,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sh,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Uk},
/* S_PIINST     = Sg*/ {Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sh,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Uk},
/* S_LONEQUEST  = Sh*/ {Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,TR,Sg,Sh,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Sg,Uk},
/* S_LQWHITE    = Sj*/ {TF,Sj,TF,TF,TF,SM,SM,TF,TF,TF,SM,TF,TF,TF,TF,TM,SM,TF,TF,SM,TF,TF,TF,TF,TF,TF,TF,Uk},
/* S_QUESTION   = Sk*/ {TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TT,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS,TS}
/*                     +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
                       | 0| 1| 2| 3| 4| 5| 6| 7| 8| 9|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|
                       +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+*/
          };

    //
    // This table implements a state machine to scan sections
    // It uses the section state table entries and constants defined above.
    //
    // It follows the (arcane) XML spec to implement CDATA, IGNORE and INCLUDE
    // sections.  This means that it keeps all non-bracket content in an 
    // INCLUDE section and discards all IGNORE content nested inside an
    // INCLUDE section:
    // 
    // INCLDUE------------------------------+-----------+-----------+----------+-------------+
    // |                                                                                     |
    // |INCLUDE-----+----------+-----------+                                                 |
    // ||                                  |                                                 |
    // ||           IGNORE-----+           |                        IGNORE------------------+|
    // ||           |          |           |                        |                       ||
    // ||           |          |           +IGNORE------+           |          INCLUDE-----+||
    // ||           |          |           ||           |           |          |           |||
    // ++-----------+----------+-----------++-----------+-----------+----------+-----------+++
    // ||  content1  | content2 |  content3|| content4  | content5  | content6 | content7  |||
    // ++===========+^^^^^^^^^^+===========++^^^^^^^^^^^+===========+^^^^^^^^^^+^^^^^^^^^^^+++
    //
    // In the above example the following content sequences are in the string result upon
    // return:  1, 3, and 5; while 2, 4, 6  and 7 have been discarded.  Note that 7 has
    // been discarded even though it is in an INCLUDE because it is nested inside an IGNORE.
    // INGORE's discard all contained content regardless of nested INCLUDEs.
    //
    // This method also implements an arbitrary section -- with the same
    // semantics as IGNORE (i.e. nesting is supported ) -- except that all the content
    // in the interior of the section up to the last closing section bracket, "]]>" is kept
    // under the assumption that a user of the scanner may want to paw through
    // the result.  So the section, "<![exml[silly <![ some more stuff ]]> and even more ]]>"
    // returns the result, "silly <![ some more stuff ]]> and even more ".  Both the
    // front and back brackets have been stripped.
    //
    // CDATA sections are implemented according to the spec in that nesting of sections
    // inside a CDATA section is not supported since the "<" charcter is just another
    // character.  (I told you that XML's rules are arcane...)  This means that the first
    // "]]>" (closing section bracket) terminates the CDATA.  So, the CDATA section,
    // "<![CDATA[silly <![ some more stuff ]]> and even more ]]>" returns the 
    // result, "silly <![ some more stuff " and the scan pointer is left pointing 
    // at the blank just past the first "]]>", leaving " and even more ]]>" to be 
    // scanned.
    //
    // If an IGNORE section is initially scanned, the entire nested content is returned
    // in the result string so the caller can decide what the result is to be.
    // So, the IGNORE section, "<![IGNORE[ content6 <![ INCLUDE [ content7 ]]> stuff ]]>"
    // returns " content6 <![ INCLUDE [ content7 ]]> stuff ".
    //
    // Finally, if a CDATA section is nested inside an INCLUDE, it, along with it's markup
    // is preserved in the returned string.  The assumption here is that the writer of the 
    // DTD has it under control.
    //
    /*----------------------------------------------------------------------------------+
     |                                                                                  |
     |                         XML Scanner Private Variables                            |
     |                                                                                  |
     +----------------------------------------------------------------------------------*/
    private static SectionDefinition[] sourceOptions = new SectionDefinition[4];
    //
    //	Scanner options
    //
    //	1.  Handle line ends as required by XML (normalized to linefeeds) or
    //		handle line ends as returns ( normalized to carraige returns ) or
    //		ignore line ends completely ( normalize them to the empty string ) or
    //		include line ends ( leave the line end characters as they appear on the input.) or
    //		handle line ends as returns followed by a linefeed ( normalize them the { cr lf } ).
    //	2.  Return general entity references or replace them with their values.
    //		Default is to replace the ge reference with it's value and not rescan 
    //       (as XML requires).
    //	3.  Return parameter entity references (minus the enclosing "%" and ";"); 
    //      or expand and rescan them.
    //		Default is to replace the pe reference with it's value and rescan the value.
    //	4.  Return character entity references as tokens (minus the enclosing "&" and ";");
    //      or expand them.
    //		Default is to expand and replace a character reference by it's value.
    //	5.  provide 4 options (3 functional) for the scanner to obtain lines to scan;  
    //	    a) none -- no option chosen, scanner cannot scan (This is default initial state.)
    //	    b) reader -- caller supplies a buffered reader stream to scan.
    //	    c) provider -- caller provides an object that implements StringProvider interface 
    //         (i.e. getString.).
    //	    d) string -- caller provides a string to tokenize. 
    //
    String  option_lineEnd = "lf"; 	
    // The options for lineEnd are "lf", "ignore", "include", "cr" and "crlf" */
    // A line may be terminated with '\r' or '\n' or both ( '\r' '\n' ) in sequence.  
    //   The options for lineEnd mean the following:
    //		lf -- each line end is normalized to '\n' (a linefeed) -- This is the XML standard
    //		ignore -- each line end terminates the string but is removed completely.  This means
    //			that a token can be split with line-ends and still be parsed as a token.
    //		include -- the line end characters are included literally as they occur.
    // 		cr -- the line end characters are normalized to '\r'.
    //		crlf -- the line end characters are normalized to '\r' '\n'.
    private boolean option_evaluateGeneralEntities   = true; /* otherwise reference is returned */
    private boolean option_evaluateParameterEntities = true; /* otherwise reference is returned */
    private boolean option_evaluateCharacterEntities = true; /* otherwise they are returned as tokens */
    private String  option_scanSource = "none";		         /* scan nothing */ 
    private boolean option_whiteSpaceIsToken = false;	     /* keep whitespace as a token */
    private boolean option_ignoreTextIsToken = false;        /* keep ignored text as a token */
    private boolean option_sectionDelimiterIsToken = false;  /* Section delimiters are tokens */
    private boolean option_keepInnerSectionMarkup = false;   /* keep interior section markup? */
    
    // setup base scan source
    private ScanSource	source = new ScanSource( null, "", option_lineEnd );
    //
    private DocType dtd = null; /* optional DocType */
    
    private byte lastState = S_WHITESPACE;		/* current scan state */
    private byte previousLastState = S_WHITESPACE;	    	/* last scan state */
	
    private StringBuffer tokenBuf = new StringBuffer( Constants.EXPECTED_TOKEN_LENGTH );
    private StringBuffer comment = new StringBuffer( Constants.EXPECTED_QUOTED_STRING_LENGTH );
    private StringBuffer quotedString = new StringBuffer( Constants.EXPECTED_QUOTED_STRING_LENGTH );	
    private StringBuffer charEntity = new StringBuffer( Constants.EXPECTED_CHARACTER_ENTITY_LENGTH );
    private StringBuffer geReference = new StringBuffer( Constants.EXPECTED_ENTITY_LENGTH );
    private StringBuffer peReference = new StringBuffer( Constants.EXPECTED_ENTITY_LENGTH );
    /* the following ought to be used for comments, and quoted strings  */
    private StringBuffer longBuf = new StringBuffer( Constants.EXPECTED_CDATA_LENGTH );
	
    private ScanToken returnToken = null;

    /* the following are used to note the initialized state of the tokens being assembled */
    private boolean  tokenStart           = false;	/* token not started */
    private boolean  commentStart         = false;	/* comment not started */
    private boolean  quotedStart          = false;	/* quoted string not started */
    private boolean  charEntityStart      = false;	/* character Entity not started */
    private boolean  generalEntityStart   = false;	/* general Entity not started */
    private boolean  parameterEntityStart = false;	/* parameter Entity not started */
    /* store the quote character to simplify the state table; 
       run &apos; and &quot strings in one state */
    private char     quoteChar = '\u0000'; /* either an apostrophe or a quotation mark */
    private char     charValue = '\u0000'; /* The emerging character reference value */
    private boolean  charEntityIsHex = false;
                     /* character entity being scanned is in hex; otherwise decimal */
    private char     entityChar = '\u0000'; 
                     /* contains either & or % according to entity seen */
    /* store the section name to simplify the state table;  
       run all sections out of the state machine  */
    private String   savedSectionName = "";	        /* No section name yet... */
    private String   lastSavedSectionName = "";     /* No last Saved Section Name */
    private byte     savedSection = SectionDefinition.UNKNOWN;
    private byte     lastSavedSection = SectionDefinition.UNKNOWN;
    private int      ignoreNest = 0;   /* records the IGNORE section nesting level */
    private int      includeNest = 0;  /* records the INCLUDE section nesting level */
    //
    // Initializer Method
    //
    private void initializeTypeRange( char [][] span, byte type )
    {
    	for ( char range = 0 ; range < span.length; range++ )
    	{
            for ( char i = span[range][0]; i <= span[range][1]; i++ )
    	    {
                characterType[i] = type;
            }
    	}
    }
    
    /*----------------------------------------------------------------------------------+
     |                                                                                  |
     |                              XML Scanner Constructors			                |
     |                                                                                  |
     +----------------------------------------------------------------------------------*/
    public XMLscanner()
    {
		if ( ! characterTypeInitialized )
		{
	   	//
	    	// The following initialize the unicode character type byte array
	    	//
	    	//	The value of the element of the characterType array indexed by a character
	    	//		defines the characters type.  It is used to 
	    	//		run the tokenizer state machine.
	    	// 
	    	// note that this code need only be run once per RICE initialization.
	    	//
	    	initializeTypeRange( whiteSpaceRange, C_WHITESPACE );
	    	initializeTypeRange( quoteRange, C_QUOTE );
	    	initializeTypeRange( digitRange, C_DIGIT );
	    	initializeTypeRange( baseCharRange, C_LETTER );
	    	initializeTypeRange( ideographicRange, C_LETTER );
	    	initializeTypeRange( combiningRange, C_NAMECHAR );
	    	initializeTypeRange( extendedDigitRange, C_NAMECHAR );
	   		initializeTypeRange( extenderRange, C_NAMECHAR );
	    	initializeTypeRange( hexCharRange, C_HEXCHAR );
			//
			// lightweight character type initializations
			//
			characterType[':'] = C_LETTER;
			characterType['_'] = C_LETTER;
			characterType['%'] = C_PERCENT;
			characterType['&'] = C_AMPERSAND;
			characterType['<'] = C_MARKUPSTART;
			characterType['!'] = C_BANG;
			characterType['['] = C_LEFTBOX;
			characterType['?'] = C_QUESTION;
			characterType['-'] = C_DASH;
			characterType[']'] = C_RIGHTBOX;
			characterType['>'] = C_MARKUPEND;
			characterType['/'] = C_SLASH;
			characterType['*'] = C_STAR;
			characterType[';'] = C_SEMICOLON;
			characterType['#'] = C_SHARP;
			characterType['.'] = C_NAMECHAR;
			characterType['+'] = C_PLUS;
			characterType['='] = C_EQUAL;
			characterType['|'] = C_BAR;
			characterType['('] = C_LEFTPAREN;
			characterType[')'] = C_RIGHTPAREN;
			characterType['x'] = C_X;
	    
			sourceOptions[0] = new SectionDefinition( "none", ScanSource.NONE );
			sourceOptions[1] = new SectionDefinition( "reader", ScanSource.READER );
			sourceOptions[2] = new SectionDefinition( "provider", ScanSource.PROVIDER );
			sourceOptions[3] = new SectionDefinition( "string", ScanSource.STRING );

			characterTypeInitialized = true;
		}
		//
		// Initialize for the default options:
		//
		option_lineEnd = "lf"; 	                /* the XML standard */
		option_evaluateGeneralEntities = true;	/* otherwise the reference is returned */
		option_evaluateParameterEntities = true;/* otherwise the reference is returned */
		option_evaluateCharacterEntities = true;/* otherwise they are handed back as tokens */
		option_scanSource = "none";			    /* default scan nothing */ 
		option_whiteSpaceIsToken = false;	    /* default don't keep whitespace for caller */
        option_ignoreTextIsToken = false;       /* default - don't keep ignore text for caller */
        option_sectionDelimiterIsToken = false; /* default - don't return section delimiters */
        option_keepInnerSectionMarkup = true;   /* default - keep interior IGNORE/unkown section markup */
    }
	
    public XMLscanner( BufferedReader reader )
    {
		this();									/* invoke standard constructor */
		option_scanSource = "reader";	
		source.setInputReader(reader);
		source.setLineEnd( option_lineEnd );	/* set line end option */
    }
	
    public XMLscanner( String scanString )
    {
		this();
		option_scanSource = "string";
		source.setInputReader(null);			/* forget the input reader */
		source.setLine(scanString);
		source.setLineEnd( option_lineEnd );	/* set line end option */
    }
	
    public XMLscanner( BufferedReader reader, String lineEndOption )
    {
		this();
		option_scanSource = "reader";
		source.setInputReader(reader);
		source.setLineEnd( lineEndOption );
    }
    
    public XMLscanner( StringProvider provider, String lineEndOption )
    {
		this();
		option_scanSource = "provider";
		source.setLineProvider( provider );
		source.setLineEnd( lineEndOption );
    }
    
    /*-----------------------------------------------------------------------------+
     |                                                                             |
     |                     XML Scanner Setters and Getters                         |
     |                                                                             |
     +-----------------------------------------------------------------------------*/
    public void setReader( BufferedReader reader )
    {
		source.setInputReader( reader );
    }
	
    public BufferedReader getInputReader( )
    {
		return source.getInputReader();
    }
	
    public void setScanLine( String scanLine )
    {
		source.setLine( scanLine );
		source.setLinePosition(0);
		source.setLineNumber(1);
		lastState = (option_whiteSpaceIsToken)? S_WHITETOKEN:S_WHITESPACE;
    }
	
    public void setWhiteSpaceIsToken( boolean isToken ) 
        { option_whiteSpaceIsToken = isToken; }

    public void setEvaluateGeneralEntities( boolean evalGE )
        { option_evaluateGeneralEntities = evalGE; }

    public void setEvaluateParameterEntities( boolean evalPE )
        { option_evaluateParameterEntities = evalPE; }

    public void setEvaluateCharacterEntities( boolean evalCE )
        { option_evaluateCharacterEntities = evalCE; }

    public void setIgnoreTextIsToken( boolean isToken )
        { option_ignoreTextIsToken = isToken; }

    public void setSectionDelimiterIsToken( boolean isToken )
        { option_sectionDelimiterIsToken = isToken; }

    public void setKeepInnerSectionMarkup( boolean keepIt )
        { option_keepInnerSectionMarkup = keepIt; }

    public void setDtd( DocType aDtd ) { dtd = aDtd; }

    public DocType getDtd() { return dtd; }

    /*----------------------------------------------------------------------+
     |                                                                      |
     |                         XML Scanner Methods                          |
     |                                                                      |
     +----------------------------------------------------------------------*/
    //	
    //  Names and Tokens:
    //	
    //
    //	Letter ::=  baseChar | ideographicChar
    //
	
    //
    //	NameChar ::= Letter | Digit | ExtendedDigit | CombiningChar | 
    //               Extender | '-' | '_' | ':' | '.'	
    //
	    
    //
    //	Name ::= ( Letter | ':' | '_' ) ( NameChar )*
    //
	
    //
    //  NameToken ::= ( NameChar )+
    //
	
    //
    //=======================================================================================
    
    //=======================================================================================
    //
    //
    //  Literals  (in the following the term "Quote" is taken to mean one of 
    //          apostrophe or quotation set for each line.
    //
	
    //
    //	EntityValue ::=  Quote ( [^ '%', '&', Quote ] * | PEreference | Reference )* Quote
    //
    
    //
    //  AttributeValue ::= Quote ( [^ '<', '&', Quote ]* | Reference )* Quote
    //
	
    //
    //  SystemLiteral ::= Quote ( [^ Quote ]* ) Quote
    //
	
    //
    //	PublicIDLiteral ::= Quote [ PublicIDChar - Quote ]* Quote
    //
	
    //
    //	PublicIDChar ;;= [ ' ', '\r', '\n', '-', '\'', '(', ')', '+', ',', '.',
    //					   '/', ':', '=', '?', ';', '!', '*', '#', '@', '$', '_', '%']
    //				  |  [ 'a'..'z', 'A'..'Z', '0'..'9' ]
    //
    
    //
    //	PEreference ::= '%' Name ';'
    //
    
    //
    //  Reference ::=  '&' Name ';'  |  CharacterReference
    //
    
    //  
    //	CharacterReference ::= '&' '#' digit+ ';' | 
    //                         '&' '#' 'x' ( digit | ['a'..'f', 'A'..'F' )+ ';' |
    //					       "&amp;" | "&lt;" | "&gt;" | "&apos;"; | "&quot;"
    //
    // 
    
    //
    //	Comment ::= "<!--" ( [ character - '-' ] | ( '-' [ character - '-'] ) )* "-->"
    //              
	
    //
    //	define control states for token scan:
    //
    //
    private byte Debug = 20;
    //
    private boolean inDTD            = false; // true if in a DTD; false otherwise
    private boolean inEntityValue    = false; // true if in an entity value; false otherwise
    private boolean inAttributeValue = false; // true if in an attribute value; false otherwise.
    
    
    // go into inDTD state so we know to expand parameter entity references
    //
    public void setInDTD( boolean dtdState ) { inDTD = dtdState; }
    public boolean getInDTD() {	return inDTD; }

    public void setInEntityValue( boolean evState ) { inEntityValue = evState; }
    public boolean getInEntityValue() { return inEntityValue; }

    public void setInAttributeValue( boolean avState ) { inAttributeValue = avState; }
    public boolean getInAttributeValue() { return inAttributeValue; }

    public void setDebug( byte db ) { Debug = db; }
    public byte getDebug( ) { return Debug; }
    
    public static boolean isNameChar( char ch )
    {
        byte ctype = characterType[ch];
        switch( ctype )
        {
            case XMLscanner.C_NAMECHAR:
            case XMLscanner.C_HEXCHAR:
            case XMLscanner.C_LETTER:
            case XMLscanner.C_DIGIT:
            case XMLscanner.C_X:
                return true;
            default:
                if ( ( ch=='_' ) ||
                     ( ch==':' ) ||
                     ( ch=='-' ) ||
                     ( ch=='.' )
                   )
                    return true;
                return false;
        }
    }
    
    public static boolean isNameStart( char ch )
    {
        byte ctype = characterType[ch];
        switch ( ctype )
        {
            case XMLscanner.C_LETTER:
            case XMLscanner.C_HEXCHAR:
            case XMLscanner.C_X:
                return true;
            default:
                if ( ch == '_' )
                    return true;
                if ( ch == ':' )
                    return true;
                return false;
        }
    }
    
    public static boolean isWhiteSpace( char ch )
    {
        return ( characterType[ch] == XMLscanner.C_WHITESPACE );
    }
    
    public static boolean isSpecial( char ch )
    {
        byte ctype = characterType[ch];
        switch ( ctype )
        {
            case XMLscanner.C_SPECIAL:
            case XMLscanner.C_QUOTE:
            case XMLscanner.C_PERCENT:
            case XMLscanner.C_AMPERSAND:
            case XMLscanner.C_SHARP:
            case XMLscanner.C_SEMICOLON:
            case XMLscanner.C_MARKUPSTART:
            case XMLscanner.C_MARKUPEND:
            case XMLscanner.C_BANG:
            case XMLscanner.C_QUESTION:
            case XMLscanner.C_DASH:
            case XMLscanner.C_LEFTBOX:
            case XMLscanner.C_RIGHTBOX:
            case XMLscanner.C_BAR:
            case XMLscanner.C_SLASH:
            case XMLscanner.C_STAR:
            case XMLscanner.C_PLUS:
            case XMLscanner.C_EQUAL:
            case XMLscanner.C_LEFTPAREN:
            case XMLscanner.C_RIGHTPAREN:
                return true;
            default:
                return false;
        }
    }
    
    public static boolean isDigit( char ch )
    {
        byte ctype = characterType[ch];
        switch ( ctype )
        {
            case XMLscanner.C_DIGIT:
                return true;
            default:
                return false;
        }
    }

    public static boolean isHexDigit( char ch )
    {
        byte ctype = characterType[ch];
        switch ( ctype )
        {
            case XMLscanner.C_DIGIT:
            case XMLscanner.C_HEXCHAR:
                return true;
            default:
                return false;
        }
    }
    /**
     * token - return an XML token object
     *
     * input:	none -- use XMLscanner constructor to setup the scanner.
     *
     * function:	break out next XML token
     *
     * notes:	A token is defined as a sequence of seperated text according to the 
     *		XML language grammar.
     *
     * algorithm:	Finite State Machine run off from state tables
     *		defined below.  This machine operates from the
     *		array characterType which classifies characters according
     *		to type.  The state machine, then keeps track of
     *		the syntax, while the states are used to keep track of the 
     *		token start and length.
     *
     *		The algorithm is regular expression analysis as implemented with the FSM.   
     *
     * output:	token containing
     *			String - The next token in the sequence.
     *			TokenType - identifier of type of token.
     *			
     * exceptions:
     *
     * references:	
     *
     */
    /* The following are auxillary routines to setup tokens */

    private ScanToken tokenBegin( ScanToken st, StringBuffer sb )
    {
        if ( st == null )
            st = new ScanToken();
        st.lineBegin = source.getLineNumber();
        st.positionBegin = source.getLinePosition();
        sb.setLength(0);
        return st;
    }

    private void tokenEnd( ScanToken st, StringBuffer s, byte rtnType, byte rtnStatus )
    {
        st.string = ( s != null ) ?  s.toString() : null;
        st.lineEnd = source.getLineNumber();
        st.positionEnd = source.getLinePosition();
        st.type = rtnType;
        st.status = rtnStatus;
    }

    private void returnTokenBegin()
    {
        returnToken = tokenBegin( returnToken, tokenBuf );
    }
	
    private void returnTokenEnd( StringBuffer s, byte returnType, byte returnStatus )
    {
        tokenEnd( returnToken, s, returnType, returnStatus );
    }
	
    private void tokenStartUp()
    {
		returnToken = tokenBegin( returnToken, tokenBuf );
        tokenStart = true;
    }
	
    private void commentStartUp()
    {
		commentStart = true;
        returnToken = tokenBegin( returnToken, comment );
    }
	
    private void quotedStringStartUp(char c)
    {
		quotedStart = true;
        returnToken = tokenBegin( returnToken, quotedString );
		quoteChar = c;
    }
	
    private void charEntityStartUp()
    {
		charEntityStart = true;
		charEntity.setLength(0);
		charValue = '\u0000';
    }
	
    private void geReferenceStartUp()
    {
        generalEntityStart = true;
        geReference.setLength(0);
        returnTokenBegin();
    }
	
    private void peReferenceStartUp()
    {
        parameterEntityStart = true;
        peReference.setLength(0);
    }
	
    public ScanToken getToken()
    {
	
        int			 sp = source.linePosition;	/* working index into scan string */
        byte		 state = lastState;		    /* current working state of scan */
        byte		 targetState = state;		/* projected next state for scan */
        // explain the above! 
        byte		 savedState = -1;			/* save state for quoted string, cdata, pcdata */
        char		 c;							/* current character */
        byte		 charType = C_END;			/* default character type */
        byte		 status = ScanToken.OK;		/* one of several token stati */
        byte		 tokenType = ScanToken.ERROR;/* one of several token types */
        int			 i;							/* random index variable */

        // Section state local variables (remove some?)
        StringBuffer name = new StringBuffer(64);/* section name */
        boolean      nameStart = false;          /* true if name started */
        //
        StringBuffer holdingPen = new StringBuffer(64); /* working holding pen */
        // Section state local variables (end)

        /* go get a token */
        /* when a token is returned, the current character pointer is always left 
           pointing to the next character to process.  This is because most lowlevel 
           tokens are terminated by the occurance of a new type of character.  
           Only special characters, comments, sections and entity references have 
           an ending terminating character sequences.  [ Note that all of 
           the angle bracket markup streams have a terminating character sequence 
           as well but they are handled above the scanner level.]
           A consequence of this rule is that getToken never has to get the next 
           character when it is initially entered and this (is designed to) 
           exactly match(es) the initial scanning condition for a new scan.
           
           There are three "state" variables that are particularly useful to understand
           in this state machine:
           
           state -- This is the direct result of computing from the "nextState" array.  It
                    contains the computed next state.  This is the state that is "active"
                    meaning that it is that state that is being acted upon.
           lastState -- This is always the state that was current just before
                    the next state value that was put into state.
           previousLastState -- This is the state before "lastState".
           
           Upon initialization...
        */ 
        previousLastState = state;				/* set previous last state */
        /* get character, character type and new state */
        if ( !source.getAtEOS() )
        {
            c = source.getChar();
            charType = characterType[c];
        }
        else
        {
            c = '\u0000';				/* it's now the empty character */
            charType = C_END;			
        }
        state = nextState[state][charType];
        //
        do 
        {
            boolean printTrace = false;
            // figure out if tracing wanted:
            if ( Debug > 15 )
                printTrace = true;
            else if ( Debug > 10 )
            {
                if ( ( state != S_COMMENT ) && ( state != S_IGNORE ) )
                    printTrace = true;
            }
            if ( printTrace )
            {
                String src = source.debugString();
                Character cc = new Character(c);
                Integer ct = new Integer( charType );
                Integer st = new Integer( state );

                System.out.println( src );
                System.out.println(" Char = " + cc.toString() + 
                                   ", charType = " + ct.toString() +
                                   ", state = " + st.toString() ); 
            }
            switch ( state )
            {
                //===============||
                // Scan states   -- the following states are the main token recognizers
                //===============||
                case S_WHITETOKEN:	/* encountering whitespace in Markup & 
                                       caller wants whitespace tokens */
                    /* if the caller wishes, we can keep track of the whitespace in text */
                    if ( !option_whiteSpaceIsToken )
                    {   
                        state = S_WHITESPACE;
                        continue;
                    }
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c); /* append the character */
                    break;
                case S_WHITESPACE:	/* encountering whitespace that is ignored */
                    if ( option_whiteSpaceIsToken )
                    {
                        state = S_WHITETOKEN;
                        continue;
                    }
                    break;
                case S_NAME:
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c); /* append the character */
                    break;
                case S_NAMETOKEN:
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);	/* append the character */
                    break;
                case S_COMMENT:		/* "<!--" + character? seen */
                    if ( !commentStart )
                    {
                        commentStartUp();
                        comment.append("<!--");
                        /* put in the initial part of the comment. */
                    }
                    /* $FIX$ -- following limitation could be made a matter of policy */
                    if ( comment.length() <= Constants.MAXIMUM_COMMENT_LENGTH )
                        comment.append(c);	/* append the character */
                    else
                        returnToken.status = ScanToken.TRUNCATED;
                    break;
                case S_MARKUPBGN:	/* < seen */
                    break;
                case S_LESSBANG:	/* <! seen */
                    break;
                case S_DIGIT:		/* { "&#" digit+ } seen, character entity reference */
                    charEntityIsHex = false; /* remember char entity kind for error */
                    if ( !charEntityStart )
                        charEntityStartUp();
                    charEntity.append(c); /* append the digit to the digit string */
                    /* compute the character value as we go */
                    charValue = (char) ( ( charValue * 10) + ( c - '0' ) );	
                    break;
                case S_HEXCHAR:		/* { "&#x" } seen, or { "&#x" hexdigit+ } seen */
                    charEntityIsHex = true;	/* remember char entity kind for error */
                    if ( !charEntityStart ) /* { "&#x" } seen */
                        charEntityStartUp();
                    else			/* { "&#x" hexdigit+ } seen */
                    {
                        charEntity.append(c); /* append the (hex) digit to the digit string */
                        /* compute the character value as we go. */
                        char cv =(char) ( ( '0' <= c && c <= '9' ) ? ( c - '0' ) :
                                          ( ( 'A' <= c && c <= 'F' ) ? ( c - 'A' + 10 ) : 
                                            ( ( ( 'a' <= c && c <= 'f' ) ? ( c - 'a' + 10 ) : 0 ) ) ) );
                        charValue = (char) ( ( charValue * 16 ) + cv );
                    }
                    break;
                case S_QUOTE:		/* Start of quoted String */
                    if ( !quotedStart )
                        quotedStringStartUp(c);			/* setup the start of the string */
                    else
                        quotedString.append(c);			/* save the character */
                    break;
                case S_SGEREF:		/* & seen */
                    savedState = lastState;			/* save state just before we got here */
                    entityChar = '&';
                    returnTokenBegin();		
                    break;
                case S_REFCHAR:		/* &# seen */
                    break;
                case S_SPEREF:		/* % seen */
                    returnTokenBegin();
                    if ( !inDTD )  // Parameter entity references only valid in DTD 
                    {
                        state = S_SPECIAL;    /* set the new state */
                        continue;	      /* start back at the top */	
                    }
                    else
                    {
                        entityChar = '%';
                        savedState = lastState;		// save state just before we got here
                    }
                    break;
                case S_PENAME:		/* { "%" namecharacter } seen */
                    if ( !parameterEntityStart )
                        peReferenceStartUp();
                    peReference.append(c);	
                    break;
                case S_REFNAME:		/* general entity reference name state */
                    if ( !generalEntityStart )
                        geReferenceStartUp();
                    geReference.append(c);
                    break;
                case S_LESSQUEST:	/* "<?" seen */
                    break;
                case S_LBDASH1:		/* "<!-" seen */
                    break;
                case S_ECDASH1:		/* "<!-- char* - " seen */
                    comment.append(c);
                    break;
                case S_ECDASH2:		/* "<!-- char* --" seen */
                    comment.append(c);
                    break;
                case S_LBBOX:		/* "<![" seen */
                    break;
                case S_SPECIAL:		/* special character seen */
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);
                    break;
                case S_LBBOXNAME:	/* { "<![" namePrefix } seen */
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);
                    break;
                case S_LBNAME:		/* { "<!" namePrefix } seen */
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);
                    break;		
                case S_LQNAME:		/* { "<?" namePrefix } seen */
                    // If came from a white space state and white space is a token then
                    if ( (lastState == S_LQWHITE) && ( option_whiteSpaceIsToken ) && tokenStart )
                    {
                        targetState = S_LQNAME;
                        returnTokenEnd( tokenBuf, ScanToken.WHITESPACE, ScanToken.OK );
                        state = IS_WHITESPACE;
                        tokenStart = false;
                        break;
                    }
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);
                    break;
                case S_LQWHITE:     /* { "<?" whitespace } seen */
                    /* if the caller wishes, we can keep track of the whitespace in text */
                    if ( option_whiteSpaceIsToken )
                    {   
                        if ( !tokenStart )
                            tokenStartUp();
                        tokenBuf.append(c); /* append the character */
                    }
                    break;
                case S_PIWHITE:     /* { "<? name whitespace } seen */
                    /* if the caller wishes, we can keep track of the whitespace in text */
                    if ( option_whiteSpaceIsToken )
                    {   
                        if ( !tokenStart )
                            tokenStartUp();
                        tokenBuf.append(c); /* append the character */
                    }
                    break;
                case S_PIINST:      /* { "<?" name whitespace nonwhite (char-'?')* } seen */
                    // if keeping whitespace and came from a whitespace state and we have white space then return it.
                    if ( (lastState == S_PIWHITE ) && option_whiteSpaceIsToken && tokenStart )
                    { 
                        // Whitespace is in the token
                        targetState = S_PIINST;
                        returnTokenEnd( tokenBuf, ScanToken.WHITESPACE, ScanToken.OK );
                        state = IS_WHITESPACE;
                        tokenStart = false;
                        break;
                    }
                    if ( !tokenStart )
                        tokenStartUp();
                    if ( (lastState == S_LONEQUEST) )
                        tokenBuf.append('?');
                    tokenBuf.append(c); /* append the character */
                    break;
                case S_LONEQUEST:   /* { "<?" name whitespace nonwhite (char-'?')* "?" } seen */
                    if ( (lastState == S_LONEQUEST) )
                    {
                        // This code is designed to handle the case of repeated '?' characters
                        // in silly constructs such as "<? ?????????>"
                        if ( !tokenStart )
                            tokenStartUp();
                        tokenBuf.append('?');
                    }
                    break;
                    
                case S_QUESTION:    /* { "?" seen -- now look for ">" } */
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);
                    break;
                    
                case S_INCLUDE:		/* { "<![" name "[" char* } seen */
                    // in theory, this state is NEVER executed.  (so, why is it here?)
                    break;
                case S_INCLUDEBOX:	/* { "<![" name "[" char* "]" } seen */
                    if ( includeNest <= 0 )  
                    {   /* if not in a section treat as special */
                        state = IS_SPECIAL;
                        continue;
                    }
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);
                    break;
                case S_INCLUDEBB:	/* { "<![" name "[" char* "]]" } seen */
                    tokenBuf.append(c);
                    break;
                case S_MARKUPEND:   /* { ">" } seen */
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c);
                    break;
                case S_ELEMENTEND:	/* { "</" } seen */
                    break;
                case S_SLASH:       /* { "/" } seen */
                    break;
                case S_LBBBLANK:    /* { "<![ S? } blanks after less bang box */
                    break;
                case S_LBBNAMEBL:   /* { "<![ S? name S? } blanks after name */
                    break;
                //=======================================++
                // IGNORE SECTION STATE MACHINE HANDLER: ||
                //=======================================++
                case S_IGNORE:  /* scanning { "<![" s? "IGNORE" s? "[" char* } */
                    if ( holdingPen.length() > 0 )
                    {
                        longBuf.append( holdingPen.toString() );
                        holdingPen.setLength(0);
                    }
                    longBuf.append(c);
                    break;
                case S_IGNLESS:  /* "<" in IGNORE or "unknown" section */
                    holdingPen.append(c);
                    // following implements fact that "<" is ordinary char in CDATA section:
                    if ( savedSection == SectionDefinition.CDATA )
                        state = S_IGNORE;
                    break;
                case S_IGNLESSB: /* "<!" in IGNORE or "unknown" section */
                    holdingPen.append(c);
                    break;
                case S_IGNLBOX:  /* "<![" in IGNORE or "unknown" section */
                    holdingPen.append(c);
                    if ( option_keepInnerSectionMarkup )
                        longBuf.append( holdingPen.toString() );
                    holdingPen.setLength(0);
                    ignoreNest++;
                    break;
                case S_IGNRBOX:     /* "]" seen in IGNORE/"unknown" section */
                    holdingPen.append(c);
                    break;
                case S_IGNRBXRBX:   /* "]]" seen */
                    holdingPen.append(c);
                    break;
                case S_IGNGREAT:    /* "]]>" seen in "IGNORE/unknown" section */
                    holdingPen.append(c);
                    if ( --ignoreNest <= 0 )
                    {
                        ignoreNest = 0; // make sure it stays at zero.
                        state = IS_SECTION;
                        if ( option_ignoreTextIsToken )
                        {   // caller wants the ignored content
                            returnTokenEnd( longBuf, savedSection, ScanToken.OK);
                            targetState = S_WHITESPACE;
                        }
                        else
                        {   // treat "]]>" as whitespace...
                            state = S_WHITESPACE;
                        }
                        // pop two level name/section stack
                        savedSection = lastSavedSection;
                        savedSectionName = lastSavedSectionName;
                        // zero for efficiency...
                        longBuf.setLength(0);
                        break;
                    }
                    else // pop up nesting level and carry on....
                    {
                        if ( option_keepInnerSectionMarkup )
                            longBuf.append( holdingPen.toString() );
                        // following clause makes sure that the inner section
                        // looks like:  name "[" char* "]"
                        else
                            longBuf.append( ']' );
                        state = S_IGNORE;
                    }
                    holdingPen.setLength(0);
                    break;
                case S_TAGNAME:
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenBuf.append(c); /* append the character */
                    break;
                    //===========================================================||
                    // ending states  -- the following states end the token scan ||
                    //===========================================================||
                case IS_CHARREFDIG:	/* a token of the form { "&#" d+ ';' } has been scanned */
                case IS_CHARREFHEX:	/* a token of the form { "&#x" h+ ';' } has been scanned */
                    if ( option_evaluateCharacterEntities )
                    {
                        switch ( savedState )
                        {
                            case S_QUOTE:
                                /* in this case we append the value represented 
                                   by the character reference */
                                quotedString.append(charValue);
                                break;
                            case S_COMMENT:
                                comment.append(charValue);
                                break;
                            default:
                                tokenBuf.append(charValue);
                                break;
                        }
                        charValue = '\u0000';		/* reset character value */
                        state = savedState;			/* return to the state from whence we came */
                    }
                    else
                    {
                        // returning the reference as a token -- 
                        // point past the semicolon for next scan.
                        source.incrementLinePosition(1);
                        returnTokenEnd( tokenBuf, ScanToken.GENERAL_ENTITY, ScanToken.OK );
                        targetState = S_WHITESPACE;
                    }
                    charEntityStart = false;
                    break;
		
                case IS_GEREF: /* a token of the form { '&' name ';' } has been scanned */
                    // point past the ; character that ended this entity token.
                    source.incrementLinePosition(1);
                    // check for the built in character references:  
                    // &gt; &lt; &apos; &quot; &amp;
                    // should we enforce the declaration of the builtins?
                    charValue = '\u0000';
                    for ( i = 0; i < builtInEntities.length; i++ )
                    {
                        if ( builtInEntities[i][1].equals(tokenBuf.toString()) )
                        {
                            charValue = builtInEntities[i][2].charAt(0);
                            break;
                        }
                    }
                    // did we find one?
                    if ( charValue != '\u0000' )
                    {
                        switch ( savedState )
                        {
                            case S_QUOTE:
                                quotedString.append(charValue);  
                                break;
                            case S_COMMENT:
                                comment.append(charValue);
                                break;
                            default:
                                tokenBuf.append(charValue);
                                break;
                        }
                        state = savedState;
                        charValue = '\u0000';				/* forget old value */
                        charEntityStart = false;			/* character not started */
                    }
                    else if ( ! inDTD && option_evaluateGeneralEntities )
                    {
                        // expand the general entity here if not scanning a 
                        // DTD and user wants the expansion.
                        String entityName = geReference.toString();
                        // Following code requires a dtd be defined.
                        if ( dtd.definedGeneralEntity( entityName ) )
                        {
                            EntityDecl entity = dtd.getGeneralEntity( entityName );
                            // The line ends have already been processed.
                            // The current source is pushed onto entityExpansion
                            ScanSource entityExpansion = new ScanSource( 
                                source, entity.getValue(), "include" );
                            // Make sure line number is set correctly in our expansion text.
                            // $FIX$ note entity expansion in ScanSource in case of error.
                            // $FIX$ so that errors in entity definition can be corrected
                            // Line number should be lnsc:PE-lnpe where lnsc is source
                            // line number and lnpe is Parameter entity line number. 
                            entityExpansion.setLineNumber( 1 );
                            // set character postition to -1 so it gets incremented to zero.
                            entityExpansion.setLinePosition( -1 );
                            //
                            // switch source to the expansion string here
                            //
                            source = entityExpansion;
                            // now continue the scan with the text inserted
                            state = savedState;
                        }
                        else
                        {
                                // The entity reference appears but is undefined....
                                // Return the entity reference and let a higher
                                // level fix this up.
                        }
                    }
                    else if ( inDTD )
                    {
                        // General Entity references are only permitted in 
                        // literal values in entity definitions or attribute 
                        // literal value definitions in a DTD.  
                        // Any other occurance in a DTD is a FATAL XML error.
                        // This is the place to check for this occurrance.
                        // $FIX$
                    }
                    generalEntityStart = false;
                    targetState = S_WHITESPACE;
                    returnTokenEnd( geReference, ScanToken.GENERAL_ENTITY, ScanToken.OK );
                    break;
                case IS_PEREF: /* a token of the form { '%' name ';' } has been scanned. */
                    source.incrementLinePosition(1); // point past ending semicolon
                    if ( option_evaluateParameterEntities )
                    {
                        String parmEntityName = peReference.toString();
                        // following required dtd be defined.
                        if ( dtd.definedParameterEntity( parmEntityName ) )
                        {
                            EntityDecl entity = dtd.getParameterEntity( parmEntityName );
                            // This scan source will have had it's lineends normalized.
                            ScanSource parmEntityExpansion = new ScanSource( 
                                source, entity.getValue(), "include" );
                            // set the line number back to one.
                            parmEntityExpansion.setLineNumber( 1 );
                            // set the char position so that it will increment to zero.
                            parmEntityExpansion.setLinePosition( -1 );
                            //
                            // switch source to the parameter entity here.
                            //
                            source = parmEntityExpansion;
                            state = savedState;
                        }
                        else
                        {
                            // The entity reference appears but is undeclared ...
                            // This is an XML validity constraint violation....
                            // You could do the interior of the if following but ....
                            /*
                            if ( false )
                            {
                                Object[] args = { "Parameter", parmEntityName };
                                String msg = MessageFormat.format( Messages.entityUndeclared, args ); 
                                XMLparseException pBad = new XMLparseException( msg );
                                throw pBad;
                            }
                            */
                            // return the reference and let an upper level handle it.
                        }
                    }
                    parameterEntityStart = false;
                    targetState = S_WHITESPACE;
                    returnTokenEnd( peReference, ScanToken.PARAMETER_ENTITY, ScanToken.OK );
                    break;
                case IS_QUOTED: /* a token of the form { quote char* quote } has been scanned */
                    /* make sure that the beginning quote is the same as this quote.... */
                    if ( c != quoteChar )
                    {
                        state = S_QUOTE; /* reset to the correct state */
                        continue;	     /* go do the correct thing */
                    }
                    source.incrementLinePosition(1);  /* point past ending quote */
                    quotedStart = false;		  /* no longer in quoted string */
                    returnTokenEnd( quotedString, ScanToken.QUOTED_STRING, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    quotedString.setLength(0); /* release content early for memory management */
                    break;
                case IS_COMMENT:  /* a comment has been scanned. */
                    source.incrementLinePosition(1); /* point past closing angle bracket */
                    commentStart = false;		 /* no longer in comment */
                    returnTokenEnd( comment, ScanToken.COMMENT, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    comment.setLength(0); /* release comment early for memory management */
                    break;
                case IS_NAME:  /* a token of the form name (and nameToken) has been scanned */
                    tokenStart = false;  /* no longer in a name scan */
                    returnTokenEnd( tokenBuf, ScanToken.NAME, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_NMTOKEN:  /* a token of the form nameToken has been scanned */
                    tokenStart = false;   /* no longer in a name token scan */
                    returnTokenEnd( tokenBuf, ScanToken.NAME_TOKEN, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_LBNAME:  /* a token of the form { "<!" name } has been scanned. */
                    tokenStart = false;  /* no longer in a name token scan */
                    returnTokenEnd( tokenBuf, ScanToken.BANG_NAME, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_LBBOXNAME:	/* token of the form { "<![" S? name S? } has been scanned. */
                    tokenStart = false; /* no longer in a name token scan */
                    /* check to see if this is a possible known section name */
                    // current scan is as follows:
                    // "<![" name "?"  
                    //             ^ where if "?" is a "[" then we have a section starting.
                    if ( charType == C_LEFTBOX )
                    {
                        // push down on two level stack...
                        lastSavedSectionName = savedSectionName;
                        lastSavedSection = savedSection;
                        //
                        savedSectionName = tokenBuf.toString();
                        savedSection = SectionDefinition.UNKNOWN;

                        // should savedSection be pushDown stack?(Only if nested unknown sections supported)
                        savedSection = SectionDefinition.getIndex( savedSectionName );
                        // figure out what kind of a section to scan:
                        switch( savedSection )
                        {
                            case SectionDefinition.CDATA:
                                // The CDATA section is terminated by the first "]]>" seen.  
                                // CDATA sections do NOT nest.
                                // Use the IGNORE section state machine to collect the content
                                // Data will be stored in longBuf.
                                state = S_IGNORE;
                                ignoreNest++;
                                break;
                            case SectionDefinition.INCLUDE:
                                // Scan tokenizing will continue.
                                state = S_WHITESPACE;
                                includeNest++;  // increment in case of nesting.
                                break;
                            case SectionDefinition.IGNORE:
                                // Run rapidly toward the balanced ending ]]>
                                // Text is (usually) ignored and discarded.
                                if ( ignoreNest == 0 ) 
                                    longBuf.setLength(0);  
                                ignoreNest++;
                                state = S_IGNORE;
                                break;
                            case SectionDefinition.UNKNOWN:
                                // The UNKNOWN section is treated similarly to an IGNORE --
                                // It is assumed to nest sections but is handed as a 
                                // balanced lump to the external processor.
                                // Text is stored in longBuf.
                                if ( ( ignoreNest == 0 ) && option_ignoreTextIsToken )
                                    longBuf.setLength(0);
                                ignoreNest++;
                                state = S_IGNORE;
                                break;
                            default:
                                // throw an exception if get here $FIX$
                                break;
                        }
                        if ( option_sectionDelimiterIsToken )
                        {   // caller wants section delimiters returned:
                            returnTokenEnd( tokenBuf, ScanToken.SECTION_NAME, ScanToken.OK );
                            targetState = state;
                            state = IS_LBBOXNAME;
                        }
                        else
                        {
                            returnTokenBegin();
                        }
                        break;	 // note that a break will increment char pointer
                    }
                    /* the current character is not a left box */
                    returnTokenEnd( tokenBuf, ScanToken.BANG_BOX_NAME, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_LQNAME:  /* a token of the form { "<?" name } has been scanned. */
                    tokenStart = false; /* no longer looking for a name token */
                    returnTokenEnd( tokenBuf, ScanToken.HUH_NAME, ScanToken.OK );
                    // Make the decision whether a PI or an XML decl (i.e. of the form "<?xml"
                    if ( (returnToken.string.toUpperCase()).equals( "XML" ) )
                        targetState = S_WHITESPACE;
                    else
                        targetState = S_PIWHITE;    /* return to scan for whitespace in PI */
                    break;
                case IS_PI:  /* a Processing Instruction has been scanned. */
                    tokenStart = false; /* no longer looking for processing instruction */
                    returnTokenEnd( tokenBuf, ScanToken.PROC_INSTR, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_SPECIAL:  /* a special character has been scanned */
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.SPECIAL, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_PERCENT:  /* a % sign was seen, scan points one past % sign and this character 
                                     did not start a Parameter Entity reference */
                    if ( ! tokenStart )
                        tokenStartUp();
                    tokenBuf.append( '%' );
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.SPECIAL, ScanToken.OK );
                    break;
                case IS_QUESTION: /* "?" seen (and not a right angle(>) */
                    if ( ! tokenStart )
                        tokenStartUp();
                    tokenBuf.append( '?' );
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.SPECIAL, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_QMARKEND:
                    source.incrementLinePosition(1); // point past last character.
                    if ( ! tokenStart )
                    {
                        tokenStartUp();
                        tokenBuf.append("?>");
                    }
                    else
                        tokenBuf.append( '>' );
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.PI_END, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_MARKUP:  /* "<" seen (and not a bang(!), slash(/) or huh(?) ) */
                    if ( ! tokenStart )
                    {
                        // $FIX$ This is one past where the token actually began....
                        tokenStartUp();
                        tokenBuf.append( '<' );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_BEGIN, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_BANGMARKUP:	/* "<!" seen (and not a name prefix ) */
                    if ( ! tokenStart )
                    {
                        // $FIX$ This is two past where the token actually began....
                        tokenStartUp();
                        tokenBuf.append( "<!" );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_BANG, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_QUESMARKUP: /* "<?" seen ( and not a name prefix ) */
                    if ( ! tokenStart )
                    {
                        // $FIX$ This is two past where the token actually began....
                        tokenStartUp();
                        tokenBuf.append( "<!" );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_QUESTION, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_EMPTYELEM:	/* "/>" seen */
                    if ( ! tokenStart )
                    {
                        // $FIX$ This is two past where the token actually began....
                        tokenStartUp();
                        tokenBuf.append( "/>" );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_EMPTY_END, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_BBMARKUP:	/* "<![" seen(and not following letter(maybe whitespace?))*/
                    if ( ! tokenStart )
                    {
                        // $FIX$ This is three past where the token actually began....
                        tokenStartUp();
                        tokenBuf.append( "<![" );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_BANG_BOX, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_LBDMARKUP:	/* "<!-" seen (and not a following dash!) */
                    if ( ! tokenStart )
                    {
                        // $FIX$ This is three past where the token actually began....
                        tokenStartUp();
                        tokenBuf.append( "<!-" );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_BANG_DASH, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_ENDTAG:		/* "</" seen */
                    if ( ! tokenStart )
                    {
                        // $FIX$ This is two past where the token actually began....
                        tokenStartUp();
                        tokenBuf.append( "</" );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_TAGEND, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_MARKUPEND:	/* ">" seen */
                    if ( ! tokenStart )
                    {
                        tokenStartUp();
                        tokenBuf.append( '>' );
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.MARKUP_END, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_SECTION:	/* { "<![" name "[" char* "]]>" } seen */
                    // here because ending { "]"*  "]]>" } seen for section.
                    // sp points directly at the '>'.
                    tokenBuf.append('>');
                    if ( tokenBuf.length() > ("]]>").length() )
                    {   // there is a non-zero right box prefix before the section end.
                        // so, first return it as a single token.  Setup so that
                        // the next token "seen" will be "]]>".  Do this by setting
                        // the target state to S_INCLUDEBB which means that scanner has seen
                        // "]]".  sp points at ">" so upon return scanner comes right back
                        // here -- but does not enter this if clause.
                        tokenBuf.setLength( tokenBuf.length() - ("]]>").length() );
                        returnTokenEnd( tokenBuf, ScanToken.RBOX_SPAN, ScanToken.OK );
                        targetState = S_INCLUDEBB;
                        tokenBuf.setLength( 0 );
                        tokenBuf.append("]]");
                        break;
                    }
                    tokenStart = false;
                    source.incrementLinePosition(1); // point past the '>'.
                    includeNest = ( includeNest > 0 ) ? includeNest - 1 : 0 ;
                    if ( option_sectionDelimiterIsToken )
                    {   // caller wants section delimiters returned.
                        // put the name of the section that we've processed in the token since
                        // the code explains that it is an end ("]]>")
                        if ( savedSectionName.length() == 0 )
                        {   // can't tell caller other than it's a section end.
                            returnTokenEnd( tokenBuf, ScanToken.SECTION_END, ScanToken.OK );
                        }
                        else
                        {   // tell caller more; i.e.,  which section by name and code.
                            tokenBuf.setLength(0);
                            tokenBuf.append( savedSectionName );
                            returnTokenEnd(tokenBuf, savedSection, ScanToken.OK );
                        }
                        targetState = S_WHITESPACE;
                    }
                    else
                    {
                        state = S_WHITESPACE;   // ignore the ending section marker.
                    }
                    // pop two level stack...
                    savedSection = lastSavedSection;
                    savedSectionName = lastSavedSectionName;
                    // make sure names always in sync:
                    if ( ( includeNest == 0 ) && ( ignoreNest == 0 ) )
                    {
                        savedSectionName = "";
                        lastSavedSectionName = "";
                        savedSection = ScanToken.SECTION_UNKNOWN;
                        lastSavedSection = ScanToken.SECTION_UNKNOWN;
                    }
                    break;
                case IS_WHITESPACE:	/* string of white space just ended */
                    tokenStart = false;
                    if ( option_whiteSpaceIsToken )
                    {
                        returnTokenEnd( tokenBuf, ScanToken.WHITESPACE, ScanToken.OK );
                        targetState = S_WHITESPACE;
                    }
                    else
                    {   // ignore whitespace
                        tokenBuf.setLength(0);  // forget it for the sake of memory management;
                        state = S_WHITESPACE;
                        continue;
                    }
                    break;
                case IS_RBOX: /*  ] seen not followed by another ] */
                    if ( !tokenStart )
                    {   /* note token really starts one past where we record it $FIX$ */
                        tokenStartUp();
                        tokenBuf.append(']');
                    }
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.SPECIAL, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IS_RBOXRBOX:
                    if ( !tokenStart )
                    {   /* note: token starts two past where it's recorded... $FIX$ */
                        tokenStartUp();
                        tokenBuf.append("]");
                    }
                    
                    tokenStart = false;
                    returnTokenEnd( tokenBuf, ScanToken.SPECIAL, ScanToken.OK );
                    // "]]" has been seen.  Tell caller about the first RBOX and
                    // 
                    targetState = S_WHITESPACE;
                    break;
                case IS_TAGNAME:
                    tokenStart = false;  /* no longer in a name scan */
                    returnTokenEnd( tokenBuf, ScanToken.TAG_NAME, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;

                    //===========================================================================||
                    // Termination states; they may terminate the scan and usually in error!     ||
                    //===========================================================================||
                case IN_CHARREF: /* scanned &#d+$ -- where $ was NOT a ; */
                    //  
                    switch ( savedState )
                    {
                        case S_QUOTE: /* came from a quote string */
                            // $FIX$ -- note that charcter references can appear in some literals.
                            quotedString.append( (charEntityIsHex)? "&#x" : "&#" );
                            quotedString.append( charEntity.toString() );
                            returnTokenEnd( quotedString, ScanToken.ERROR, ScanToken.QUOTE_ERROR );
                            break;
                        case S_COMMENT:	/* came from a comment */
                            comment.append( (charEntityIsHex)? "&#x" : "&#" );
                            comment.append( charEntity.toString() );
                            returnTokenEnd( comment, ScanToken.ERROR, ScanToken.COMMENT_ERROR );
                            break;
                        case S_INCLUDE:
                            longBuf.append( (charEntityIsHex)? "&#x" : "&#" );
                            longBuf.append( charEntity.toString() );
                            returnTokenEnd( longBuf, ScanToken.ERROR, ScanToken.CHAR_ENTITY_ERROR );
                            break;				
                        default: /* just random characters */
                            tokenBuf.append( (charEntityIsHex)? "&#x" : "&#" );
                            tokenBuf.append( charEntity.toString() );
                            returnTokenEnd( tokenBuf, ScanToken.ERROR, ScanToken.CHAR_ENTITY_ERROR );
                            break;
                    }
                    // following forces scanner out of IN_CHARREF;
                    state = savedState;
                    break;
                case IN_QUOTED:
                    returnTokenEnd( quotedString, ScanToken.QUOTED_STRING, ScanToken.QUOTE_ERROR );
                    targetState = S_WHITESPACE;
                    break; 
                case IN_COMMENT:
                    returnTokenEnd( comment, ScanToken.COMMENT, ScanToken.COMMENT_ERROR );
                    targetState = S_WHITESPACE;
                    break;
                case IN_SECTION:
                    returnTokenEnd( longBuf, savedSection, ScanToken.ENDED_IN_SECTION );
                    targetState = S_WHITESPACE;
                    break;
                case IN_MARKUP:
                case IN_MARKUPBANG:
                case IN_MARKUPQUES:
                case IN_LBDASH1:
                case IN_LBBOX:
                case IN_LBBOXNAME:
                case IN_LBNAME:
                case IN_LQNAME:
                {
                    String ts;
                    if ( !tokenStart )
                        tokenStartUp();
                    tokenStart = false;
                    switch( state )
                    {
                        case IN_MARKUP:
                            tokenType = ScanToken.MARKUP_BEGIN;
                            status = ScanToken.ENDED_IN_MARKUP;
                            ts = "<";
                            break;
                        case IN_MARKUPBANG:
                            tokenType = ScanToken.MARKUP_BANG;
                            status = ScanToken.ENDED_IN_MARKUP;
                            ts = "<!";
                            break;
                        case IN_MARKUPQUES:
                            tokenType = ScanToken.MARKUP_QUESTION;
                            status = ScanToken.ENDED_IN_MARKUP;
                            ts = "<?";
                            break;
                        case IN_LBDASH1:
                            tokenType = ScanToken.MARKUP_BANG_DASH;
                            status = ScanToken.ENDED_IN_MARKUP;
                            ts = "<!-";
                            break;
                        case IN_LBBOX:
                            tokenType = ScanToken.MARKUP_BANG_BOX;
                            status = ScanToken.ENDED_IN_MARKUP;
                            ts = "<![";
                            break;
                        case IN_LBBOXNAME:
                            tokenType = ScanToken.BANG_BOX_NAME;
                            status = ScanToken.LBBNAME_ERROR;
                            ts = "";
                            break;
                        case IN_LBNAME:
                            tokenType = ScanToken.BANG_NAME;
                            status = ScanToken.DECL_ERROR;
                            ts = "";
                            break;
                        case IN_PI:
                            tokenType = ScanToken.PROC_INSTR;
                            status = ScanToken.ENDED_IN_PI;
                            ts = "";
                            break;
                        case IN_LQNAME:
                            tokenType = ScanToken.HUH_NAME;
                            status = ScanToken.ENDED_IN_PI;
                            ts = "";
                            break;
                        default:
                            tokenType = ScanToken.ERROR;
                            status = ScanToken.ENDED_IN_MARKUP;
                            ts = "";
                            break;
                    }
                    if ( ts.length() > 0 )
                        tokenBuf.append( ts );
                    targetState = S_WHITESPACE;
                    returnTokenEnd( tokenBuf, tokenType, status );
                }
                    break;
                case IN_GEREF:
                case IN_PEREF:
                case IN_ENTITY:
                    status = (entityChar == '%') ? 
                        ScanToken.PARM_ENTITY_ERROR : ScanToken.GEN_ENTITY_ERROR ;
                    switch ( savedState )
                    {
                        case S_QUOTE: /* came from a quote string */
                            quotedString.append( entityChar );
                            quotedString.append( charEntity.toString() );
                            break;
                        case S_COMMENT:	/* came from a comment */
                            comment.append( entityChar );
                            comment.append( charEntity.toString() );
                            break;
                        case S_INCLUDE:
                            longBuf.append( entityChar );
                            longBuf.append( charEntity.toString() );
                            break;				
                        default: /* just random characters */
                            tokenBuf.append( entityChar);
                            tokenBuf.append( charEntity.toString() );
                            break;
                    }
                    returnTokenEnd( tokenBuf, ScanToken.ERROR, status );
                    state = savedState;
                    targetState = S_WHITESPACE;
                    break;
                case IN_NAMETOKEN:
                    returnTokenEnd( tokenBuf, ScanToken.NAME_TOKEN, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IN_NAME:
                    returnTokenEnd( tokenBuf, ScanToken.NAME, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;
                case IN_SLASH:
                    returnTokenEnd( tokenBuf, ScanToken.SPECIAL, ScanToken.OK );
                    targetState = S_WHITESPACE;
                    break;		
                case IN_IGNORE:        /* end of stream in IGNORE section */
                    returnTokenEnd( longBuf, savedSection, ScanToken.ENDED_IN_IGNORE );
                    targetState = S_WHITESPACE;
                    break;

                    //============================================================||
                    // Scan ending states                                         ||
                    //============================================================||
                case END_EMPTY: /* the all done state -- nothing left in current line */
                    returnTokenEnd( tokenBuf, ScanToken.EMPTY, ScanToken.ENDED );
                    targetState = S_WHITESPACE;
                    break;
                case END_ILLEGAL:
                    source.incrementLinePosition(1);  /* point past the illegal character */
                    targetState = S_WHITESPACE;
                    returnTokenEnd( tokenBuf, ScanToken.ILLEGAL, ScanToken.OK );
                    break;
                    //============================================================||
                    // The default catchall state                                 ||
                    //============================================================||
                default:	/* terrible error in token scan */
                    /* parameters:  state, line, sp */
                    /* throw an exception here ! */
                    /* String is unknownScanState in Messages.java */
                    break;
            }
            //
            // stop looping if there is a token ready to go or a termination.
            if ( state >= IS_START ) 
            {
                if ( Debug > 5 )
                { 
                    String msg = "";
                    if ( (state >= IS_START) && (state < T_STATES) )
                    {
                        System.out.println( returnToken.toString() );
                    }
                    else
                    {
                        System.out.println( tokenBuf.toString() );

                        Object[] args = { new Integer( (int) state), tokenBuf.toString(), 
                                          new Integer(tokenBuf.length()), source.line };
                        msg = MessageFormat.format( Messages.tokenScanned, args );
                        System.out.println( msg );
                    }
                    // $FIX$ log the message
                };
                break;
            }
            //
            // The following code checks for EOS and if found, pops to any previously pushed
            // scan source.  This implements the termination of inserted character sequences
            // from an entity reference by popping back to previously pushed scan source.
            // Because it is possible for an entity reference to contain an entity reference,
            // multiple pushes may be required 
            c = source.getNextChar();  /* get next character */
            if ( source.getAtEOS() )
            {
                while ( source.getAtEOS() && source.okToPop() )
                {
                    source = source.getPrevious();
                }
                // linePosition was left at the first unscanned character in the 
                // pushed down source.  Thus we only need a getChar() call and NOT
                // a getNextChar call.
                c = source.getChar();
            }
            previousLastState = lastState;  /* set last state */
            lastState = state;      /* save current state */
            /* get character, character type and next state */
            charType = ( source.getAtEOS() )? C_END : characterType[c];
            state = nextState[state][charType];
        } while ( state < END_START  ); 	/* end of looking through characters */
	if ( state < IN_START ) 
	    lastState = targetState;			/* reset entry state */
	return returnToken;
    }
}
	


