/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/XMLdisassembly/Handler.java,v 1.6 2000/12/02 23:36:53 bhunt Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

*/

package com.adobe.rice.common.XMLdisassembly;

//
// java imports
//
import java.util.*;
import java.io.*;
import org.xml.sax.*;

//
// icecubes  imports
//
import com.adobe.rice.IceCubes;
import com.adobe.rice.common.*;
import com.adobe.rice.common.policies.*;
import com.adobe.rice.common.storage.logging.*;     // so we can write to the log.

public class Handler 
{   
    /**
     * <dl><dt>parent</dt> 
     *     <dd>Reference to the parent of this handler ( so it knows where to return the
     *         work result.  If null, this handler is the root and better know how to
     *         deal with all the results.
     *     </dd>
     * </dl>
     */
    protected Handler parent;
    
    /**
     * <dl><dt>prepared</dt>
     *     <dd> An array of objects that the parent sets to prepare the children.
     *          The value, if used, is set by the time the first child element's 
     *          startElement call is made so that it is available to all contained 
     *          elements; or it is set in a result call for all subsequent elements.  
     *          The children can access this array by calling parent.getPrepared().
     *     </dd>
     * </dl>
     **/
    protected Object[] prepared = null;

    /**
     * <dl><dt>element</dt>
     *     <dd>the element that the Handler can process.
     *     </dd>
     * </dl>
     */
    protected String element;
    
    /**
     * <dl><dt>elementAttributes</dt>
     *     <dd>a mapping from elementType to Array of AttributeEntries.
     * </dl>
     **/
    protected ParserAttributes elementAttributes;
    
    /**
     * <dl><dt>content</dt>
     *     <dd>the element's content model.
     *     </dd>
     * </dl>
     **/
    protected Content content;
	
    /**
     * <dl><dt>disassembler</dt>
     *     <dd>Reference to the disassembler with which the Handler is associated.
     *     </dd>
     * </dl>
     */
    protected XMLDisassembler disassembler;
    
    /**
     * <dl><dt>errorCount</dt>
     *     <dd>the number of errors encountered in this handler.
     *     </dd>
     * </dl>
     **/
     protected int errorCount = 0;
     
    /**
     * <dl><dt>errorLevel</dt>
     *     <dd>the maximum severity of the errors encountered.
     *     </dd>
     * </dl>
     **/
    protected int errorLevel = 0;
        
    //
    // Methods
    //
    
    public Handler getParent( ) { return parent; }
    public void setParent( Handler p ) { parent = p; }
    public Object[] getPrepared() { return prepared; }
    public void setPrepared( Object[] prep ) { prepared = prep; }
    public String getElement() { return element; }
    public void setElement( String e ) { element = e; }
    public ParserAttributes getElementAttributes() { return elementAttributes; }
    public void setElementAttributes( ParserAttributes a ) { elementAttributes = a; }
    public Content getContent(){ return content; }
    public void setContent( Content c ) { content = c; } 
    public XMLDisassembler getDisassembler() { return disassembler; }
    public void setDisassembler( XMLDisassembler x ) { disassembler = x; }
    public void setErrorCount( int count ) { errorCount = count; }
    public int bumpErrorCount( int count ) { errorCount += count; return errorCount; }
    public int bumpErrorCount() { errorCount++; return errorCount; }
    public int getErrorCount() { return errorCount; }
    public void setErrorLevel( int level ) { errorLevel = level; }
    public void adjustErrorLevel( int level ) { errorLevel = ( level > errorLevel )? level: errorLevel; }
    public int getErrorLevel( ) { return errorLevel; }
    
    //
    // Utility Methods:
    //

    /**
     * Returns Attribute entry with name attrName in the element with name
     * elementName.  Returns null if not found.
     *
     * $PERFORMANCE$ - change the storage of attribute entry's to a hash table
     *              for faster look up.
     *
     * @param   attrName - the name of the attribute whose entry is returned.
     **/
    public AttributeEntry findAttribute( String attrName )
    {
        AttributeEntry ae = null;
        for ( int i = 0; i < elementAttributes.size(); i++ )
        {
            ae = this.elementAttributes.get(i);
            if (ae.getName().equals(attrName))
                return ae;
        }
        return null;
    }
    
    /**
     * extracts the attributes from the parser and places them in the
     * handler attribute list.
     *
     * @param attributes - of Class ParserAttributes - the "list" of attributes.
     *
     **/
    public void setupAttributes( ParserAttributes attributes )
    {
        TagEntry contentTags[] = null; 
	    AttributeEntry ae;
	    String choice;
	    int logLevel = ((Number) Tray.policyManager.getSelection("LoggingLevel")).intValue();
	    
		// run through the attributes and get their values
		for ( int i = 0; i < attributes.getLength(); i++ )
		{
            String aName = attributes.getName(i);
            String aLName = attributes.getLocalName(i);
            String aType = attributes.getType(i);
            String aValue = attributes.getValue(i);
		    ae = this.findAttribute(aName);
		    if ( ae != null )
		    {
		        ae.setValue(aValue);
		        ae.setType(aType);
		    }
		    else /* This attribute is not defined */
		    {   /* $POLICY$ */
		        if ( logLevel > 6 )
		        {
                    Object[] parms = { aName, aValue, this.element };    
                    XMLdisassemblyMessages.log( XMLdisassemblyMessages.get("attributeUnknown") , parms, ImpactRange.WARNING );
		        }
		            
		        choice = (String) Tray.policyManager.getSafeSelection("AttributeNew", "exception" );
		        if ( choice.equals("add") )
		        {
		            this.elementAttributes.add(new AttributeEntry(aName, aLName, aType, aValue ));
		        }
		        else if ( choice.equals("ignore") )
		        {
		        }
		        else /* if ( choice.equals("exception") ) */ // throw an exception here
		        {
		            // $FIX$ -- throw an exception here.
		        }
		    }
		}
		
    }
    
    /**
     * Return the value of a named attribute
     *
     * @param name
     * @return String - the value of attribute with name or the null string if none found.
     **/
    public String getAttributeValue( String name )
    {
        AttributeEntry ae = this.findAttribute(name);
        if ( ae != null )
            return ae.getValue();
        return "";
    }
    
    //
    // Constructors
    //
    
    /**
     * Creates a new Handler.
     *
     * @param	disassemble  - the disassembler with which the Handler is associated.
     * @param   parent       - the parent handler of this handler.
     */
    public Handler( XMLDisassembler disassembler, Handler parent ) 
    {
        this.disassembler = disassembler;
        this.parent = parent;
    }
	
    /**
     * Creates a new Handler.
     **/
	public Handler( )
	{
	    disassembler = null;
	    parent = null;
	}
	
	/**
	 * Creates new Handler
	 *
	 * @param   parent  - Handler's parent.
	 *
	 **/
	public Handler( Handler parent )
	{
	    this.parent = parent;
	    if ( parent != null )
	        this.disassembler = parent.getDisassembler();
	    else
	        this.disassembler = null;
	}
	
	/**
    * Receive an object for locating the origin of SAX document events. 
	* 
	* Function to set the document locator required by SAXManager interface.
	* It appears this function is designed to set an object to track document physical
	* location information (URL, FileName, Line number, column numer) for
	* parsing errors.  Why it is implemented at this level is unclear.
	*
	**/
	public void setDocumentLocator( String docName, Locator l )
	{
	}
	
	/**
	* Receive notification of the beginning of a document.
	*
	* Function called by payload disassembler at the beginning of an XML document (Payload)
	*
	* @param    docName string containing the name of the document.  (s.b. root element but it is not)
	*
	**/
	public void startDocument( String docName )
	{
	}
		
	/**
    * Receive notification of a processing instruction. 
	* 
	* Function called by payload disassembler when a processing Instruction is seen
	*
	* @param    target  string containing name of the processing instruction target
	* @param    pi      string containing the processing instruction
	**/
	public void processingInstruction( String target, String pi )
	{
	}
	
	/**
    * Begin the scope of a prefix-URI Namespace mapping.
	* 
	* Function called by payload disassembler when a prefix name is seen
	*
	* @param    tag     string containing element name "tag".
	* @param    prefix  string containing prefix name for the name space.
	* @param    uri     string containing the prefix uri for the name space.
	**/
	public void startPrefixMapping(String tag, String prefix, String uri)
	{
	}
			
    /**
     * Receive notification of the beginning of an element. 
     * 
     * Called by the payload disassembler at the beginning of an element understood by the Handler.
     *
     * @param   namespaceURI    the URI of the name space (or null or empty ).
     * @param	localName	    the local name of the element (the second part of prefix:localname).
     * @param   rawName         the complete name of the element ( complete prefix:local )
     * @param	atts            the list of attributes for the element.
     */
    public void startElement(String namespaceURI, String localName, 
                             String rawName,      ParserAttributes atts  ) 
    {
        Object[] parms = { "startElement", "Handler", this.element };    
        XMLdisassemblyMessages.log( XMLdisassemblyMessages.get("MethodNotOverridden") , parms, ImpactRange.WARNING );
    }

    /**
     * Receive notification of the end of an element. 
     *
     * Called by the payload disassembler at the end of an element.
     *
     * @param   namespaceURI    the URI of the name space (or null or empty ).
     * @param	localName	    the local name of the element (the second part of prefix:localname).
     * @param   rawName         the complete name of the element ( complete prefix:local )
     */
    public void endElement(String namespaceURI, String localName, String rawName) 
    {
        Object[] parms = { "endElement", "Handler", this.element };    
        XMLdisassemblyMessages.log( XMLdisassemblyMessages.get("MethodNotOverridden") , parms, ImpactRange.WARNING );
    }
    
    /**
     * Receive notification of character data. 
     *
	 * Called by payload disassembler for character data.
	 *
	 * @param   localName String containing name of element type this #PCDATA is for.
	 * @param	ch		  array of character data.
	 * @param	start	  begining index of character data.
	 * @param	length	  length of character data.
	 *
	 * @see		SAXManager
	 * @see     XMLDisassembler
     **/
	public void characters(String localName, char ch[], int start, int length) 
	{
    }
    
    /**
     * Receive notification of ignorable whitespace in element content.
     *
     * Called by payload disassembler for ignorable whitespace character data
     *
     *
     * @param   localName String containing name of element containing whitespace
     * @param   ch        array of character data
     * @param   start     beginning index of whitespace
     * @param   length    length of whitespace
     *
     * @see     SAXManager
     * @see     XMLDisassembler
     **/
     public void ignorableWhitespace( String localName, char ch[], int start, int length )
     {
     }
    
    
    /**
     * Called by an XMLDisassembler when a sub-handler completes and returns it's work.
	 *
	 * @param   localName   String containing the name of an element just completed.
	 * @param   objectArray Array of objects, which are the result of the element just
	 *                      completed.  Presumably, the called handler's result 
	 *                      method knows how to handle the objects in the array,
	 *                      with the name of the completed element being the
	 *                      distinguishing hint.    
	 *
	 * @see		Handler in package XMLdisassembly (you are here!)
	 *
	 **/
	 public void result( String localName, Object[] objectArray ) 
	 {
	    if ( true )
	    {
            Object[] parms = { "result", "Handler", this.element };    
            XMLdisassemblyMessages.log( XMLdisassemblyMessages.get("MethodNotOverridden") , parms, ImpactRange.WARNING );
        }
	 }
	 
	 /**
	  * Receive notification of a skipped entity
	  *
	  * Called by an XMLDisassembler when an entity is skipped..
	  *
	  * @param  name    String containing the entity name.
	  *
	  **/
      public void skippedEntity( String name ) 
	  {
	  }
          
	 /**
	  * End the scope of a prefix-URI mapping.
	  *
	  * Receive notification of the end of a document.
	  *
	  * Called by an XMLDisassembler when a document completes.
	  *
	  * @param  name    String containing the prefix name that is ending
	  *
	  */
      public void endPrefixMapping( String prefix ) 
	  {
	  }

	 /**
	  * Receive notification of the end of a document.
	  *
	  * Called by an XMLDisassembler when a document completes.
	  *
	  * @param  name    String containing the document name.
	  *
	  */
	  public void endDocument( String docName )
	  {
	  }

	/*-------------------------------------------------------------------+
     |                                                                   |
	 |             E r r o r H a n d l e r  I n t e r f a c e            |
	 |                                                                   |
	 +-------------------------------------------------------------------*/
		 
	/**
     * Receive notification of a warning.
     *
     * SAX parsers will use this method to report conditions that are 
     * not errors or fatal errors as defined by the XML 1.0 recommendation. 
     * The default behaviour is to take no action.
     *
     * The SAX parser must continue to provide normal parsing events 
     * after invoking this method: it should still be possible for the 
     * application to process the document through to the end.
     *
     * @param localName - The local name of the tag - active with the warning.
     * @param exception - The warning information encapsulated in a SAX 
     *                    parse exception.
     *
     * @return boolean - returns true to indicate it is ok to bubble up the error.
     * 
     * @exception SAXException - Any SAX exception, possibly wrapping another exception.
     * 
     * @see SAXParseException
     *
	 **/
		 
    public boolean warning(String localName, Exception exception) 	
    {
        Object[] parms = { "warning", "Handler", localName };    
        XMLdisassemblyMessages.log( XMLdisassemblyMessages.get("MethodNotOverridden") , parms, ImpactRange.WARNING );
        
        parms = new Object[] { exception.getMessage() };    
        XMLdisassemblyMessages.log( Constants.PARSER_WARNING_DEFAULT_MESSAGE, parms, ImpactRange.WARNING );
        
        return true;
    }
        
    /**
     * Receive notification of a recoverable error.
     *
     * This corresponds to the definition of "error" in section 1.2 of 
     * the W3C XML 1.0 Recommendation. For example, a validating parser 
     * would use this callback to report the violation of a validity 
     * constraint. The default behaviour is to take no action.
     *
     * The SAX parser must continue to provide normal parsing events 
     * after invoking this method: it should still be possible for the 
     * application to process the document through to the end. If the 
     * application cannot do so, then the parser should report a fatal 
     * error even if the XML 1.0 recommendation does not require it to 
     * do so.
     * 
     * @param localName - The local name of the tag - active with the error.
     * @param exception - The error information encapsulated in a SAX parse exception.
     *
     * @return boolean - returns true to indicate it is ok to bubble up the error.
     * 
     * @exception SAXException - Any SAX exception, possibly wrapping another exception.
     *
     * @see SAXParseException
     *
     **/
        
    public boolean error(String localName, Exception exception) 
    {
        Object[] parms = { "error", "Handler", localName };    
        XMLdisassemblyMessages.log( XMLdisassemblyMessages.get("MethodNotOverridden") , parms, ImpactRange.SERIOUS );
        
        parms = new Object[] { exception.getMessage() };    
        XMLdisassemblyMessages.log( Constants.PARSER_ERROR_DEFAULT_MESSAGE, parms, ImpactRange.SERIOUS );
        
        return true;
    }
        
    /**
     * Receive notification of a non-recoverable error.
     *
     * This corresponds to the definition of "fatal error" in section 
     * 1.2 of the W3C XML 1.0 Recommendation. For example, a parser would 
     * use this callback to report the violation of a well-formedness 
     * constraint.
     *
     * The application must assume that the document is unusable after
     * the parser has invoked this method, and should continue (if at 
     * all) only for the sake of collecting addition error messages: 
     * in fact, SAX parsers are free to stop reporting any other events 
     * once this method has been invoked.
     *
     * @param localName - The local name of the tag - active with the fatal error.
     * @param exception - The fatal error information encapsulated in a SAX parse exception.
     *
     * @return boolean - returns true to indicate it is ok to bubble up the error.
     * 
     * @exception SAXException - Any SAX exception, possibly wrapping another exception.
     *
     * @see SAXParseException
     *
     **/
        
    public boolean fatalError(String localName, Exception exception) 
    {
        Object[] parms = { "fatalerror", "Handler", localName };    
        XMLdisassemblyMessages.log( XMLdisassemblyMessages.get("MethodNotOverridden") , parms, ImpactRange.FATAL );
        
        parms = new Object[] { exception.getMessage() };    
        XMLdisassemblyMessages.log( Constants.PARSER_FATAL_DEFAULT_MESSAGE, parms, ImpactRange.FATAL );
        
        return true;
    }
    
}

