/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/XMLdisassembly/XMLDisassembler.java,v 1.7 2000/12/02 23:36:53 bhunt Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

*/

package com.adobe.rice.common.XMLdisassembly;

import java.util.*;
import java.io.*;
import java.text.*;                 // for MessageFormat -- used by logging package
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import org.apache.xerces.parsers.*;
import org.apache.xerces.validators.*;

//$JDK1.1$ import com.sun.java.util.collections.*;

//
// ICEcubes imports:
//
import com.adobe.rice.common.storage.logging.*;     // so we can write to the log.


public class XMLDisassembler 
{
    // The XMLDisassembler keeps a hashmap of { name, handlerEntry } pairs.
    // A handlerEntry contains a handler and a handler use count.
    // "use count" contains the number of times the handler is registered.
    // It is incremented each time the tagName is registered and decremented
    // each time the tagname is unregistered.
    // When the "use count" drops to zero, the handler is removed from the
    // hashmap.  A hashmap was chosen to make the look-up cost of the handler
    // run in constant time regardless of the number of (unique) handlers
    // in the map.
    
    // The registeredHandlers HashMap is static (i.e. it lets the use count
    // operate across all the threads).  The advantage is that there is not
    // a needless repetition of the HashMap of registeredHandlers for each
    // thread.
    
    // namePrefix may be used to distinguish between document name spaces. 
    // An element type from one document having the same name as an 
    // element type from another document can be distinguished through
    // having different name prefixes. 
    
    // To keep track of which handler to call when character data is delivered
    // a stack of the current element types is kept.  Whenever a start tag is
    // detected, the element type is pushed on the stack;
    // and when the end tag is detected, the element type is popped from the stack.
    // Whenever character data is delivered, the handler for the element type
    // on the top of the stack provides the character method call.
    
    
	private HashMap     registeredHandlers;
	
	private class HandlerEntry 
	{
	    Handler         handler;        
	    int             useCount;
	    Stack           handlerStack;
	    
	    HandlerEntry()
	    {
	        handler = null;
	        useCount = 0;
	        handlerStack = null;
	    }
	    HandlerEntry( int uc )
	    {
	        this();
	        useCount = uc;
	    }
	    HandlerEntry( int uc, Handler he )
	    {
	        handler = he;
	        useCount = uc;
	    }
	    
	    Handler getHandler() { return handler; }
	    void setHandler( Handler he ) { handler = he; }
	    
	    int getUseCount() { return useCount; }
	    void setUseCount( int uc ) { useCount = uc; }
	    
	    int push( Handler newHandler )
	    {
	        if ( this.handler != null )
	        {
	            if ( handlerStack == null )
	                handlerStack = new Stack();
	            handlerStack.push(this.handler);
	            this.handler = newHandler;
	            useCount++;
	        }
	        else
	        {
	            this.handler = newHandler;
	            this.useCount = 1;
	        }
	        return useCount;
	    }
	    
	    Handler pop()
	    {
	        Handler returnHandler = this.handler;
	        if ( returnHandler != null )
	        {
	            returnHandler = this.handler;
	            if ( handlerStack != null )
	            {
	                this.handler = (Handler) this.handlerStack.pop();
	                if ( this.handlerStack.isEmpty() )
	                    this.handlerStack = null;
	            }
	            useCount--;
	        }
	        return returnHandler;
	    }
	}
	
	//
	// implement a SAXManager Class
	//
	
	
    private class SAXManager extends DefaultHandler
    {
        private XMLDisassembler	disassembler;
        
        private Locator locator;
		
		public SAXManager(XMLDisassembler disassembler) 
		{
			this.disassembler = disassembler;
		}
		
		/**
		* Begins parsing data from stream.
		*
		* @param	input	stream of data to parse.
		*/
		public void beginParsing(Reader input) 
		{
			SAXParser parser = new SAXParser();
			
			//
			// now turn on the features that we want:
			//
			
			// Turn on validation
			
            try 
            {
                parser.setFeature("http://xml.org/sax/features/validation", true);
            } 
            catch (org.xml.sax.SAXException se) 
            {
                se.printStackTrace();
                // $FIX$ log this error
            }
            
            // Turn on namespace support
            
            try 
            {
                parser.setFeature( "http://xml.org/sax/features/namespaces", false );
            }
            catch( SAXNotRecognizedException e )
            { /* $FIX$ log this error */ }
            catch( SAXNotSupportedException  e )
            { /* $FIX$ log this error */ }	
            
            //
            // setup the handlers:
            //
			parser.setContentHandler((ContentHandler) this);
			parser.setErrorHandler( (ErrorHandler) this);
			
			
			//
			// begin the parse
			//
			try 
			{
				parser.parse(new InputSource(input));
			} 
			catch (SAXException saxException) 
			{
			    try { disassembler.fatalError(saxException); }
			    catch ( SAXException se ) 
			    { 
			    //
				    System.out.println("saxException while parsing.");
				    System.out.println(saxException.getMessage() );
				}
				
			} 
			catch (IOException ioException) 
			{
			    try { disassembler.fatalError( new SAXException( ioException.getMessage() ) ); }
			    catch ( SAXException se )
			    {
				    System.out.println("ioException while parsing.");
				    System.out.println(ioException.getMessage() );
				}
			}
		}
		/*-------------------------------------------------------------------+
		 |                                                                   |
		 |       S A X  C o n t e n t H a n d l e r   I n t e r f a c e      |
		 |                                                                   |
		 +-------------------------------------------------------------------*/
		/**
		* Function to set the document locator required by SAXManager interface.
		* It appears this function is designed to set an object to track document physical
		* location information (URL, FileName, Line number, column numer) for
		* parsing errors.  Why it is implemented at this level is unclear.
		*
		**/
		public void setDocumentLocator( Locator l )
		{
		    locator = l;
		    disassembler.setDocumentLocator( locator );
		}
		/**
		* Function called by SAX parser at the beginning of an XML document
		*
		*
		**/
		public void startDocument()
		{
		    disassembler.startDocument();
		}
		
		/**
		* Function called by SAX parser when a processing Instruction is seen
		*
		*
		* @param    target  string containing name of the processing instruction target
		* @param    pi      string containing the processing instruction
		**/
		public void processingInstruction( String target, String pi )
		{
		    disassembler.processingInstruction( target, pi );
		}

	    /**
        * Begin the scope of a prefix-URI Namespace mapping.
	    * 
	    * Function called by SAX Parser when a prefix name is seen
	    *
	    * @param    prefix  string containing prefix name for the name space.
	    * @param    uri     string containing the prefix uri for the name space.
	    **/
	    public void startPrefixMapping(String prefix, String uri)
	    {
	        disassembler.startPrefixMapping( prefix, uri );
	    }
			
		
		/**
		* Function called by SAX parser at the beginning of an element.
		*
		* @param	name	the name of the element.
		* @param	atts	the attribute list for the element.
		*/
        public void startElement(String namespaceURI, String localName, 
                                 String qName, org.xml.sax.Attributes atts) 
		{
		    ParserAttributes pa = new ParserAttributes();
		    for ( int i = 0; i < atts.getLength(); i++ )
		    {
		        AttributeEntry ae = new AttributeEntry(
		                                    atts.getRawName(i), 
		                                    atts.getLocalName(i),
		                                    atts.getType(i),
		                                    atts.getValue(i));
		        pa.add(ae);
		    }
		    pa.trimToSize();
		    
			disassembler.startElement(namespaceURI, localName, qName, pa);
			pa = null;      // release the array.
		}
		
		/**
		* Function called by SAX parser for character data.
		* 
		*
		* @param	ch		array of character data.
		* @param	start	begining index of character data.
		* @param	length	length of character data.
		*/
		public void characters(char ch[], int start, int length) 
		{
			disassembler.characters(ch, start, length);
		}
		
		/**
		* Function called by SAX parser for ignorable white space
		* Note:  Don't access out of the range ch[start] .. ch[ start+length-1 ]
		*
		* 
		* @param    ch      array of white space and other stuff
		* @param    start   beginning index of white space
		* @param    length  length of the white space
		*
		**/
		public void ignorableWhitespace( char ch[], int start, int length )
		{
		    disassembler.ignorableWhitespace( ch, start, length );
		}
		/**
		* Function called by SAX parser at the end of an element.
		*
		* @param	name	the name of the element.
		*/
		public void endElement(String namespaceURI, String localName, String qName) 
		{
			disassembler.endElement(namespaceURI, localName, qName);
		}
		
		/**
		* Function called by SAX parser at the end of a document
		*
		* 
		**/
		
		public void endDocument()
		{
		    disassembler.endDocument();
		}
		
		/*-------------------------------------------------------------------+
		 |                                                                   |
		 |             E r r o r H a n d l e r  I n t e r f a c e            |
		 |                                                                   |
		 +-------------------------------------------------------------------*/
		 
		/**
         * Receive notification of a warning.
         *
         * SAX parsers will use this method to report conditions that are 
         * not errors or fatal errors as defined by the XML 1.0 recommendation. 
         * The default behaviour is to take no action.
         *
         * The SAX parser must continue to provide normal parsing events 
         * after invoking this method: it should still be possible for the 
         * application to process the document through to the end.
         *
         * @param exception - The warning information encapsulated in a SAX 
         *                    exception.
         *
         * @exception SAXException - Any SAX exception, possibly wrapping another exception.
         * 
         * @see SAXParseException
		 *
		 **/
		 
        public void warning(SAXException exception) throws SAXException	
        {
            disassembler.warning( exception );
        }
        
        /**
         * Receive notification of a recoverable error.
         *
         * This corresponds to the definition of "error" in section 1.2 of 
         * the W3C XML 1.0 Recommendation. For example, a validating parser 
         * would use this callback to report the violation of a validity 
         * constraint. The default behaviour is to take no action.
         *
         * The SAX parser must continue to provide normal parsing events 
         * after invoking this method: it should still be possible for the 
         * application to process the document through to the end. If the 
         * application cannot do so, then the parser should report a fatal 
         * error even if the XML 1.0 recommendation does not require it to 
         * do so.
         * 
         * @param exception - The error information encapsulated in a SAX parse exception.
         * 
         * @exception SAXException - Any SAX exception, possibly wrapping another exception.
         *
         * @see SAXParseException
         *
         **/
        
        public void error(SAXException exception) throws SAXException
        {
            disassembler.error( exception );
        }
        
        /**
         * Receive notification of a non-recoverable error.
         *
         * This corresponds to the definition of "fatal error" in section 
         * 1.2 of the W3C XML 1.0 Recommendation. For example, a parser would 
         * use this callback to report the violation of a well-formedness 
         * constraint.
         *
         * The application must assume that the document is unusable after
         * the parser has invoked this method, and should continue (if at 
         * all) only for the sake of collecting addition error messages: 
         * in fact, SAX parsers are free to stop reporting any other events 
         * once this method has been invoked.
         *
         * @param exception - The error information encapsulated in a SAX parse exception.
         * 
         * @exception SAXException - Any SAX exception, possibly wrapping another exception.
         *
         * @see SAXParseException
         *
         **/
        
        public void fatalError(SAXException exception) throws SAXException
        {
            disassembler.fatalError( exception );
        }
        
    }
	
	//
	// per instance variables
	//
	private     String      docPrefix;          // document level name prefix.
	private     Reader      input;
	private     Writer      output;
	private     Stack       nameStack;
	private     Stack       nameSpaceStack;     // stack of name space prefixes
	
	
	public XMLDisassembler()
	{
	    if ( registeredHandlers == null )
	    {
    		registeredHandlers = new HashMap(
		                    Constants.EXPECTED_HANDLER_COUNT, 
		                    Constants.HANDLER_HASHMAP_LOADFACTOR );
		}
		docPrefix = "";
		nameStack = new Stack();
		nameStack.ensureCapacity(Constants.MAXIMUM_HANDLER_STACKSIZE);
	}
	/**
	* @param	input	stream of input to be disassembled.
	* @param    prefix  Document prefix.
	*
	*/ 
	public XMLDisassembler(Reader input) 
	{
	    this();
		this.input = input;
	}
	
	public XMLDisassembler( String prefix )
	{
	    this();
	    this.docPrefix = prefix;
	}
	
	public XMLDisassembler( String prefix, Reader input )
	{
	    this();
	    this.docPrefix = prefix;
	    this.input = input;
	}

	public XMLDisassembler( String prefix, Reader input, Writer output )
	{
	    this();
	    this.docPrefix = prefix;
	    this.input = input;
	    this.output = output;
	}

	
	public void setOuput( Writer out ) { output = out; }
	public Writer getOutput() { return output; }
	
	public void setInput( Reader in ) { input = in; }
	public Reader getInput() { return input; }
	
	public void setDocPrefix( String prefix ) { docPrefix = prefix; }
	public String getDocPrefix( ) { return docPrefix; }
		
	
	/**
	 * Gets the Handler for an element.
	 *
	 * @param   name  - A String containing the element name whose handler to get.
	 *
	 * @return  Handler - the handler for the element or null if none.
	 *
	 **/
	public Handler getHandler( String name )
	{
	    String tagName = this.docPrefix + name;
	    Handler theHandler = null;
	    if ( this.registeredHandlers.containsKey(tagName) )
	    {
	        theHandler = ((HandlerEntry) registeredHandlers.get(tagName)).getHandler();
	    }
	    return theHandler;
	}
	
	/**
	* Registers a handler for the element it understands.
	* 
	* @param	tagName - A String containing the name of the tag to be handled.
	* @param    handler	- the Handler to be registered.
	*
	* @return   Handler - the previous handler, if any.
	
    * In general, there should be a one-to-one correspondence between an 
    * element type and it's associated handler.  This class permits the
    * handler for an element type to be changed.  When this is done, the
    * new handler completely replaces the old.  Specifically, if a start
    * tag with element type, tagName and handler, tagNameHandler can contain
    * itself as a child element and a new handler, say tagNameNewHandler is
    * registered to handle tagName, then when the end element tag is detected
    * for the containing element type, tagName, the handler tagNameNewHandler
    * will be called.   To permit the object registering a handler to manage
    * this circumstance, the previous handler is returned.
	*/
	public synchronized void registerHandler(String name, Handler handler) 
	{
	    String tagName = this.docPrefix + name;
	    if ( registeredHandlers.containsKey( tagName ) )
	    {
	        HandlerEntry he = (HandlerEntry) registeredHandlers.get( tagName );
	        he.push(handler);
	        registeredHandlers.put( tagName, he );
	    }
	    else
	    {
	        HandlerEntry he = new HandlerEntry( 1, handler );
	        registeredHandlers.put( tagName, he );
	    }
	}
	
	/**
	* Sets a handler - ie replaces a tagHandler without altering useCount of tag.
	*   unless the tag has no handler -- in which case it registers the handler.
	*
	* @param    tagName - String containing name of tag whose handler is being restored.
	* @param    handler - Handler containing handler which is being restored.
	*
	* @return   handler - The previously set Handler or null if there was none.
	*/
	public synchronized Handler setHandler( String name, Handler handler )
	{
	    Handler previousTagHandler = null;
	    HandlerEntry he = null;
	    String tagName = this.docPrefix + name;
	    if ( registeredHandlers.containsKey( tagName ))
	    {
	        he = (HandlerEntry) registeredHandlers.get( tagName );
	        previousTagHandler = he.getHandler();
	        he.setHandler(handler);
	    }
	    else
	    {
	        he = new HandlerEntry( 1, handler );
	    }
	    registeredHandlers.put( tagName, he );
	    return previousTagHandler;
	}
	
	/**
	* Unregisters a handler (synchronous)
	*
	* @param    name - a String containing the (unprefixed) name of the element type whose
	*           handler is being unregistered.
	*
	* @return   previous handler for element type, name; or null if there was none.
	*
	* look up the handler by name, if found, 
	* the use count is decremented and if zero the handler (and name) are removed.
	*/
	public synchronized Handler unregisterHandler(String name )                                             
	{
	    Handler prevHandler = null;
	    String tagName = this.docPrefix + name;
	    if ( registeredHandlers.containsKey( tagName ) )
	    {
	        HandlerEntry he = (HandlerEntry) registeredHandlers.get( tagName );
            // decrement the use count.

	        prevHandler = he.pop();
	        if ( he.getUseCount() == 0 )
	        {
	            registeredHandlers.remove( tagName );
	        }
	    }
	    return prevHandler;
	}
	
	/*-------------------------------------------------------------------+
	 |                                                                   |
	 |                 S A X M a n a g e r  I n t e r f a c e            |
	 |                       I m p l e m e n t a t i o n                 |
	 |                                                                   |
	 +-------------------------------------------------------------------*/
	/**
	 * Function to set the document locator required by SAXManager interface.
	 * It appears this function is designed to set an object to track document physical
	 * location information (URL, FileName, Line number, column numer) for
	 * parsing errors.  Why it is implemented at this level is unclear.
	 *
	 **/
	public void setDocumentLocator( Locator l )
	{
		String tagName = this.docPrefix + Constants.DOCUMENT_LOCATOR;
		if ( registeredHandlers.containsKey(tagName) )
		{
	        Handler handler = ( (HandlerEntry) registeredHandlers.get(tagName) ).getHandler();
		    nameStack.push( tagName );
		    
			handler.setDocumentLocator(tagName, l);
		}
		else
		{
		    Object[] parms = { "setDocumentLocater", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"), 
		        parms, 
		        200000000 
		    );
    	    // Indicates that no document locator is registered.  
	    }
	}
	/**
	* Function called by SAX parser at the beginning of an XML document
	*
	*
	**/
	public void startDocument()
	{
		String tagName = this.docPrefix + Constants.DOCUMENT_NAME;
		if ( registeredHandlers.containsKey(tagName) )
		{
	        Handler handler = ( (HandlerEntry) registeredHandlers.get(tagName) ).getHandler();
		    nameStack.push( tagName );
		    
 		    handler.startDocument(tagName);
		}
		else
		{
	        // Indicates that we have an unknown document starting.  This is a
	        // Syntax error and suggests that we should ignore it and it's
            // content. -- WHOA!!!
    		    
	        // We however, log the error and it's content.
	        // $FUTURE$ Probably what we should do is call an unknown document handler!
		    Object[] parms = { "startDocument", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"), 
		        parms, 200000000 
		        );
	    }
	}
		
	/**
	* Function called by SAX parser when a processing Instruction is seen
	*
	*
	* @param    target  string containing name of the processing instruction target
	* @param    pi      string containing the processing instruction
	**/
	public void processingInstruction( String target, String pi )
	{
	    String piName = this.docPrefix + target;
	    String tagName = this.docPrefix + Constants.DOCUMENT_NAME;
		if ( registeredHandlers.containsKey(piName) )
		{
	        Handler handler = ( (HandlerEntry) registeredHandlers.get(piName) ).getHandler();
	        
			handler.processingInstruction(target, pi);
		}
	    else if ( registeredHandlers.containsKey(tagName) )
	    {   // call the document handler if there is one
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        
	        handler.processingInstruction(target, pi);
	    }
	    else
	    {
	        // There is no handler for process instructions -- It is a matter of policy of what
	        // to do in this case.  Implement the default "silent ignore" policy.
	        // $POLICY$
	        // A:  Throw a parsing exception ( warning - level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the characters and the error.
		    Object[] parms = { "processingInstruction", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"),
		        parms,
		        200000000 
		        );
	    }
	}
	
	/**
    * Begin the scope of a prefix-URI Namespace mapping.
	* 
	* Function called by payload disassembler when a prefix name is seen
	*
	* @param    prefix  string containing prefix name for the name space.
	* @param    uri     string containing the prefix uri for the name space.
	**/
	public void startPrefixMapping(String prefix, String uri)
	{
	    String tagName = this.docPrefix + Constants.DOCUMENT_NAME;
	    if ( !nameStack.empty() )
	    {
	        tagName = ( (String) nameStack.peek() ); 
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        String qName = tagName;
	        if ( qName.startsWith(this.docPrefix) )
	        {
	            // extract the non-doc prefixed "raw" name.
	            qName = tagName.substring(this.docPrefix.length());
	        }

	        handler.startPrefixMapping( qName, prefix, uri );
	    }
	    else if ( registeredHandlers.containsKey(tagName) )
	    {   // call the document handler if there is one
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();

	        handler.startPrefixMapping( Constants.DOCUMENT_NAME, prefix, uri );
	    }
	    else
	    {
	        // There is no handler for prefix mapping -- It is a matter of policy of what
	        // to do in this case.  Implement the default "silent ignore" policy.
	        // $POLICY$
	        // A:  Throw a parsing exception ( error - level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the characters and the error.
		    Object[] parms = { "startPrefixMapping", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"),
		        parms,
		        200000000
		        );
	    }
		
	}
			
		
	/**
	* Function called by SAXmanager at the beginning of an element.
	*
    * @param namespaceURI URI of the name space (or null or empty ).
    * @param localName	  local name of the element (the second part of prefix:localname).
    * @param qName      complete name of the element ( complete prefix:local )
    * @param atts         attributes for the element.
	*
	* @see		SAXManager
	*/
	public void startElement(String namespaceURI, String localName, 
                             String qName, ParserAttributes atts ) 
	{
	    String tagName = this.docPrefix + qName;
	    if ( registeredHandlers.containsKey(tagName) )
	    {
	        HandlerEntry he = (HandlerEntry) registeredHandlers.get(tagName);
	        Handler		 curHandler = he.getHandler();
            // push the name on the name stack.
		    nameStack.push( tagName );

			curHandler.startElement(namespaceURI, localName, qName, atts);
		}
		else
		{
	        // Indicates that we have an unknown tag starting.  This is a
	        // Syntax error and suggests that we should ignore it and it's
            // content.
		    
	        // $FIX$ We should, however, log the error and it's content.
	        // Probably what we should do is call an ignore handler!
		    Object[] parms = { "startElement", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"),
		        parms, 
		        200000000
		    );
	    }
	}
	
	/**
	* Function called by SAXmanager for character data.
	*
	* @param	ch		array of character data.
	* @param	start	begining index of character data.
	* @param	length	length of character data.
	*
	* @see		SAXManager
	*/
	public void characters(char ch[], int start, int length) 
	{
	    String tagName = this.docPrefix + Constants.DOCUMENT_NAME;
	    if ( !nameStack.empty() )
	    {
	        // look on top of the stack for the current element name.
	        tagName = ( (String) nameStack.peek() ); 
	        // go get its handler
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        // obtain the qName from the tagName
	        String qName = tagName;
	        if ( qName.startsWith(this.docPrefix) )
	        {
	            qName = tagName.substring(this.docPrefix.length());
	        }

	        handler.characters( qName, ch, start, length );
	    }
	    else if ( registeredHandlers.containsKey(tagName) )
	    {   // call the document handler if there is one
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        
	        handler.characters( Constants.DOCUMENT_NAME, ch, start, length );
	    }
	    else
	    {
	        // There is no handler for characters -- It is a matter of policy of what
	        // to do in this case.  Implement the default "silent ignore" policy.
	        // $POLICY$
	        // A:  Throw a parsing exception ( error - level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the characters and the error.
		    // Object[] parms = { "characters", "XMLDisassembler", tagName };
		    // XMLdisassemblyMessages.log("MethodNotRegistered", parms, 200000000 );
	    }
	}
	
	/**
	* Function called by SAXmanager for ignorable whitespace
	*
	* @param	ch		array of character data.
	* @param	start	begining index of character data.
	* @param	length	length of character data.
	*
	* @see		SAXManager
	*/
	public void ignorableWhitespace(char ch[], int start, int length) 
	{
	    String tagName = this.docPrefix + Constants.DOCUMENT_NAME;
	    if ( !nameStack.empty() )
	    {
	        tagName = ( (String) nameStack.peek() ); 
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        String qName = tagName;
	        if ( qName.startsWith(this.docPrefix) )
	        {
	            qName = tagName.substring(this.docPrefix.length());
	        }

	        handler.ignorableWhitespace( qName, ch, start, length );
	    }
	    else if ( registeredHandlers.containsKey(tagName) )
	    {   // call the document handler if there is one
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();

	        handler.ignorableWhitespace( Constants.DOCUMENT_NAME, ch, start, length );
	    }
	    else
	    {
	        // There is no handler for whitespace -- It is a matter of policy of what
	        // to do in this case.  Implement the default "silent ignore" policy.
	        // $POLICY$
	        // A:  Throw a parsing exception. ( Warning level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the characters and the error.
		    // Object[] parms = { "ignorableWhitespace", "XMLDisassembler", tagName };
		    // XMLdisassemblyMessages.log("MethodNotRegistered", parms, 200000000 );
	    }
	}
	
	/**
	* Function called by SAXmanager at the end of an element.
	*
	* @param	name	the name of the element.
	*
	* @see		SAXManager
	*/
	public void endElement(String namespaceURI, String localName, String qName) 
	{
	    String tagName = this.docPrefix + qName;
	    if ( registeredHandlers.containsKey(tagName))
	    {
	        HandlerEntry he = (HandlerEntry) registeredHandlers.get(tagName);
	        Handler		 curHandler = he.getHandler();
	        try 
	        { 
	            curHandler.endElement(namespaceURI, localName, qName); 
	        } 
	        catch ( Exception e )
	        { // What to do with the exception here?
	          // $FIX$ - perhaps an error message and an error response as well.
	        }
	        finally
	        {
	            nameStack.pop();
	        }
	    }
	    else
	    {
	        // There is no handler for this ending element  -- It is a matter of policy of what
	        // to do in this case.  Implement the default "log the error and ignore it" policy.
	        // $POLICY$
	        // A:  Throw a parsing exception. ( error level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the end tag and the error.
		    Object[] parms = { "endElement", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"),
		        parms,
		        200000000
		    );
	    }
	}
	
	/**
	 * Function called by a handler when it completes work on and element and 
	 * wants to resturn the result.  The result is returned to the element handler
	 * of the element which contained the element just completed.  The element stack is
	 * used to determine the containing element - look up it's handler and call it's
	 * result method.
	 *
	 * @param   name        String containing the name of the element whose work was just completed.
	 * @param   objectArray Array of Objects that is the result.  Presumably, the handler's
	 *                      completion routine knows how to handle the objects in the array
	 *                      with the completing element name as the distinguishing hint.
	 *
	 **/
	 public void result( String name, Object[] objectArray )
	 {
	    String tagName = this.docPrefix + name;
	    // if the following fails -- A fatal error has occured.
	    String currentName = (String) nameStack.pop();  // return result to element handler of element
	                                           // in which the element was part of content.
	    if ( !nameStack.empty() )
	    {
	        Handler handler = ( ( HandlerEntry ) registeredHandlers.get( (String) nameStack.peek() ) ).getHandler();

	        handler.result(name, objectArray );

	        nameStack.push( currentName );  // restore current name.
	    }
	    else
	    {
	        // Wow, the stack is empty and we have a completing element. 
	        // What should be done here?   
	        // A.  call it's own result routine?  
	        // B.  Log an error and ignore?  
	        // C.  Throw an exception?
	        // Should this be a matter of policy? (choice A, B or C) -- yes 
	        // $POLICY$
	        // Implement choice A for right now:  call the current handler's result.
	        nameStack.push( currentName );  // put it back -- tos is the "current" element for call.
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( currentName ) ).getHandler() ;

	        handler.result(name, objectArray);
	        
		    Object[] parms = { "result", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"),
		        parms,
		        ImpactRange.WARNING 
		    );
	    }
	 }
	 
	/**
	* Function called by SAX parser at the end of an XML document
	*
	* @param (none)
	*
	**/
	public void endDocument()
	{
		String tagName = this.docPrefix + Constants.DOCUMENT_NAME;
		if ( registeredHandlers.containsKey(tagName) )
		{
	        HandlerEntry he = (HandlerEntry) registeredHandlers.get(tagName);
	        Handler		 curHandler = he.getHandler();

			curHandler.endDocument(tagName);

	        nameStack.pop();
		}
		else
		{
	        // Indicates that we have an unknown document ending.  This is a
	        // Syntax error and suggests that we should ignore it and it's
            // content. -- WHOA!!!
		    
	        // We, however, log the error and it's content.
	        // $FUTURE$ Probably what we should do is call an unknown document handler!
		    Object[] parms = { "endDocument", "XMLDisassembler", tagName };
		    XMLdisassemblyMessages.log(
		        XMLdisassemblyMessages.get("MethodNotRegistered"),
		        parms, 
		        ImpactRange.WARNING  
		    );
	    }
	}
	
	
	/*-------------------------------------------------------------------+
	 |                                                                   |
	 |             E r r o r H a n d l e r  I n t e r f a c e            |
	 |                                                                   |
	 +-------------------------------------------------------------------*/
		 
	/**
     * Receive notification of a warning.
     *
     * SAX parsers will use this method to report conditions that are 
     * not errors or fatal errors as defined by the XML 1.0 recommendation. 
     * The default behaviour is to take no action.
     *
     * The SAX parser must continue to provide normal parsing events 
     * after invoking this method: it should still be possible for the 
     * application to process the document through to the end.
     *
     * @param exception - The warning information encapsulated in a SAX 
     *                    parse exception.
     *
     * @exception SAXException - Any SAX exception, possibly wrapping another exception.
     * 
     * @see SAXParseException
     *
     **/
		 
    public void warning(SAXException exception) throws SAXException	
    {
        boolean bubbleUp = true;
	    String tagName = this.docPrefix + Constants.WARNING_HANDLER_NAME;
	    if ( !nameStack.empty() )
	    {
	        tagName = ( (String) nameStack.peek() ); 
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        String qName = tagName;
	        if ( qName.startsWith(this.docPrefix) )
	        {
	            qName = tagName.substring(this.docPrefix.length());
	        }

	        bubbleUp = handler.warning( qName, (Exception) exception );
	    }
	    if ( bubbleUp && registeredHandlers.containsKey(tagName) )
	    {   // call the document handler if there is one
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();

	        bubbleUp = handler.warning( Constants.WARNING_HANDLER_NAME, (Exception) exception );
	    }
	    else
	    {
	        // There is no handler for warnings -- It is a matter of policy of what
	        // to do in this case.  Implement the default "log" policy. (B below)
	        // $POLICY$
	        // A:  Throw a parsing exception. ( Warning level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the characters and the error.
            Object[] parms = { exception.getMessage() };    
            XMLdisassemblyMessages.log( Constants.PARSER_WARNING_DEFAULT_MESSAGE, parms, ImpactRange.WARNING  );
	    }
        
     }
        
    /**
        * Receive notification of a recoverable error.
        *
        * This corresponds to the definition of "error" in section 1.2 of 
        * the W3C XML 1.0 Recommendation. For example, a validating parser 
        * would use this callback to report the violation of a validity 
        * constraint. The default behaviour is to take no action.
        *
        * The SAX parser must continue to provide normal parsing events 
        * after invoking this method: it should still be possible for the 
        * application to process the document through to the end. If the 
        * application cannot do so, then the parser should report a fatal 
        * error even if the XML 1.0 recommendation does not require it to 
        * do so.
        * 
        * @param exception - The error information encapsulated in a SAX exception
        * @exception SAXException - Any SAX exception, possibly wrapping another exception.
        *
        * @see SAXParseException
        *
        **/
        
    public void error(SAXException exception) throws SAXException
    {
        boolean bubbleUp = true;
	    String tagName = this.docPrefix + Constants.ERROR_HANDLER_NAME;
	    if ( !nameStack.empty() )
	    {
	        tagName = ( (String) nameStack.peek() ); 
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        String qName = tagName;
	        if ( qName.startsWith(this.docPrefix) )
	        {
	            qName = tagName.substring(this.docPrefix.length());
	        }

	        bubbleUp = handler.error( qName, (Exception) exception );
	    }
	    if ( bubbleUp && registeredHandlers.containsKey(tagName) )
	    {   // call the document handler if there is one
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();

	        bubbleUp = handler.error( Constants.ERROR_HANDLER_NAME, (Exception) exception );
	    }
	    else
	    {
	        // There is no handler for warnings -- It is a matter of policy of what
	        // to do in this case.  Implement the default "log" policy. (B below)
	        // $POLICY$
	        // A:  Throw a parsing exception. ( Warning level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the characters and the error.
            Object[] parms = { exception.getMessage() };    
            XMLdisassemblyMessages.log( Constants.PARSER_ERROR_DEFAULT_MESSAGE, parms, ImpactRange.SERIOUS );
	    }
    }
        
    /**
        * Receive notification of a non-recoverable error.
        *
        * This corresponds to the definition of "fatal error" in section 
        * 1.2 of the W3C XML 1.0 Recommendation. For example, a parser would 
        * use this callback to report the violation of a well-formedness 
        * constraint.
        *
        * The application must assume that the document is unusable after
        * the parser has invoked this method, and should continue (if at 
        * all) only for the sake of collecting addition error messages: 
        * in fact, SAX parsers are free to stop reporting any other events 
        * once this method has been invoked.
        *
        * @param exception - The error information encapsulated in a SAX parse exception.
        * 
        * @exception SAXException - Any SAX exception, possibly wrapping another exception.
        *
        * @see SAXParseException
        *
        **/
        
    public void fatalError(SAXException exception) throws SAXException
    {
        boolean bubbleUp = true;
	    String tagName = this.docPrefix + Constants.FATAL_HANDLER_NAME;
	    if ( !nameStack.empty() )
	    {
	        tagName = ( (String) nameStack.peek() ); 
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();
	        String qName = tagName;
	        if ( qName.startsWith(this.docPrefix) )
	        {
	            qName = tagName.substring(this.docPrefix.length());
	        }

	        bubbleUp = handler.fatalError( qName, (Exception) exception );
	    }
	    if ( bubbleUp && registeredHandlers.containsKey(tagName) )
	    {   // call the document handler if there is one
	        Handler handler = ( (HandlerEntry) registeredHandlers.get( tagName ) ).getHandler();

	        bubbleUp = handler.fatalError( Constants.FATAL_HANDLER_NAME, (Exception) exception );
	    }
	    else
	    {
	        // There is no handler for warnings -- It is a matter of policy of what
	        // to do in this case.  Implement the default "log" policy. (B below)
	        // $POLICY$
	        // A:  Throw a parsing exception. ( Warning level )
	        // B:  Log the error and ignore it.
	        // C:  Silently Ignore the characters and the error.
            Object[] parms = { exception.getMessage() };    
            XMLdisassemblyMessages.log( Constants.PARSER_FATAL_DEFAULT_MESSAGE, parms, ImpactRange.FATAL );
	    }
    }
    
    /**
     * start up the parse for this disassembler.
     *
     **/
    public void parse ()
    {
        SAXManager manager = new SAXManager (this);
        manager.beginParsing(input);
    }
    	
}
















