    /*
    $Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/collections/BitVector.java,v 1.1.1.1 2000/09/19 19:39:44 bhunt Exp $

    <<ICECOPYRIGHT>>
    */

    /*----------------------------------------------------------------------------------------+
    |                                                                                        |
    |                              Reference Implementation of ICE                           |
    |                                                                                        |
    |                         Collections -- BitVector Class Definition                      |
    |                                                                                        |
    +----------------------------------------------------------------------------------------*/ 
    package com.adobe.rice.common.collections;

    import com.adobe.rice.common.storage.logging.*;
    import com.adobe.rice.common.policies.*;
    import com.adobe.rice.common.*;

    //$JDK1.1$ import com.sun.java.util.collections.*;
    import java.util.*;
    import java.lang.Object;

    /**
    * This class is used to keep an arbitrarily long array of bit chunks.
    * It is designed to be reasonably efficient in storage space as well
    * as being able to grow very large.  
    **/

    public class BitVector
    {
    private static final int pageSize = 32;         // we want 32 blocks in a page
    private static final int itemSize = 2;          // we want two bits per item
    private static final int blockSize = 4;         // number of bytes in a block
    private static final int itemsPerBlock = (blockSize * 8 )/itemSize;
    private static final int itemMask = 0x00000003; // item mask 
    private static final int initialPages = 10;     // initial number of pages
    static final int[] masks = 
        {
            0xC0000000,
            0x30000000,
            0x0C000000,
            0x03000000,
            0x00C00000,
            0x00300000,
            0x000C0000,
            0x00030000,
            0x0000C000,
            0x00003000,
            0x00000C00,
            0x00000300,
            0x000000C0,
            0x00000030,
            0x0000000C,
            0x00000003
        };
    static final int[] inverseMasks =
        {
            0x3FFFFFFF,
            0xCFFFFFFF,
            0xF3FFFFFF,
            0xFCFFFFFF,
            0xFF3FFFFF,
            0xFFCFFFFF,
            0xFFF3FFFF,
            0xFFFCFFFF,
            0xFFFF3FFF,
            0xFFFFCFFF,
            0xFFFFF3FF,
            0xFFFFFCFF,
            0xFFFFFF3F,
            0xFFFFFFCF,
            0xFFFFFFF3,
            0xFFFFFFFC
        };
              
    ArrayList pageArray; // implementation is an array of pages.
    
    int capacity;  // the total number of items that can be in this pageArray
                   // capacity = pageArray.length()*pageSize*itemPerBlock 
                   //
    int size;      // the total number of items "in play" in the bit vector
                   // size = ( index > size ? index : size );
                   // size will be reduced by a complex algorithm to be developed later.
    private class Page // a fixed array of integers.
    {
        int[] block;
            
        Page() { block = new int[pageSize]; }
        /**
        *  Method that returns an item out of a page.
        *
        * @param    index   The offset of the item to return.
        */
        int getItem( int index )
        {
            int actualIndex = ( index / itemsPerBlock ) % pageSize;
            int theBlock = block[actualIndex];
            int theItemIndex = index % itemsPerBlock;
            int shiftAmount = itemSize * ( itemsPerBlock - theItemIndex - 1 );
            return ( theBlock >>> shiftAmount ) & itemMask;
        }
        
        /**
        *  Method that sets an item in a page.
        *
        * @param    index   The offset of the item to set.
        * @param    item    The item to set.
        */
        void setItem( int index, int item )
        {
            item &= itemMask;
                
            int actualIndex = ( index / itemsPerBlock ) % pageSize;
            int theBlock = block[actualIndex];
                
            int theItemIndex = index % itemsPerBlock;
                
            int shiftAmount = itemSize * ( itemsPerBlock - theItemIndex - 1 );
                
            item <<= shiftAmount;
            // clean out the slot for the item and or it in.
            theBlock = (( theBlock & inverseMasks[theItemIndex] ) | item );
            // put it back
            block[actualIndex] = theBlock;
                
        }
    }

    public BitVector()
    {
        pageArray = new ArrayList(initialPages);
        size = 0;       // nothing in it now.
        capacity = ( pageArray.size()*this.pageSize*this.itemsPerBlock );
    }
        
    private void makeNewPage( int loc )
    {
        pageArray.add(loc, new Page() );
        capacity = ( pageArray.size()*this.pageSize*this.itemsPerBlock );
    }
               
    /**
    *  Method that sets an item in a page.
    *
    * @param    index   The offset of the item to set.
    * @param    item    The item to set.
    */
    public void setItem( int index, int item )
    {
        if ( index < 0 )
            return;     // we could invert the value and hope, also!
            
        size = ( index > size ) ? index : size;
                
        int pageNumber = index / ( pageSize * blockSize ) ; 
            
        if ( pageNumber >= pageArray.size() )
        {
            // the page does not exist -- so create it
            makeNewPage( pageNumber );
        }
        else if ( pageArray.get(pageNumber) == null )
        {
            // the page does not exist -- so create it
            makeNewPage( pageNumber );
        }
        Page thePage = ( Page ) pageArray.get( pageNumber );
        thePage.setItem(index, item);
    }
        
    /**
    *  Method that returns an item out of a page.
    *
    * @param    index   The offset of the item to return.
    */    
    public int getItem( int index )
    {
        if ( index < 0 )
            return -1;  // this means the maximum size bit string is 31 bits.
            
        int pageNumber = index / ( pageSize * blockSize ) ; 
        if ( pageNumber >= pageArray.size() )
        {
            return -1; // the page does not exist
        }
        else if ( pageArray.get(pageNumber) == null )
        {
            return -1; // the page does not exist
        }
        Page thePage = ( Page ) pageArray.get( pageNumber );
        int the = thePage.getItem(index);
        return the;
    }
    
    public int getSize() { return size; }
    
    public int getCapacity() { return capacity; }
        
}
