/*
$Log: NumericParameter.java,v $
Revision 1.6  2000/11/20 21:26:12  andrewf
some work on changing parameters

Revision 1.5  2000/11/09 19:11:50  andrewf
Corrected errors in EnumerationParameter

Revision 1.4  2000/10/26 00:03:31  andrewf
added documentation and copyright

$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/negotiation/NumericParameter.java,v 1.6 2000/11/20 21:26:12 andrewf Exp $
$Author: andrewf $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org
*/


package com.adobe.rice.common.negotiation;
import java.util.*;

public class NumericParameter extends Parameter
{  
    // NumericSet is the set of acceptable
    // values from which default values must
    // come from
    private NumericSet    range;
    
    // defaultvalues is the set of values
    // which a parameter will accept.
    private Set           defaultValues;
    
    private boolean       hasDefaults;
    
    // constructor for ice-span order = "numeric"
	public NumericParameter (String              name, 
	                         String              id, 
	                         int                 status,
	                         double              minimum,
	                         double              maximum,
	                         double              defaultValue,
	                         double              precision,
	                         boolean             hasDefaults,
	                         NegotiationStrategy policy,
	                         Application         application) 
	                         throws OutOfRange,
	                         InternalError
	{     
	    super (name, id, status, policy, application);
	    
	    TreeSet temp = null;
	    
	    // check if the default value is valid as per section
	    // 4.4.5 of the Ice specification
		if (hasDefaults && !isValidDefault (minimum, maximum, defaultValue))
		{
		   throw new OutOfRange ();
		}
		
	    // okay - now set up the current set
	    setRange (new NumericSpan (minimum, maximum, precision));
    
        // no defaults is simply an empty set
        temp = new TreeSet ();
        
        if (hasDefaults)
        {
	       // bite the bullet, create a Set 
	       temp.add (new Double (defaultValue));
	    }
	    
	    setHasDefaultValues (hasDefaults);
	    setDefaultValues (temp);
	}
   
   
    // constructor for ice-enum order = "numeric"  
    public NumericParameter (String    name, 
	                         String    id, 
	                         int       status,
	                         Set       defaultValues,
	                         Set       rangeValues,
	                         int       selection,
	                         boolean   hasDefaults,
	                         NegotiationStrategy icePolicy,
	                         Application application) 
	                         throws OutOfRange, InternalError
	{
	    super (name, id, status, icePolicy, application);
		  
		// do default values come from the acceptable set
		// of values. Depend on && short circuit
	    if (hasDefaults && !isValidDefault (rangeValues, defaultValues, selection))
	    {
	       throw new OutOfRange ();
        }	    
	   
	    // initialise the IceSet
	    setRange (new NumericEnum (rangeValues, selection));
	    
	    // and default values
	    if (!hasDefaults)
	    {
	       defaultValues = new TreeSet ();
	    }
	    
	    setHasDefaultValues (hasDefaults);
	    setDefaultValues (defaultValues);
	}
     
     
    // Constructor built specially for payloadDisassembler  
    public NumericParameter (String      name, 
	                         String      id, 
	                         int         status,
	                         NumericSet  defaultValues,
	                         IceSet      rangeValues)
	                         throws OutOfRange, InternalError
	{
	    super (name, id, status, null, null);
	} 
     
     
    /**************** Private methods *****************/
     
    // check for numeric span
    private boolean isValidDefault (double minimum, 
                                    double maximum, 
                                    double defaultValue)
    {
       boolean result = false;
       
       // the minimum value must be less than or equal to the max
       result = (minimum <= maximum);
       
       // and the defaultValue must be between the minimum and
       // the maximum
       return result && 
              (minimum <= defaultValue && maximum >= defaultValue);
    }
   
    /*************** protected methods *******************/
    protected void setHasDefaultValues (boolean hasDefaults)
    {
        this.hasDefaults = hasDefaults;
    }
   
    // getDefaultValues and setDefaultValues are required for 
    //accepting IceParameters as per section 4.5.x of ICE specification 
    public Set getDefaultValues ()
    {
        return defaultValues;   
    }

    public void setDefaultValues (Set defaultValues)
    {
        // should deep copy. To avoid clone ()
        TreeSet temp = new TreeSet (defaultValues); 
        this.defaultValues = temp;    
    }
  
    public boolean hasDefaultValues ()
    {
	    return hasDefaults;
    }
  
    /*************** public methods **********************/
    
    public void setRange (IceSet range)
    {
        this.range = (NumericSet) range;
    }
    
    public IceSet getRange ()
    {
        return range;
    }
     
    // hasEquivalentValues determines if two IceParameters
    // have equivalent ice-default-values
    public boolean hasEquivalentValues (Parameter sender)
    {
        return (getDefaultValues ().equals (sender.getDefaultValues ()));
    }
    
    // are sender's default values within receiver's range 
    // of values?
    public boolean in (Parameter sender)
    {
        return getRange ().in (sender.getDefaultValues ());
    }
 
    // return the size of the range set
    public int rangeSize ()
    {
        return getRange ().rangeSize ();
    }
 
    // return the size of the defaultValues
    public int defaultValueSize ()
    {
        return getDefaultValues ().size ();   
    }
    
    /******************* API methods *****************/
    
    public boolean in (double value)
    {
        TreeSet list;
        // wrap the value in set
        list = new TreeSet ();
        list.add (new Double (value));
        return getRange ().in (list);
    }
    
    public boolean in (Set values)
    {
        return getRange ().in (values);
    }
    
    // what is minimum range value
    public double minimum ()
    {
        return (((Double)((NumericSet) getRange ()). minimum ())).doubleValue ();
    }
    
    // what is maximum range value
    public double maximum ()
    {
        return (((Double)((NumericSet) getRange ()).maximum ())).doubleValue ();
    }
    
    public double previous (double value)
    {
        return (((Double)((NumericSet) getRange ()).previous (value))).doubleValue ();
    }
    
    public double next (double value)
    {
        return (((Double)((NumericSet) getRange ()).next (value))).doubleValue ();
    }
    
    public String toString ()
    {
        return super.toString () + " range: " + getRange ().toString ()
               + " defaults :" + getDefaultValues ();
    }
}
 
 