/*
$Log: Offer.java,v $
Revision 1.3  2000/11/22 00:46:33  andrewf
Changed IceNegotiable to Negotiable

Revision 1.2  2000/10/26 00:03:31  andrewf
added documentation and copyright

$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/negotiation/Offer.java,v 1.3 2000/11/22 00:46:33 andrewf Exp $
$Author: andrewf $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org
*/


package com.adobe.rice.common.negotiation;
import java.util.*;

public abstract class Offer extends IceNegotiable 
                            implements Offerable
{  
    private Resolver       resolver;
    private OfferStrategy  offerStrategy;
    private boolean        hasFocus;
    private Parameter      focus;
    private int            impasse;
    private Application    application;
	 
	public Offer (String name, 
	              String id, 
	              Collection     parameterList,
	              OfferStrategy  strategy,
	              Application application)
	{
	    //initially we do not know the status of the
	    //offer - it is determined by the resolver
		super (name, id, IceNegotiable.UNKNOWN);
		
	    resolver = new Resolver  (parameterList);
	    setStrategy (strategy);
	    setStatus (resolver.getStatus ());
	    
	    // initially offers do not have a focus
	    setFocus (null);
	    
	    // impasse experimental - lets not use getter/setter
	    impasse = 0;
	    
	    this.application = application;
	}
	
	private void setFocus (Parameter newFocus)
	{
	    focus = newFocus;
	    hasFocus = (focus == null) ? false : true;
	}
	
	public Parameter getFocus ()
	{
	   return focus;
	}
	
	// private section
	
	 // isValid () is a utility method used to enforce
    // parameter negotiation rules in 4.4.5 of ICE 1.1
    // specification
    public final boolean isValid (Negotiable sender,
                                  Negotiable receiver)
    {
        boolean   result;
        Parameter r = (Parameter) receiver;
        Parameter s = (Parameter) sender;
        
        //does the sender have a default?
        result = s.hasDefaultValues ();
         
        return result && r.isValid (s);
    }
	
	private final boolean timeExpired ()
	{
	    boolean result = false;
	    
	    return result;   
	}
	
	private final boolean roundsExceeded ()
	{
	    boolean result = false;
	    
	    return result;
	}
	
	private final boolean impasseDetected (Offer sender)
	{
	    boolean result = false;
	    boolean changed = false;
	    
	    // do we already have a focus?
	    if (hasFocus () && getFocus ().equals (sender.getFocus ()))
	    {
	       if (impasse >= Negotiable.IMPASSE_LIMIT)
	       {
	          result = true;
	       }
	       else
	       {
	          ++impasse;
	       }
	    }
	    else
	    {
	       // reaching here means one of the following:
	       // we did not have a focus - new offer OR
	       // we are negotiating a different parameter 
	       impasse = 0;
	       setFocus (getFocus ());
	    }
	    
	    return result;
	}
	
	// for now make public
	public boolean hasFocus ()
	{
	    return focus == null ? false : true;
	}

	private OfferStrategy getStrategy ()
	{
	    return offerStrategy;
	}
	
	private void setStrategy (OfferStrategy newStrategy)
	{
	    offerStrategy = newStrategy;
	}
	
	private void pushUnresolved (Negotiable element)
	{
	    resolver.PushUnresolved (element);    
	}
	
    private boolean hasResolvedNegotiables ()
    {
        
        return resolver.hasResolvedNegotiables();   
    }
    
    private Negotiable nextResolvedNegotiable ()
    {
        return resolver.nextResolvedNegotiable();   
    }
    
    private Negotiable getResolvedNegotiable (String name)
    {
        return resolver.getResolvedNegotiable(name);   
    }
    
    private boolean hasUnresolvedNegotiables ()
    {
        
        return resolver.hasUnresolvedNegotiables();   
    }
    
    private Negotiable nextUnresolvedNegotiable ()
    {
        return resolver.nextUnresolvedNegotiable();   
    }
    
    private Negotiable getUnresolvedNegotiable (String name)
    {
        return resolver.getUnresolvedNegotiable(name);   
    }
    
    private Negotiable removeUnresolvedNegotiable (String name)
    {
        return resolver.removeUnresolvedNegotiable(name);
    }
    
    private void addResolvedNegotiable (Negotiable receiver)
    {
        resolver.addResolvedNegotiable(receiver);        
    }
    
    private Negotiable getNegotiable (String name)
    {
        return resolver.getNegotiable(name);   
    }
    
    private boolean addToResolvedNegotiables (Negotiable element)
    {
        return resolver.addResolvedNegotiable(element);     
    }
  
    private void initialiseResolved ()
    {
        resolver.initialiseResolvedList ();
    }
    
    // Proceed calls a user defined pre-condition test
    protected boolean proceed (Negotiable sender,
                               Negotiable receiver)
    {
        return getStrategy ().proceed (sender, receiver);
    }
    
	// canProceed implements protocol rules
	// subclasses are not allowed to override it
	// when work is finished, this will be made final
	protected boolean canProceed (Offer sender)
	{
	    boolean result = true;
	    
	    do
	    {
	       //first check if time has expired
	       if (timeExpired ())
	       {
	          result = false;
	          break;
	       }
	       
	       // check if permitted round count has been exceeded
	       if (roundsExceeded ())
	       {
	          result = false;
	          break;
	       }
	    
	       // check if an impasse has been detected
	       if (impasseDetected (sender))
	       { 
	          result = false;
	          break;
	       }
	    
	       // does the user wish to proceed?
	       if (!proceed (sender, this))
	       {
	          result = false;
	          break;
	       }
	    } while (false);
	    
	    return result;   
	}
	
	protected Offer resolveImpasse (Negotiable sender)
	                                  throws InternalError
	{
	    return getStrategy ().resolveImpasse (sender, this);    
	}
	                                                      
	protected Offer ratify (Offer sender)
	                   throws InternalError
	{
	    return (Offer) getStrategy ().ratify (sender, this);
	}

    public void accept (Negotiable sender)
    {
	    setStatus (IceNegotiable.ACCEPT);
	    setErrorStatus (IceNegotiable.OK);
    }

    // when a parameter is accepted - it taken off
    // the unresolved list and placed on the resolved list
    public void accepted(Negotiable negotiable)
    {
        Negotiable receiver;
      
        receiver = removeUnresolvedNegotiable (negotiable.getName());
        receiver.accepted ();
        addResolvedNegotiable (receiver);  
        setFocus (null);
    }

    public boolean isAcceptable (Negotiable sender)
    {
        return getStrategy ().isAcceptable (sender, this);           
    }
    
    // this sets the focus as per section 
    // August 15th, 2000 - set the error status to
    // counter-proposal
    
    public void negotiating (Negotiable negotiable)
    {
        setErrorStatus (Negotiable.COUNTER_PROPOSAL); 
        setStatus (Negotiable.NEGOTIATE);
        
        // August 17th, 2000 
        // and make it the focus
        setFocus ((Parameter) negotiable);
    }
    
    public Negotiable negotiate (Negotiable sender) throws
                                    InternalError,
                                    OutOfRange,
                                    Sorry
    {
        Offer        reply = null;
        Offer        senderOffer;
        Negotiable   r;
        Negotiable   s;
        Negotiable   counterNegotiable;
        
        // cast to a form that is usable
        senderOffer = (Offer) sender;
        
        // August 16th - September 18
        // the offer is sending back a serialisation of its state
        // so this is typically correct
        reply = this;
       
        // August 14th, 2000 change
        // cover the case the sender accepted 
        // revisit later and re-structure code
        // for single exit point
        if (sender.isAccepted ())
        {
            // resolve outstanding parameters
            resolve (senderOffer);
            return ratify (senderOffer);
        }
         
        // match sender's resolved parameters against
        // receiver's unresolved parameters
        if (!resolve (senderOffer))
        {
           return this;           
        }
        
        // if there are no more unresolved receiver
        // parameters, then we are done
        if (isResolved ())
        {
           reply = ratify (senderOffer);     
        }
        else
        {
           // can negotiations proceed?
           if (!canProceed (senderOffer))
           {
              // must resolve impasse according to 
              // section 4.4.6.2 rules
              return resolveImpasse (senderOffer);
           }
           
           // try to accept as many parameters as possible
           // in the order which the sender transmitted them
           while (senderOffer.hasUnresolvedNegotiables ())
           {
               s = senderOffer.nextUnresolvedNegotiable (); 
               r =  getUnresolvedNegotiable(s.getName ());
               
               // a null can represent two conditions:
               // illegal - a parameter is renegotiated
               // illegal - a new parameter has been introduced
               
               if (r == null)
               {
                  Negotiable temp = getNegotiable (s.getName ());
                  
                  if (temp == null)
                  {
                     // illegal introduction
                     this.reject (Negotiable.RENEGOTIATION);
                     break;
                  }
                  else if (temp.isAccepted ())
                  {
                     // illegal re-negotiation
                     this.reject (Negotiable.RENEGOTIATION);
                     break;
                  }
                  else
                  { 
                     // there should not be any other conditions
                     // for which r is null
                     throw new InternalError ();
                  }
               }
               
               // check validity of parameters
               if (!isValid (s, r))
               {
                  // september 19th correction
                  this.reject (Negotiable.INVALID); 
                  break;
               }
               
               if (r.isAcceptable (s))
               {
                  r.accept (s);
                  accepted (r);
               }
               else
               {
                  // this becomes the focus, compute a counterOffer
                  counterNegotiable = r.negotiate (s);
                  if (counterNegotiable.isRejected ())
                  {
                     reply.reject (counterNegotiable.getErrorStatus ()); 
                  }
                  else
                  {
                     negotiating (counterNegotiable); 
                     
                     //for debugging purposes, push back r
                     ((Offer)sender).pushUnresolved (s);
                  }
                  break;
               }
           }
           
           // if all parameters are resolved, then a decision
           // should be made concerning accepting offer
           if (isResolved ())
           {
              //ratifying logic can overturn at this stage
              reply = ratify (senderOffer);
           }
           // if the reply was not rejected
           else if (!reply.isRejected ())
           {
              reorder (senderOffer);
              reply = ratify (senderOffer);
           }
        }   
        
        return reply;
    }
    
    
    public Negotiable propose (Negotiable sender)
    {
        return getStrategy ().propose (sender, this);
    }
    
    //
    //  August 17th, 2000
    //  Reorder - section 4.4.6.3
    //
    public void reorder (Offer sender)
    {
        Negotiable element = getFocus ();
        
        // to avoid double counting remove the focus 
        //from the unresolved list
        resolver.removeUnresolvedNegotiable (element.getName ());
        
        // and add it to the front
        resolver.PushUnresolved(element);
        
        // and call getPolicy's reorder
        getStrategy ().reorder (this);
    }
    
    //public Negotiable consult (Negotiable sender)
    //{
    //    return null; 
    //}
     
    public boolean resolve (Offer senderOffer)
    {
       boolean result = true;
       Negotiable sender;
       Negotiable receiver;
        
       // november 21, re-visit 
   
       try
       {
         
       // a kludge
       senderOffer.initialiseResolved ();
       
       while (senderOffer.hasResolvedNegotiables ())
       {
          sender = senderOffer.nextResolvedNegotiable();
          
          //september 23 correction
          //check to see if negotiable is already accepted
          receiver = getNegotiable (sender.getName ());
               
          if (receiver != null)
          {
             if (receiver.isAccepted ())
             {
                // System.out.println (getName () + " resolved " + receiver);
                continue;
             }
          }
          
          // October 17th change - are we still willing to
          // honour this parameter's value
          // November 22 add try block
         
          if (!receiver.isAcceptable (sender))
          {
             this.reject (Negotiable.SORRY);
             return false;
          }
          
          // receiver = removeUnresolvedNegotiable (sender.getName());
          removeUnresolvedNegotiable (sender.getName());
          
          if (receiver != null)
          {
             receiver.accepted ();
             addResolvedNegotiable (receiver);
             
          // August 18th, 2000 change - if there was a focus
          // it would have reset
          
             setFocus (null);
          }
       }  
       
       }
       catch (OutOfRange e)
       {
         this.reject (Negotiable.SORRY);
         result = false;
       }
       
       return result;
    }
 
    // may be promoted to Negotiable class
    public boolean isResolved ()
    {
        return hasUnresolvedNegotiables () ? false : true;  
    }
 
    // utility function
    public String toString ()
    {
        return super.toString () + " " + getErrorStatus () +
               " " + resolver.toString ();
    }
}