/*
$Log: TimeParameter.java,v $
Revision 1.3  2000/10/26 00:03:32  andrewf
added documentation and copyright

$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/negotiation/TimeParameter.java,v 1.3 2000/10/26 00:03:32 andrewf Exp $
$Author: andrewf $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org
*/


package com.adobe.rice.common.negotiation;
import java.util.*;

public class TimeParameter extends Parameter
{  
    // TimeSet is the set of acceptable
    // values from which default values must
    // come from
    private TimeSet    range;
    
    // defaultvalues is the set of values
    // which a parameter will accept.
    private Set           defaultValues;
     
    private boolean       hasDefaults;
    
    // constructor for ice-span order = "time"
	public TimeParameter (String    name, 
	                      String    id, 
	                      int       status,
	                      Date      minimum,
	                      Date      maximum,
	                      Date      defaultValue,
	                      long      precision,
	                      boolean   hasDefaults,
	                      NegotiationStrategy policy,
	                      Application application) 
	                      throws OutOfRange, InternalError
	{     
	    super (name, id, status, policy, application);
	    
	    TreeSet temp;
	    
	    // check if the default value is valid as per section
	    // 4.4.5 of the Ice specification
		if (hasDefaults && !isValidDefault (minimum, maximum, defaultValue))
		{
		   throw new OutOfRange ();
		}
		
	    // okay - now set up the current set
	    setRange (new TimeSpan (minimum, maximum, precision));
    
	    // bite the bullet, create a Set 
	    temp = new TreeSet ();
	    if (hasDefaults)
	    {
	       temp.add (defaultValue);
	    }
	    
	    setHasDefaultValues (hasDefaults);
	    setDefaultValues (temp);
	}
   
   
    // constructor for ice-enum order = "time"  
    public TimeParameter (String    name, 
	                      String    id, 
	                      int       status,
	                      Set       defaultValues,
	                      Set       rangeValues,
	                      int       selection,
	                      boolean   hasDefault,
	                      NegotiationStrategy icePolicy,
	                      Application application) 
	                      throws OutOfRange, InternalError
	{
	    super (name, id, status, icePolicy, application);
		  
		// do default values come from the acceptable set
		// of values
	    if (hasDefaults && !isValidDefault (rangeValues, defaultValues, selection))
	    {
	       throw new OutOfRange ();
        }	    
	   
	    // initialise the IceSet
	    setRange (new TimeEnum (defaultValues, selection));
	    
	    // and default values
	    if (!hasDefaults)
	    {
	       // just has an empty set
	       defaultValues = new TreeSet ();
	    }
	    
	    setDefaultValues (defaultValues);
	    setHasDefaultValues (hasDefaults);
	}
      
    /**************** Private methods *****************/
    
    // check for numeric span
    private boolean isValidDefault (Date minimum, 
                                    Date maximum, 
                                    Date defaultValue)
    {
       boolean result = false;
       
       // the minimum value must be less than or equal to the max
       result = minimum.compareTo (maximum) <= 0;
       
       // and the defaultValue must be between the minimum and
       // the maximum
       return result && (minimum.compareTo (defaultValue) <= 0 
                         && maximum.compareTo (defaultValue) >= 0);
    }
   
    /*************** protected methods *******************/
    protected void setHasDefaultValues (boolean hasDefaults)
    {
        this.hasDefaults = hasDefaults;
    }
    
    public boolean hasDefaultValues ()
    {
        return hasDefaults;
    }
    
    // getDefaultValues and setDefaultValues are required for 
    //accepting IceParameters as per section 4.5.x of ICE specification
   
    public Set getDefaultValues ()
    {
        return defaultValues;   
    }

    public void setDefaultValues (Set defaultValues)
    {
        // perhaps I should change this to a clone?
        defaultValues = defaultValues;    
    }
  
    /*************** public methods **********************/
    
    public void setRange (IceSet range)
    {
        this.range = (TimeSet) range;
    }
    
    public IceSet getRange ()
    {
        return range;
    }
     
    // hasEquivalentValues determines if two IceParameters
    // have equivalent ice-default-values
    public boolean hasEquivalentValues (Parameter sender)
    {
        return (getDefaultValues ().equals (sender.getDefaultValues ()));
    }
    
    // are sender's default values within receiver's range 
    // of values?
    public boolean in (Parameter sender)
    {
        return getRange ().in (sender.getDefaultValues ());
    }
 
    // return the size of the range set
    public int rangeSize ()
    {
        return getRange ().rangeSize ();
    }
 
    // return the size of the defaultValues
    public int defaultValueSize ()
    {
        return getDefaultValues ().size ();   
    }
    
    /******************* API methods *****************/
    
    public boolean in (Date value)
    {
        TreeSet list;
        // wrap the value in set
        list = new TreeSet ();
        list.add (value);
        return getRange ().in (list);
    }
    
    public boolean in (Set values)
    {
        return getRange ().in (values);
    }
    
    // what is minimum range value
    public Date minimum ()
    {
        return ((TimeSet) getRange ()). minimum ();
    }
    
    // what is maximum range value
    public Date maximum ()
    {
        return ((TimeSet) getRange ()).maximum ();
    }
    
    public Date previous (Date value)
    {
        return ((TimeSet) getRange ()).previous (value);
    }
    
    public Date next (Date value)
    {
        return ((TimeSet) getRange ()).next (value);
    }
}
 
 