/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/payloadAssembly/PayloadFactory.java,v 1.5 2000/12/02 23:24:23 bhunt Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

*/

package com.adobe.rice.common.payloadAssembly;
//
// java imports
//
import java.io.*;
import java.util.*;

//
// icecubes  imports
//
import com.adobe.rice.IceCubes;
import com.adobe.rice.util.*;
import com.adobe.rice.util.time.*;
import com.adobe.rice.common.*;
import com.adobe.rice.common.XMLassembly.*;
import com.adobe.rice.common.policies.*;
import com.adobe.rice.common.payloadAssembly.*;
import com.adobe.rice.common.storage.logging.*;     // so we can write to the log.

public class PayloadFactory
{
    //
    // The payload factory class is a singleton that keeps track of all of the 
    // variants of ICE that are currently active in this instance of ICE cubes. 
    //
    // A brief explanation follows:
    //
    // Each ICE version is permitted to have variants.  An ICE variant is an extension
    // of the ICE DTD that is agreed upon by two communicating parties.  In ICE versions
    // later than 1.01, extensions are negotiated between parties.  To implement this,
    // icecubes requires that each variant be named.  Since the cost of compiling a
    // DTD is fairly high and since a compiled DTD supports any number of documents,
    // icecubes keeps each variant DTD once it has compiled it.  The variants are
    // named and stored as follows.  A table variants is kept for each version of 
    // ICE that is active.  That is, the value in the version table at each version key
    // is a table of variants.  The value for each variant key in a specific version's 
    // variant table is a "compiled DTD".  This mechanism makes it extremely easy to keep
    // a (growing) list of compiled DTDs, one for each variant in a version. 
    //
    // This class keeps, indefinitely, all compiled variants for known versions 
    // (i.e. those that are defined in the Constants class.  These are generally
    // the versions that have been released by the Ice authoring group.
    // However, it is permitted and possible to add a new version, and with it new variants to
    // that version.  However, when all variants for one of these dynamically 
    // added versions are deleted, the version is also deleted.
    //
    // This provides the foundation for limiting the number of compiled dtd's
    // consuming memory but not being used.  A future addition could be to track
    // use of a specific variant compiled DTD and delete it if it is not being used.
    // 
    
    // the (dynamic) ICE version table
    private static HashMap iceVersions = new HashMap();
    //
    // these are the defaults.  Without them, IceCubes cannot run.
    //
    private static DocType defaultIceDTD = null;
    private static String  defaultVariant = Constants.fixedIceDTDvariant;
    private static String  defaultVersion = Constants.fixedIceDTDversion;
    //
    // This is the current DTD that will be used to create new payloads if
    // no version and variant are specified in a get.  It is usually the
    // default DTD above (but not necessarily!)
    //
    private static DocType iceDTD = null;
    //
    // Implement the Singleton pattern
    //
    private static PayloadFactory payloadFactory = null;
    
    //
    // The following is a small class to manage compiled DTDs for payload factory
    //
    
    private class CompiledDTD
    {
        // The internal name and version of the DTD:
        String name;                // the name of this DTD variant
        String version;             // the version of this DTD variant.
        // How to find it -- This is (possibly) a uri
        String fileName;            // the file name of this DTD
        String filePath;            // the file path of this DTD
        // The actual DTD
        DocType dtd;                // the compiled DTD -- ready for use.
        
        CompiledDTD()
        {
            version = Constants.fixedIceDTDversion;
            fileName = Constants.fixedIceDTDname;
            filePath = Constants.fixedIceDTDpath;
            // neither the dtd nor the variant name is set in this constructor.
        }
        
        CompiledDTD( String ver, String nm, String fPath, String fName )
        {
            String theFullPath;
            
            name = nm;
            version = ver;
            fileName = fName;
            filePath = fPath;
            
            // now set up the DTD:
            theFullPath = getIceDTDfullPath(filePath, fileName); 
            //
            // $FIX$ the issue is how should we recover from one of the possible errors?
            // Probably the best thing to do is to leave the DTD null as a 
            // signal (to the unsuspecting?) that this variant cannot be used.
            // $POLICY$ -- maybe recovery should be a matter of policy?
            try { dtd = setupDTD( theFullPath ); }
            catch ( DocTypeException dte ) { }
            catch ( XMLparseException xpe ) { }
            catch ( IOException ioe ) { }
            
        }
        
        private DocType setupDTD( String fullPath ) 
            throws DocTypeException, XMLparseException, IOException
        {
            DocType dtd = null;
            try { dtd = new DocType(fullPath); }
            catch ( DocTypeException dte ) 
            { 
                Object parms[] = {"ICE DTD", fullPath, dte.getMessage() };
                AssemblyMessages.log(
                    AssemblyMessages.get("invalid"), 
                    parms, ImpactRange.FATAL );
                throw dte; 
            }
            catch ( XMLparseException xpe ) 
            { 
                Object parms[] = {"ICE DTD", fullPath, xpe.getMessage() };
                AssemblyMessages.log(
                    AssemblyMessages.get("invalid"), 
                    parms, ImpactRange.FATAL );
                throw xpe; 
            }
            catch ( IOException ioe ) 
            { 
                Object parms[] = {"ICE DTD", fullPath, ioe.getMessage() };
                AssemblyMessages.log(
                    AssemblyMessages.get("invalid"), 
                    parms, ImpactRange.FATAL );
                throw ioe; 
            }
            return dtd;
        }
        
        void setName( String nm ) { name = nm; }
        void setVersion( String v ) { version = v; };
        void setFileName( String fn ) { fileName = fn; }
        void setFilePath( String path ) { filePath = path; }
        void setDTD( DocType theDTD ) { dtd = theDTD; }
        
        String getName() { return name; }
        String getFileName() { return fileName; }
        String getFilePath() { return filePath; }
        String getVersion() { return version; }
        DocType getDTD() { return dtd; }
    }
    
    //
    // During initialization, PayloadFactory compiles the
    // default ICE dtd first from policy declaration; or from the builtin 
    // default path.  All other variants are added "on the fly" as they are
    // negotiated or otherwise added (How this happens for an 1.0 or 1.01 dtd 
    // is a matter of local instance policy.)
    // 
    // Presumably, when an extension is negotiated, the semantics of the
    // extension will call the DTD compilation and setup routines here.
    //
    
    private PayloadFactory() throws DocTypeException
    {
        // setup the table of variant tables:
        for ( int i = 0; i < Constants.versions.length; i++ )
        {
            HashMap variants = new HashMap();
            this.iceVersions.put( Constants.versions[i], variants );
        }
        // Must construct the standard payload variant
        String iceDTDpath = this.getIceDTDpath();
            
        DocType stdDTD = this.ConstructPayloadVariant(Constants.fixedIceDTDversion,
                                        Constants.fixedIceDTDvariant,
                                        iceDTDpath,
                                        Constants.fixedIceDTDname );
        if ( stdDTD == null )
        {
            //log the failure and collapse!
            // This will happen very early in initialization.
            Object parms[] = 
            {   Constants.fixedIceDTDvariant, 
                Constants.fixedIceDTDversion, 
                AssemblyMessages.get("noStandardIceDTD") + 
                    iceDTDpath + "/" + Constants.fixedIceDTDname
            };
            AssemblyMessages.log(
                AssemblyMessages.get("dtdCompileFailed"), 
                parms, ImpactRange.FATAL );
            DocTypeException dte = new DocTypeException("Cannot Run without a DTD!");
            throw dte;
        }
        else
        {
            // setup the default DTD
            DocType newDTD = setDefaultPayloadDTD( Constants.fixedIceDTDversion, 
                                                   Constants.fixedIceDTDvariant );
            // now the DTD we got back had better be identical by reference pointer
            // to the stdDTD.
            if ( stdDTD != newDTD )
            {
                Object parms[] = 
                {   Constants.fixedIceDTDvariant, 
                    Constants.fixedIceDTDversion, 
                    AssemblyMessages.get("StdIceDTDSetupError") + 
                        iceDTDpath + "/" + Constants.fixedIceDTDname
                };
                AssemblyMessages.log(
                    AssemblyMessages.get("dtdCompileFailed"), 
                    parms, ImpactRange.FATAL+1 );
                DocTypeException dte = new DocTypeException("Cannot Run without a DTD!");
                throw dte;
            }
            this.iceDTD = stdDTD;
        }
                              
    }
    
    static public PayloadFactory instance()
    {
        if ( payloadFactory == null )
        {
            try { payloadFactory = new PayloadFactory(); }
            catch ( DocTypeException dte )
            {
                // $FIX$ output the exeception message and log it if possible
                System.out.println(" Payload Factory Failure -- ICE DTD Exception!");
                System.out.println( dte.getMessage() );
                System.exit(-8002);
            }
        }
        return payloadFactory;
    }
    
    //
    // Now build a Payload for the caller.
    //
    public PayloadDoc get() 
    {
        // now construct the prototype XML document.
        
        PayloadDoc payloadDoc = null;
        try
        {
            payloadDoc = new PayloadDoc( defaultIceDTD );
        }
        catch ( ElementException ele )
        {
            //$LOG$ log the failure and go on.
            Object parms[] = { this.defaultVariant, this.defaultVersion, ele.getMessage() };
            AssemblyMessages.log(
                AssemblyMessages.get("PayloadDocMakeFailed"), 
                parms, ImpactRange.FATAL );
        }
        catch ( XMLdocException xde )
        {
            //$LOG$ log the failure and go on.
            Object parms[] = { this.defaultVariant, this.defaultVersion, xde.getMessage() };
            AssemblyMessages.log(
                AssemblyMessages.get("PayloadDocMakeFailed"), 
                parms, ImpactRange.FATAL );
        }
        return payloadDoc;
            
    }
    
    public PayloadDoc get( String version )
    {
        // this routine is equivalent to:
        // this.get( version, defaultVariant );
        String theVersion = this.defaultVersion;
        if ( ( version != null) && ( version.length() > 0 ) )
            theVersion = version;
        return this.get( theVersion, defaultVariant );
    }
    
    public PayloadDoc get( String version, String variant )
    {
        PayloadDoc theDoc = null;
        HashMap variants;
        if ( this.iceVersions.containsKey(version) )
        {
            // get the variants for this version:
            variants = (HashMap) this.iceVersions.get(version);
            if ( variants.containsKey( variant ) )
            {
                CompiledDTD cVdtd = (CompiledDTD) variants.get( variant );
                if ( cVdtd.getDTD() != null )
                {
                    try { theDoc = new PayloadDoc( cVdtd.getDTD() ); }
                    catch ( XMLdocException xde ) 
                    { 
                        //$LOG$ log the failure and go on.
                        Object parms[] = {cVdtd.getName(), cVdtd.getVersion(), xde.getMessage() };
                        AssemblyMessages.log(
                            AssemblyMessages.get("PayloadDocMakeFailed"), 
                            parms, ImpactRange.FATAL );
                    }
                    catch ( ElementException ee )
                    {
                        //$LOG$ log the failure and go on
                        Object parms[] = {cVdtd.getName(), cVdtd.getVersion(), ee.getMessage() };
                        AssemblyMessages.log(
                            AssemblyMessages.get("PayloadDocMakeFailed"), 
                            parms, ImpactRange.FATAL );
                    }
                }
            }
        }
        return theDoc;
    }
    
    /**
    * This routine only sets the defaultPayloadDTD if it exists.  If it does
    * exist, the new dtd is returned and the defaults are set.  If it does not
    * exist, then null is returned and the DTD is not changed.
    * 
    **/
    public synchronized DocType setDefaultPayloadDTD( String version, String variant )
    {
        DocType newDTD = null;
        
        if ( this.iceVersions.containsKey(version) )
        {
            HashMap ht = (HashMap) this.iceVersions.get( version );
            if ( ht.containsKey( variant ) )
            {
                CompiledDTD cDTD = ( CompiledDTD) ht.get( variant );
                if ( cDTD != null )
                    newDTD = cDTD.getDTD();
                if ( newDTD != null )
                {
                    this.defaultIceDTD = newDTD;
                    this.defaultVariant = variant;
                    this.defaultVersion = version;
                }
            }
        }
        return newDTD;
    }
    /**
    * This routine only sets the defaultPayloadDTD if it exists.  If it does
    * exist, the new dtd is returned and the defaults are set.  If it does not
    * exist, then null is returned and the DTD is not changed.
    * 
    **/
    public DocType setDefaultPayloadDTD( String variant )
    {
        return this.setDefaultPayloadDTD(this.defaultVersion, variant);
    }
    
    //
    // Utility Routines
    //
    /**
     * Construct a payload variant.  This method is called to ready a new
     * payload document variant for use by iceCubes.  The new variant is 
     * readied by compiling a new DTD for the type and entering the new
     * DTD in the table of variants for the particular version.
     * If the version does NOT exist, a new variant hashtable is created to
     * hold the variants of this version and the variant hashtable and the new version
     * are entered into the version hashtable.
     *
     * If the variant already exists, it is replaced by the new variant!
     *
     * Note, Only one updater at a time on the table of tables is allowed.
     **/
    public synchronized DocType ConstructPayloadVariant( String version, String variant, String filePath, String fileName )
    {
        DocType newDTD;
        HashMap ht;
        
        CompiledDTD cDTD = new CompiledDTD( version, variant, filePath, fileName );
        CompiledDTD oldCompiledDTD = null;
        
        newDTD = cDTD.getDTD();
        if ( newDTD != null )
        {   // There is something to add.
            // Make sure that we have a hashmap for variants of this version.
            newDTD.setName("ice-payload");
            if ( !this.iceVersions.containsKey(version) )
            {
                ht = new HashMap();
                this.iceVersions.put(version, ht);
            }
            else
                ht = (HashMap) this.iceVersions.get( version );
            // we now have a HashMap for variants of this version.
            // write in the new DTD.
            oldCompiledDTD = (CompiledDTD) ht.put( variant, cDTD );
            if ( oldCompiledDTD != null )
            {
                //$LOG$ - better log overwrites!
                Object parms[] = { version, variant, filePath + "/" + fileName };
                AssemblyMessages.log(
                    AssemblyMessages.get("VariantReplaced"), 
                    parms, ImpactRange.WARNING );
            }
        }
        return newDTD;
    }
    
    public synchronized DocType DeletePayloadVariant( String version, String variant )
    {
        DocType oldDTD = null;
        
        if ( this.iceVersions.containsKey(version) )
        {   // get the version's variant hashmap:
            HashMap hm = (HashMap) this.iceVersions.get(version);
            CompiledDTD cDTD = (CompiledDTD) hm.get( variant );
            if ( cDTD != null )
            {
                oldDTD = cDTD.getDTD();
                hm.remove(variant);
            }
            // if this version is not a well known version (ie, - in our table of
            // standard ice versions) then if we deleted all variants, we dump the
            // new version as well.
            if ( !checkVersion( version ) )
            {
                if ( hm.isEmpty() )
                    this.iceVersions.remove( version );
            }
        }
        return oldDTD;
    }
    /**
     * checkVersion - make sure that this is a correct version
     *
     * @param ver - String containing the suspected version to check.
     *
     * @return boolean - true if the version is valid.
     *
     **/
     public boolean checkVersion( String ver )
     {
        for ( int i = 0; i < Constants.versions.length; i++ )
        {
            if ( ver.equals(Constants.versions[i]) )
                return true;
        }
        return false;
     }
    
    /**
     * getIceDTDpath - return the String to the ice path.
     *  This is either a URL or a local file name.
     *
     **/
     
    private String getIceDTDpath( )
    {
        // To do this, we have to get the path from the policies.
        // There are two sources for the ICE dtd:
        // -- a URL; where it constitutes the external subset; 
        // -- or from a local file.
            
        String iceDTDpath = Constants.fixedIceDTDpath;
                    
        String loc = (String) Tray.getPolicyManager().getSafeSelection("IceDTDlocation", "builtin");
            
        if ( loc.equals("local") )
        {
            String path = (String) Tray.getPolicyManager().getSelection("LocalDTDlocation");
            if ( path != null )
            {
                // The following seems to be to restrictive -- shouldn't this be done
                // elsewhere?  This code is attempting to figure out if the local location
                // is a relative file path -- starting from the default location -- namely
                // the current directory.
                if (  ( path.length() > 0 ) &&
                      ( !path.startsWith("/") ) &&
                      ( !path.startsWith(".") ) &&
                      ( !path.startsWith("http:") ) &&
                      ( !path.startsWith("https:") ) &&
                      ( !path.startsWith("file:") ) &&
                      ( !path.startsWith("ftp:") ) 
                   )
                    path = "./"+path;
                iceDTDpath = path;
            }
        }
        else if ( loc.equals("remote") )
        {
            String path = (String) Tray.getPolicyManager().getSelection("RemoteDTDlocation");
            if ( path != null )
                iceDTDpath = path;
        }
        return iceDTDpath;
    }
    
    private String getIceDTDfullPath( String fPath, String fName )
    {
        String path;
        String name = Constants.fixedIceDTDname;
        
        if ( ( fPath != null ) && ( fPath.length() > 0 ) )
            path = fPath;
        else
            path = getIceDTDpath();
        if ( ( fName != null ) && ( fName.length() > 0 ) )
            name = fName;
        return path + "/" + name;        
    }
}