/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/policies/PolicyManager.java,v 1.5 2000/11/22 21:16:16 bhunt Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

*/

/**
 * PolicyManger.java
 *
 *
 * Created: Sun Aug 21 01:10:41 1999
 *
 * @author Will Conant
 * @author V. Bruce Hunt 
 * @version 0.1
 *
 * @update vbh - implemented Manager interface
 * @update vbh - added package support
 *
 * The policy manager implements a "bilevel" store strategy
 *
 * Each policy is identified by at most two names, package.name
 *
 * where package -- is the package name to which the policy applies; 
 * while name -- is the name of the specific policy.
 *
 * The special package name, "*" is used for all global policies.
 *
 */
package com.adobe.rice.common.policies;

//$JDK1.1$ import com.sun.java.util.collections.*;

import java.util.*;
import java.io.*;
//
// ICEcubes imports:
//
import com.adobe.rice.IceCubes;
import com.adobe.rice.util.*;
import com.adobe.rice.common.*;
import com.adobe.rice.common.catalog.*;
import com.adobe.rice.common.storage.*;
import com.adobe.rice.common.storage.logging.*; 
import com.adobe.rice.common.scheduling.*;
import com.adobe.rice.common.transportation.*;
import com.adobe.rice.common.XMLdisassembly.*;
import com.adobe.rice.common.policies.configure.*; 
import com.adobe.rice.kernel.*;

public class PolicyManager extends IceCubeManager
{
    //
    // Static class variables
    //
    
    // there is only ONE policyManager
    private static PolicyManager policyManager;
    
    //
    // the object that contains and manages the policies the define the managers
    //
    private static ManagerPolicies managerPolicies;
	
	//
	// private instance variables
	//
    private HashMap packageMap;  // This maps package names to policy hash maps.
	
	//
	// initializer -- implement the singleton pattern
	//
    private PolicyManager() throws PolicyException, DispatcherException
    {
        super ("PolicyManager");
        
        //
        // first -- set ourselves up
        //
        PolicyManager.policyManager = this;
        // following is minimum apparatus to both add and access policies:
        // Thus, by the time this initializes, we have a small PolicyManager
        // that is runnable.
           //
        HashMap policyMap = new HashMap();
        packageMap = new HashMap();
        packageMap.put(Constants.GLOBAL_POLICY, policyMap );        // store the global policy map
        //
        // Setup initial default policies
        //      These policies are used during initialization and may be overwritten
        //      by the configuration files.
        //
        this.initialBuiltinPolicies();
     }
    
    // 
    // static methods
    //
    	
    public static PolicyManager getPolicyManager() 
    {
        return PolicyManager.instance();
    }
    
    public static PolicyManager instance()
    {
        if ( policyManager == null ) 
        {
            try
            {
                policyManager = new PolicyManager();
            }
            catch ( PolicyException pe ) 
            {
                // $FIX$ output the exeception message and log it if possible
                System.out.println(" Policy Exception -- Policy Manager Failure!");
                System.out.println( pe.getMessage() );
                System.exit(-9000);
            }
            catch ( DispatcherException de )
            {
                // $FIX$ output the exeception message and log it if possible
                System.out.println(" Policy Manager Exception -- Dispatcher Manager Failure !");
                System.out.println( de.getMessage() );
                System.exit(-9001);
            }
        }
        return PolicyManager.policyManager;
    }
    
    public static void initializeConfiguration() throws PolicyException, FileNotFoundException
    {
        //
        //
        //
        // read in the policies from the policy definition file.
        //
        //
        // 1.  get an XML parser for the Policy DTD.
        //
        String fileName = (String) policyManager.getSafeSelection("ConfigurationFileName", Constants.CONFIGURATION_FILENAME );
        String filePath = (String) policyManager.getSafeSelection("ConfigurationFilePath", Constants.CONFIGURATION_FILEPATH );
        // 
        File configFile = new File(filePath, fileName);
        //
        // 2.  Make sure that the configuration file exists
        //
        if ( !configFile.exists() )
        {
            // This is a FATAL icecubes error
            PolicyException fatalError = new PolicyException(
                "IceCubes Fatal Error" + "Configuration file, " + 
                Constants.CONFIGURATION_FILEPATH + System.getProperty("file.separator") + 
                Constants.CONFIGURATION_FILENAME + ", does not exist!  " 
                );
            throw fatalError;
        }
        //
        // 3.  Make sure that it is a file
        //
        if ( !configFile.isFile() )
        {
            // This is also a Fatal icecubes error
            PolicyException fatalError = new PolicyException(
                "IceCubes Fatal Error" + "Configuration file, " + 
                Constants.CONFIGURATION_FILEPATH + System.getProperty("file.separator") + 
                Constants.CONFIGURATION_FILENAME + ", is not a file!  " 
                );
            throw fatalError;
        }
        // construct an input stream for the file.
        FileReader configReader = new FileReader(configFile);
        // create a disassembler instance
        XMLDisassembler policyDisassembler = new XMLDisassembler("Policy", configReader);
        // construct the root element (tag) handler:
        Configuration iceCubesConfig = new Configuration( policyDisassembler, null );
        // register the first tag handler:
        policyDisassembler.registerHandler("ice-configuration", iceCubesConfig);
        // run the disassembler to parse and build the policies.
        // note that this must run synchronously to build the policies BEFORE
        // allowing other managers to run -- otherwise, they cannot follow the
        // configured policies.
        policyDisassembler.parse();

    }
	
    //
    // instance Methods 
    //
    
    /**
     * setup the initial default policies that are "builtin" to ICEcubes
     * @author Bruce Hunt 
     * @param void
     * @return void
     *
     **/
    private void initialBuiltinPolicies() throws PolicyException
    {
        //
        // Note that one or more of the policy setups here may throw a PolicyException.
        // If one is thrown, it is an ICEcubes fatal error.  
        //
        //
        // Setup the Logging Level
        //
        SimplePolicy loggingLevel = new SimplePolicy( Integer.class,  new Integer(9));
        this.setPolicy("LoggingLevel", loggingLevel);
        //
        // Setup the Configuration File Name and File Path:
        //
        SimplePolicy configurationFileName = new SimplePolicy(String.class, Constants.CONFIGURATION_FILENAME);
        this.setPolicy("ConfigurationFileName", configurationFileName);
        //
        SimplePolicy configurationFilePath = new SimplePolicy(String.class, Constants.CONFIGURATION_FILEPATH);
        this.setPolicy("ConfigurationFilePath", configurationFilePath);
        //
        // Setup the default policy for undefined attributes:
        //
        ChoicePolicy attributeNew = new ChoicePolicy( String.class );
            attributeNew.add("exception");
            attributeNew.add("add");
            attributeNew.add("ignore");
            attributeNew.select("add");
        this.setPolicy("AttributeNew", attributeNew);
        
    }
    /** 
     * get the current object representing the selected policy option for the package.
     * @author Bruce Hunt -- rewrote to support package names.
     * @param packageName -- name of package (or group of policies) this policy applies to.
     * @param policyName  -- name of the policy
     * @return the current object representing the policy or null if there is none.
     * @see getSafeSelection for a way to assure null is not returned.
     *
     **/
    public Object getSelection(String packageName, String policyName )
    {
        if ( packageMap.containsKey(packageName) )
        {
            HashMap packagePolicies = (HashMap) packageMap.get(packageName);
            if ( packagePolicies.containsKey(policyName) )
            {
                return ( (Policy)packagePolicies.get(policyName)).getSelection();
            }
        }
        return null;  // should a policyException be thrown here?
        // probably it should, but think of the typing overhead 
        // with all of the try catch clauses.
    }
    /** 
     * get the current object representing the selected global policy option.
     * @author Will Connant, Bruce Hunt -- rewrote to support package names.
     * @param policyName  -- name of the global policy
     * @return the current object representing the policy or null if there is none
     * @see getSelection for a way to assumre that null is not returned.
     *
     **/
    public Object getSelection( String policyName )
    {
        return getSelection( Constants.GLOBAL_POLICY, policyName );
    }
    
    /** 
     * get the current object that is the policy selection for the package; or if
     * the policy doesn't exist, return a default object.
     *.
     * @author Bruce Hunt -- to support package names.
     * @param packageName -- name of package (or group of policies) this policy applies to.
     * @param policyName  -- name of the policy
     * @param defaultObject -- object to be returned if no policy.
     * @return the current object representing the policy or null if there is none.
     * @see getSafeSelection for a way to assure null is not returned.
     *
     **/
    public Object getSafeSelection( String packageName, String policyName, Object defaultObject )
    {
        if ( this.hasPolicy( packageName, policyName ) )
        {
            return this.getPolicy(packageName, policyName).getSelection();
        }
        return defaultObject;
    }
    
    /** 
     * get the current object that is the global policy selection; or if
     * the policy doesn't exist, return a default object.
     *.
     * @author Bruce Hunt -- to support package names.
     * @param policyName  -- name of the policy
     * @param defaultObject -- object to be returned if no policy.
     * @return the current object representing the policy or null if there is none.
     * @see getSafeSelection for a way to assure null is not returned.
     *
     **/
	public Object getSafeSelection( String policyName, Object defaultObject )
	{
	    return getSafeSelection( Constants.GLOBAL_POLICY, policyName, defaultObject );
	}
    
    /** 
     * get the current policy for a package with a given name.
     * @author Will Connant, Bruce Hunt -- rewrote to support package names.
     * @param packageName -- name of package (or group of policies) this policy applies to.
     * @param policyName  -- name of the policy
     * @return the current object representing the policy or null if there is none.
     *      null can be returned for multiple reasons.  First, there may be no policies for
     *      the given packageName.  Second, there may be no policy with the given name in
     *      the package's policies.
     *
     **/
    public Policy getPolicy( String packageName, String policyName )
    {
        if ( packageMap.containsKey(packageName) )
        {
            HashMap packagePolicies = (HashMap) packageMap.get(packageName);
            if ( packagePolicies.containsKey(policyName) )
            {
                return (Policy) packagePolicies.get(policyName);
            }
        }
        return null;
    }
    
    /** 
     * get the current global policy with a given name.
     * @author Will Connant, Bruce Hunt -- rewrote to support package names.
     * @param policyName  -- name of the policy
     * @return the current object representing the policy or null if there is none.
     *      null can be returned for multiple reasons.  First, there may be no global policies
     *      (not likely). Second, there may be no policy with the given name in
     *      the global policies.
     *
     **/
    public Policy getPolicy( String policyName )
    {
        return getPolicy( Constants.GLOBAL_POLICY, policyName );
    }
    
    /** 
     * Is there a policy named policyName in the package named packageName?.
     * @author Bruce Hunt
     * @param packageName -- name of package (or group of policies) examining for this policy.
     * @param policyName  -- name of the policy
     * @return true if there is a policy named, policyName in the package named, packageName;
     *      false is returned otherwise.
     *
     **/
    public boolean hasPolicy( String packageName, String policyName )
    {
        if ( packageMap.containsKey(packageName) )
        {
            HashMap packagePolicies = (HashMap) packageMap.get(packageName);
            if ( packagePolicies.containsKey(policyName) )
            {
                return true;
            }
        }
        return false;
    }
    
    /** 
     * Is there a policy named policyName in the global policies list?
     * @author Bruce Hunt
     * @param packageName -- name of package (or group of policies) examining for this policy.
     * @param policyName  -- name of the policy
     * @return true if there is a policy named policyName in the package named packageName;
     *      false is returned otherwise.
     *
     **/
    public boolean hasPolicy( String policyName )
    {
        return hasPolicy( Constants.GLOBAL_POLICY, policyName );
    }
    
    /** 
     * set the current object that is the policy selection for the package; or if
     * the policy doesn't exist; or the object is the incorrect class, throw a PolicyException.
     *.
     * @author Bruce Hunt -- to support package names.
     * @param packageName -- name of package (or group of policies) this policy applies to.
     * @param policyName  -- name of the policy
     * @param object -- object to be the selection of the named policy in the named package.
     * @return void
     * @exception -- PolicyException 
     * @see getSafeSelection for a way to assure null is not returned.
     *
     **/
	public void select( String packageName, String policyName, Object o ) throws PolicyException
	{
	    this.getPolicy(packageName, policyName).select(o);
	}
	
    /** 
     * set the current object that is the global policy selection; or if
     * the policy doesn't exist; or the object is the incorrect class, throw a PolicyException.
     *.
     * @author Bruce Hunt -- to support package names.
     * @param policyName  -- name of the policy
     * @param object -- object to be the policy.selection.
     * @return void
     * @exception -- PolicyException 
     * @see getSafeSelection for a way to assure null is not returned.
     *
     **/
    public void select(String policyName, Object o) throws PolicyException 
    {
        this.select(Constants.GLOBAL_POLICY, policyName, o);
    }
	
    /** 
     * set the policy named policyName for the package named packageName.
     *.
     * @author Bruce Hunt -- to support package names.
     * @param packageName -- name of package (or group of policies) this policy applies to.
     * @param policyName  -- name of the policy
     * @param policy -- the Policy instance with name policyName
     * @return void
     *
     **/
    public void setPolicy( String packageName, String policyName, Policy policy )
    {
        HashMap packagePolicies;
        if ( ! packageMap.containsKey(packageName) )
        {
            packagePolicies = new HashMap();
            packageMap.put(packageName, packagePolicies );
        }
        else
            packagePolicies = (HashMap) packageMap.get(packageName);
        if ( packagePolicies.containsKey( policyName ) )
        {
            packagePolicies.remove( policyName );
        }
        packagePolicies.put( policyName, policy );
    }

    /** 
     * set the global policy named policyName.
     *.
     * @author Bruce Hunt -- to support package names.
     * @param policyName  -- name of the policy
     * @param policy -- the Policy instance with name policyName
     * @return void
     *
     **/
    public void setPolicy(String policyName, Policy policy) 
    {
        this.setPolicy(Constants.GLOBAL_POLICY, policyName, policy);
    }

    
    //
    // Execution Activity Logic
    //
    public void preInitialize() { }
    public void initialize() { }
    public void postInitialize() { }
    //
    public void run() { }
    //
    public void preFinalize() { }
    public void finalize() { }
    public void postFinalize() { }

    //
    // Implement core manager registration policy
    //
    public void managerRegister( Tray tray ) 
    {
        // Step one:  Register the built-in managers;
        managerPolicies = new ManagerPolicies();
        
        managerPolicies.setTray(tray);
        
        managerPolicies.initialize();

    }

}





