/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/common/scheduling/DispatchSchedule.java,v 1.2 2000/11/27 20:19:41 bhunt Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

*/

package com.adobe.rice.common.scheduling;
/** 
 *  DispatchSchedule contans ordered ScheduleItems which should be executed 
 *  `soon', viz. policies. Unlike the PerishableDispatchSchedule, once run, 
 *  this will not terminate. It will just wait until it is provided 
 *  DispatchScheduleItems to run.
 **/

import java.util.*;
import com.adobe.rice.util.time.*;

public class DispatchSchedule extends Thread
{
	private IceDateTime birth;
	private List items;

    public DispatchSchedule( )
    {
        this.birth = new IceDateTime();     // assume it is born now!
        items = new LinkedList();
    }
    
	public DispatchSchedule(IceDateTime birth) 
	{
		this.birth=birth;   // assume it was born then.
		items= new LinkedList();
	}
	
	/**
	 * inserts a DispatchScheduleItem into this DispatchSchedule.  
	 * This routine is synchronized around the core schedule so that
	 * the DispatchSchedule is not corrupted.
	 *
	 * @param item - the DispatchScheduleItem to insert
	 *
	 * @author J. Herbach
	 *
	 **/
	public void insert(DispatchScheduleItem item) 
	{ // "producer"
		synchronized (items) 
		{
			boolean alreadyexists=false;
			ListIterator it = items.listIterator();				
			while (it.hasNext())
			{
				DispatchScheduleItem cur = (DispatchScheduleItem)it.next();
				if (cur.equals(item)) 
				{
					alreadyexists= true; // do not add
					break;
				}
				if (item.getStarttime().before(cur.getStarttime())) 
					break; // and add
			}
			if (!alreadyexists)
				it.add(item);
			items.notifyAll(); // tell the consumer that life might have changed.
		}
	}
	/**
	 * Runs the DispatchSchedule forever.
	 *
	 * @param - none
	 *
	 * @author - J. Herbach
	 *
	 * @notes:
	 *
	 *  $FIX$ items needed:
	 *  1.  The dispatch scheduler needs to have a serializer
	 *  to save the current state of the schedule.
	 *  2.  This thread ought to pay attention to some sort of
	 *  interruption to signal that it is to terminate and 
	 *  then gracefully complete all dispatched threads and
	 *  then serialize its schedule and finally terminate.
	 *  It should also announce on a regular basis how many
	 *  threads and/or schedule items it has left to do
	 *  for termination.
	 *  3.  There ought to be an "abort" interruption that
	 *  tells the dispatch schedule to abort all runners, 
	 *  serialize the schedule for storage and terminate.
	 *
	 **/
	
	public void run() 
	{ // "consumer"
		while (true) 
		{ // unlike a PerishableDispatchSchedule, never ends!
			// isItemReady() will wait() and release items.
			synchronized (items) 
			{ 
				// wait until we are guarenteed something to run *NOW*
				while (isItemReady() == false) 
				{
				    // We should not go run others while we wait because
				    // isItemReady waits for us.
				    // this.yield();
				}
				
				ListIterator it = items.listIterator();
				DispatchScheduleItem  first = (DispatchScheduleItem) it.next();
				first.execute();
				it.remove(); // remove the first item, as it was just executed!
			}
		}		
	}
	
	/** 
	 * Waits until it is time to consume the first item (possibly not provided yet!)
	 *
	 * @param - none
	 * @returns boolean - true if an item is ready
	 *
	 * @author J. Herbach
	 *
	 * @see object.wait for how wait works.
	 *.
	 **/
	private boolean isItemReady() 
	{
		Iterator checkIt = items.iterator();
		
		// if the queue is empty, then just wait until it isn't empty!
		if (!checkIt.hasNext()) 
		{   // wait "forever" -- i.e., until something deposited.
			try { items.wait(); }
			catch (InterruptedException e) {}
		}
		else 
		{
			DispatchScheduleItem first = (DispatchScheduleItem) checkIt.next();
			IceDateTime now = new IceDateTime();
			// if we have to wait, then wait accordingly.
			if (first.getStarttime().before(now)) 
			    return true; // we have something to process!
			// If we have to wait, then wait accordingly.
			// What happens if another DSI is inserted before first while we are
			// waiting?  We get notified when it happens and we resume even
			// if the total time has NOT elapsed.
			try 
			{
			    // wait for notification or timeout
				items.wait(first.getStarttime().subtract(now).getTotal());
			} 
			catch (InterruptedException e) {}
			// we're not sure if we have loop more
		}
		return false; //
	}
}
			
	