/*
$Log: IceCubeManager.java,v $
Revision 1.5  2000/11/27 20:21:34  bhunt
Updates

Revision 1.4  2000/11/22 21:19:11  bhunt
Updates for IceNOP

Revision 1.3  2000/11/01 02:39:07  andrewf
added constants

Revision 1.2  2000/10/25 22:04:27  andrewf
added documentation and instructions

$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/kernel/IceCubeManager.java,v 1.5 2000/11/27 20:21:34 bhunt Exp $
$Author: bhunt $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org

   The IceCubeManager is the abstract base class for all Managers
   in an ICE Processor. An IceCubeManager have a name, at least, one 
   mailbox, and one thread of execution associated with them. Moreover
   IceCubeManagers inherit all the "inter-process-communications" 
   mechanisms needed to send and receive messages. 
   
   NOTE one: this implementor thinks of IceCubeManagers as being
   akin to a task or sub-system in the Mach operating system. 

   NOTE two: many of the descriptions here are repeated in
   MessageDispatcher since IceCubeManagers depend on the
   former to send and receive messages.

 *<dl>
 *   <dt>createMailbox (String name) </dt>
 *      <dd>A mailbox is a queue of messages. Mailboxes not
 *           Managers, are the recipient of all communications.
 *           In turn, managers obtain messages by doing a
 *           receive on a specific mailbox the manager created.
 *           createMailbox allows a Manager to create a mailbox 
 *           with a specific identifier (usually Manager's name).
 *
 *
 *   <dt>destroyMailbox</dt>
 *      <dd>Destroy a specific mailbox</dd>
 *
 *   <dt>asyncSend</dt>
 *      <dd>Send a message asynchronously. asyncSend is
 *           used for one-way communications and when one
 *           does not wish to wait for a reply. In order
 *           to associate potential replies, a transaction
 *           id is used.</dd>
 *
 *   <dt>synchronousSend ()</dt>
 *       <dd> Send a message and wait for a response. This
 *            method is provided synchronous and blocking 
 *            communications</dd> The intended user of
 *            synchronousSend are threads that can afford
 *            to block without affecting other threads and
 *            managers.
 *
 *   <dt>asyncReceive</dt>
 *      <dd> Non blocking read of mailbox. Note: this 
 *            function name should be renamed. Also this 
 *            function assumes the mailbox name is the same 
 *            as the object calling the routine.</dd>
 *
 *   <dt>asyncReceive (String name)</dt>
 *      <dd> General form of asyncReceive. Do a receive
 *            on a mailbox with the specified name (provided
 *            the manager is authorized to do this) </dd>
 *
 *   <dt>receive ()</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           same name as caller. </dd>
 *
 *   <dt>receive (long timeout)</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           same name as caller. However the receive will
 *           return either if a message is returned in the
 *           appropriate time, or the time expires</dd>
 *
 *   <dt>receive (String name)</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           specified name. </dd>
 *
 *   <dt>receive (String name, long timeout)</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           specified name. However the receive will
 *           return either if a message is returned in the
 *           appropriate time, or the time expires</dd>
 *
 *   <dt>reply (IceMessage) </dt>
 *   <dd> reply to a received message. Preferred way to
 *        answer received messages, since reply relies on
 *        the instance variables in the IceMessage object</dd>
 *  
 *   <dt> registerCommand (String name, Mailbox)</dt>
 *   <dd> registerCommand associates a string with a mailbox
 *        registered with the messageDispatcher. RegisterCommand
 *        is used to help dispatch messages to a mailbox where
 *        the sender has an expectation that the message will
 *        be processed </dd>
 *
 *   <dt> deregisterCommand (String mailboxName) </dt>
 *   <dd> deregister the Command </dd>
 *
 *   <dt> getIceAction (String commandName) </dt>
 *   <dd> get an integer value associated with a string name.
 *        The purpose of getIceAction is to aid a manager in
 *        processing a message. </dd>
 *
 *  </dl>
 */


package com.adobe.rice.kernel;
import com.adobe.rice.common.*;


public abstract class IceCubeManager implements Manager
{
    private Thread            thread;
    private String            name;
    private MessageDispatcher dispatcher;
    
    public IceCubeManager (String name, MessageDispatcher dispatcher)
                           throws DispatcherException
    {
        // Managers must have names and a dispatcher
        if (name == null || dispatcher == null)
           throw new DispatcherException ("null arguments not permitted");
           
        this.name = name;
        this.dispatcher = dispatcher;
    
        //create a mail queue associated with the manager
        dispatcher.createMailbox (name);
    }
    
    public IceCubeManager( String name ) throws DispatcherException
    {
        // Managers must have names and a dispatcher
        if (name == null )
           throw new DispatcherException ("null arguments not permitted");
           
        this.name = name;
        this.dispatcher = MessageDispatcher.instance();
    
        //create a mail queue associated with the manager
        dispatcher.createMailbox (name);
    }
        
    
    public abstract void finalize ();
    
       
    protected void CreateMailbox (String name)
    {
        dispatcher.createMailbox (name);
    }
    
    protected void DestroyMailbox (String name)
    {
        dispatcher.destroyMailbox (name);
    }
    
    // should throw an exception on a null name or message
    protected long send (String name, IceMessage message)
    {
        return dispatcher.asyncSend (name, message);    
    }
    
    protected IceMessage receive () throws DispatcherException
    {
        return dispatcher.receive ();
    }
    
    protected IceMessage receive (long timeout) 
                         throws DispatcherException
    {
        return dispatcher.receive (timeout);
    }
    
    protected IceMessage receive (String queueName)
                         throws DispatcherException
    {
        return dispatcher.receive (queueName);
    }
    
    protected IceMessage receive (String queueName,
                                  long   timeout)
                         throws DispatcherException
    {
        return dispatcher.receive (queueName, timeout);
    } 
    
    
    protected IceMessage synchronousSend (String     queueName, 
                                          IceMessage message)
                         throws DispatcherException
    {
        return dispatcher.synchronousSend (queueName, message);  
        
    }
    
    protected IceMessage synchronousSend (String     queueName, 
                                          IceMessage message,
                                          long       msecs)
                         throws DispatcherException
    {
        return dispatcher.synchronousSend (queueName, message, msecs);          
    }
    
    protected long reply (IceMessage message)
    {
        return dispatcher.reply (message);   
    }
    
    protected int getIceAction (String command)
    {
        return dispatcher.getIceAction (command);
    }
    
    protected boolean registerCommand (String command,
                                       String mailBoxName)
    {
        return dispatcher.registerCommand (command, mailBoxName);
    }
              
    protected boolean deregisterCommand (String command)
    {
        return dispatcher.deregisterCommand (command);
    }
    
    protected String getCommand (String name)
    {
        return dispatcher.getCommand (name);
    }
    
    public String getName ()
    {
        return name;
    }
    
    public Thread setThread( Thread th )
    {
        Thread oldThread = thread;
        this.thread = th;
        return oldThread;
    }
    
    public Thread getThread() { return thread; }
}