/*
$Header: /opt/repository/icecubes/Rice/code/com/adobe/rice/kernel/MessageDispatcher.java,v 1.2 2000/10/25 22:04:27 andrewf Exp $

ICEcubes - Reference implementation of the Information and Content Exchange Protocol
Copyright (C) 1998 - 2000 ICE Authoring Group

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

The ICE Authoring Group thanks Adobe System's Inc. for supporting 
the initial development of ICEcubes.

This library was written on behalf of the ICE Authoring Group, affialiated
with GCA's IdeaAlliance.  For the latest information on ICE see 

http://www.icestandard.org


*/


package com.adobe.rice.kernel;
import java.util.*;

/*
   $Log: MessageDispatcher.java,v $
   Revision 1.2  2000/10/25 22:04:27  andrewf
   added documentation and instructions

   
   The MessageDispatcher handles communications between
   IceCube's various managers.

   The MessageDispatcher provides the low level primitives
   for supporting both synchronous and asynchronous sending 
   and receiving messages between managers and threads.
   
   Note: It is recommended that manager writers "hide" 
   the underlying message passing mechanics, so that
   objects believe they are making normal method calls
   rather than a MessageDispatcher call. i.e
   
   doSomething (A a, B b, C c)
   {
       synchronousSend (some_struct_with a b c)
   }

 *<dl>
 *   <dt>createMailbox (String name) </dt>
 *      <dd>A mailbox is a queue of messages. Mailboxes not
 *           Managers, are the recipient of all communications.
 *           In turn, managers obtain messages by doing a
 *           receive on a specific mailbox the manager created.
 *           createMailbox allows a Manager to create a mailbox 
 *           with a specific identifier (usually Manager's name).
 *
 *
 *   <dt>destroyMailbox</dt>
 *      <dd>Destroy a specific mailbox</dd>
 *
 *   <dt>asyncSend</dt>
 *      <dd>Send a message asynchronously. asyncSend is
 *           used for one-way communications and when one
 *           does not wish to wait for a reply. In order
 *           to associate potential replies, a transaction
 *           id is used.</dd>
 *
 *   <dt>synchronousSend ()</dt>
 *       <dd>Send a message and wait for a response. This
 *            method is provided synchronous and blocking 
 *            communications</dd> The intended user of
 *            synchronousSend are threads that can afford
 *            to block without affecting other threads and
 *            managers.
 *
 *   <dt>asyncReceive</dt>
 *      <dd> Non blocking read of mailbox. Note: this 
 *            function name should be renamed. Also this 
 *            function assumes the mailbox name is the same 
 *            as the object calling the routine.</dd>
 *
 *   <dt>asyncReceive (String name)</dt>
 *      <dd> General form of asyncReceive. Do a receive
 *            on a mailbox with the specified name (provided
 *            the manager is authorized to do this) </dd>
 *
 *   <dt>receive ()</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           same name as caller. </dd>
 *
 *   <dt>receive (long timeout)</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           same name as caller. However the receive will
 *           return either if a message is returned in the
 *           appropriate time, or the time expires</dd>
 *
 *   <dt>receive (String name)</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           specified name. </dd>
 *
 *   <dt>receive (String name, long timeout)</dt>
 *      <dd> Synchronous, blocking receive on mailbox with
 *           specified name. However the receive will
 *           return either if a message is returned in the
 *           appropriate time, or the time expires</dd>
 *
 *   <dt>reply </dt>
 *   <dd> reply to a received message. Preferred way to
 *        answer received messages </dd>
 *  
 *   <dt> registerCommand (String name, Mailbox)</dt>
 *   <dd> registerCommand associates a string with a mailbox
 *        registered with the messageDispatcher. RegisterCommand
 *        is used to help dispatch messages to a mailbox where
 *        the sender has an expectation that the message will
 *        be processed </dd>
 *
 *   <dt> deregisterCommand (String mailboxName) </dt>
 *   <dd> deregister the Command </dd>
 *
 *   <dt> getIceAction (String commandName) </dt>
 *   <dd> get an integer value associated with a string name.
 *        The purpose of getIceAction is to aid a manager in
 *        processing a message. </dd>
 *
 *  </dl>
 */

public class MessageDispatcher
{
    // IceDispatcher is a singleton class
    private static MessageDispatcher dispatcher;
    
    // all named mailboxes (queues,slots) 
    private Hashtable mailboxTable;
          
    // October 12th addition, IceDispatcher
    // will handle the mapping of commands
    // to queues
    private Hashtable commandTable;
                        
    private long transactionId = 0;
    
    private long generateTransactionId ()
    {
        return ++transactionId;
    }
   
    // on creation, a single instance is created
    static 
    {
        dispatcher = new MessageDispatcher ();
    }                                   
     
        
    // underlying primitive for receive    
    private IceMessage _receive (Object key, long timeout)
                       throws DispatcherException
    {
        Mailbox    target;
        IceMessage message;
        
        // get the mailbox
        if ((target = (Mailbox) mailboxTable.get (key)) == null)
        {
           throw new DispatcherException (" key not found");
        }
        
        try
        {
           if (timeout == Constants.BLOCK)
           {
              // block
              message = (IceMessage) target.remove ();      
           }
           else if (timeout == Constants.POLL)
           {
              // return immediately
              message = (IceMessage) target.asyncRemove (); 
           }
           else if (timeout < Constants.BLOCK)
           {
              throw new DispatcherException ("negative time");
           }
           else
           {
              message = (IceMessage) target.remove (timeout);
           }
        }
        catch (InterruptedException e)
        {
            message = null;
        } 
        
        return message;
    }                                  
                                                                             
    
    // underlying primitive for synchronous
    private IceMessage _syncSend (String     name, 
                                  IceMessage message,
                                  long       timeout)
                                  throws DispatcherException 
    {
        IceMessage newMessage = null;
        Object     returnAddress;
        Mailbox    myMailBox;
        Mailbox    mailBox;
        
         // if the mailbox does not exist then return an error
        if ((mailBox = (Mailbox) mailboxTable.get (name)) == null)
        {
           return null;
        }
        
        
        returnAddress = Thread.currentThread ();
        register ((Thread) returnAddress);
        message.setReturnAddress (returnAddress);
        message.setSender (Thread.currentThread().getName ());
        
        // okay "send" the message
        mailBox.add(message);
         
        // now do a receive first by getting my return mailslot
        myMailBox = (Mailbox) mailboxTable.get (returnAddress);
        
        try
        {
           if (timeout == Constants.BLOCK)
           {
              newMessage = (IceMessage) myMailBox.remove ();
           }
           else if (timeout == Constants.POLL)
           {
              newMessage = (IceMessage) myMailBox.asyncRemove ();
           }
           else if (timeout < -1)
           {
              throw new DispatcherException ("invalid timeout");
           }
           else
           {
              newMessage = (IceMessage) myMailBox.remove (timeout);
              if (newMessage == null)
              {
                 newMessage = new IceMessage (true);
              }
           }
        }
        catch (InterruptedException e)
        {
            newMessage = null;
        }
        
        // just in case a message arrived between returning from
        // timeout and before deregistering.
        
        if (newMessage.hasTimedout ())
        {
           IceMessage temp = asyncReceive ();
           if (temp != null)
           {
              newMessage = temp;
           }
        }
        
        // deregister the thread
        deregister ((Thread) returnAddress);
       
          
        // and return the message
        return newMessage;
    }                                                                        
                                                                                                                                                       
                                                                             
    protected MessageDispatcher ()
    {
        mailboxTable = new Hashtable ();
        commandTable = new Hashtable ();
        transactionId = 0;
    }
    
    
    public static MessageDispatcher instance ()
    {
        return dispatcher;
    }
    
    
    // registers the Thread with the dispatcher
    public void register (Thread thread)
    {
        mailboxTable.put (Thread.currentThread (), 
                          new MailSlot ());
    }
    
    
    // deregister the Thread
    public void deregister (Thread thread)
    { 
        mailboxTable.remove (thread);
    }
    
   
    public void createMailbox (String name)
    {
        mailboxTable.put (name, new MailQueue ());
    }
    
    
    public void destroyMailbox (String name)
    {
        mailboxTable.remove(name);
    }
    
    
    public long asyncSend (Object name, IceMessage message)
    {
        long    transactionId = -1;
        String  returnAddress;
        Mailbox mailbox;
        
        // if the mailbox does not exist then return an error
        if ((mailbox = (Mailbox) mailboxTable.get (name)) == null)
        {
           return -1;
        }
        
        // if message is new, generate an unique id
        if (message.getTransactionId () != -1)
        {
           transactionId = generateTransactionId ();   
        }
        
        // use the manager's name as the identify
        returnAddress = Thread.currentThread ().getName ();
        message.setSender (returnAddress);
        message.setReturnAddress (returnAddress);
        message.setTransactionId (transactionId);

        // okay "send" the message
        mailbox.add (message);
        
        return transactionId;
    }
 
 
    // blocking synchronous call
    public IceMessage synchronousSend (String name, 
                                       IceMessage message)
                      throws DispatcherException 
    { 
        return _syncSend (name, message, Constants.BLOCK);
    }
    
   
    public IceMessage synchronousSend (String     name, 
                                       IceMessage message, 
                                       long       timeout)
                                       throws DispatcherException
    {
        return _syncSend (name, message, timeout);
    }
    
    
    // asynchronous receive - assume receiving 
    // runner's queue
    public IceMessage asyncReceive ()
                      throws DispatcherException
    {
        String name = Thread.currentThread ().getName ();     
        return _receive (name, Constants.POLL);
    }
    
    
    // asynchronous receive from a specific mailbox
    public IceMessage asyncReceive (String mailbox)
                      throws DispatcherException
    {
        return _receive (mailbox, Constants.POLL);
    }
    
    
    // synchronous receive from reply queue
    public IceMessage receive ()
                      throws DispatcherException
    {
        return _receive (Thread.currentThread ().getName (), 
                         Constants.BLOCK);
    }
    
    
    // synchronous receive from reply queue
    public IceMessage receive (long timeout)
                      throws DispatcherException
    {
        return _receive (Thread.currentThread ().getName (), 
                         timeout);
    }     
   
   
    public IceMessage receive (String mailboxName)
                      throws DispatcherException
    {
        return _receive (mailboxName, Constants.BLOCK);
    }
    
    
    // synchronous receive from a specific mailbox with a timeout
    public IceMessage receive (String mailboxName, long timeout)
                      throws DispatcherException
    {
        return _receive (mailboxName, timeout);
    }   
    
    // reply is synchronous
    public long reply (IceMessage message)
    {  
        Mailbox    mailBox;
        long       result;
               
         // if the mailbox does not exist then return an error
        mailBox = (Mailbox) mailboxTable.get (message.getReturnAddress ()); 
        
        if (mailBox == null)
        {
            result = -2;
        }
        else
        {
            // don't want a response
            message.setReturnAddress (null);
            message.setSender (Thread.currentThread().getName ());
        
            // okay "send" the message
            mailBox.add(message);
            result = message.getTransactionId ();
        } 
        
        return (result);
    }
    
    
    public boolean registerCommand (String command,
                                    String mailBoxName)
    {
        Mailbox mailbox;
        boolean result = false;
        String  key;
         
        if (mailboxTable.get (mailBoxName) == null)
        {
           result = false; 
        }
        else
        {
           // normalize strings to all lowercase
           key = command.toLowerCase(); 
           
           // associate command with mailbox
           commandTable.put (key, mailBoxName);
           
           result = true;
        }
        
        return result;
    }
         
         
    public boolean deregisterCommand (String command)
    {
        return (commandTable.remove (command) == null) ? false : true;
    }
         
         
    public String getCommand (String name)
    {
        return (String) commandTable.get (name.toLowerCase ());
    }
    
    
    //
    // will do timing tests on this one to see if it is 
    // quicker than a hashtable since hashtables synchronize
    // and this is essentially an immutable data structure.
    // If so this may give a clue to other tables. Perhaps
    // I should also toy with a binary search.
    public int getIceAction (String name)
    {
        int    result = -1;
        String commandName = name.toLowerCase();
 
        for (int n = 0 ; n < Constants.commands.length ; n++)
        {
            if (Constants.commands[n].equals (commandName))
            {
               result = n;
               break;
            }
        }
       
        return result;
    }
}